;;; julia-mode-latexsubs.el --- Generated variables for Julia mode -*- lexical-binding: t; -*-
;; This file is automatically generated. Do not edit by hand.

(define-abbrev-table 'julia-latexsub-abbrev-table ()
  "Abbrev table for Julia LaTeX substitution."
  :regexp "\\(\\\\[[:graph:]]+\\)")

(defconst julia-mode-latexsubs
  (let ((table (make-hash-table :test 'equal)))
    (puthash "\\exclamdown" "¡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\exclamdown" "¡" nil :system t)
    (puthash "\\sterling" "£" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sterling" "£" nil :system t)
    (puthash "\\yen" "¥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\yen" "¥" nil :system t)
    (puthash "\\brokenbar" "¦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\brokenbar" "¦" nil :system t)
    (puthash "\\S" "§" table)
    (define-abbrev julia-latexsub-abbrev-table "\\S" "§" nil :system t)
    (puthash "\\copyright" "©" table)
    (define-abbrev julia-latexsub-abbrev-table "\\copyright" "©" nil :system t)
    (puthash "\\:copyright:" "©" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:copyright:" "©" nil :system t)
    (puthash "\\ordfeminine" "ª" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ordfeminine" "ª" nil :system t)
    (puthash "\\guillemotleft" "«" table)
    (define-abbrev julia-latexsub-abbrev-table "\\guillemotleft" "«" nil :system t)
    (puthash "\\neg" "¬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neg" "¬" nil :system t)
    (puthash "\\:registered:" "®" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:registered:" "®" nil :system t)
    (puthash "\\circledR" "®" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledR" "®" nil :system t)
    (puthash "\\highminus" "¯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\highminus" "¯" nil :system t)
    (puthash "\\degree" "°" table)
    (define-abbrev julia-latexsub-abbrev-table "\\degree" "°" nil :system t)
    (puthash "\\pm" "±" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pm" "±" nil :system t)
    (puthash "\\^2" "²" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^2" "²" nil :system t)
    (puthash "\\^3" "³" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^3" "³" nil :system t)
    (puthash "\\P" "¶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\P" "¶" nil :system t)
    (puthash "\\cdotp" "·" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cdotp" "·" nil :system t)
    (puthash "\\^1" "¹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^1" "¹" nil :system t)
    (puthash "\\ordmasculine" "º" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ordmasculine" "º" nil :system t)
    (puthash "\\guillemotright" "»" table)
    (define-abbrev julia-latexsub-abbrev-table "\\guillemotright" "»" nil :system t)
    (puthash "\\1/4" "¼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/4" "¼" nil :system t)
    (puthash "\\1/2" "½" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/2" "½" nil :system t)
    (puthash "\\3/4" "¾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\3/4" "¾" nil :system t)
    (puthash "\\questiondown" "¿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\questiondown" "¿" nil :system t)
    (puthash "\\AA" "Å" table)
    (define-abbrev julia-latexsub-abbrev-table "\\AA" "Å" nil :system t)
    (puthash "\\AE" "Æ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\AE" "Æ" nil :system t)
    (puthash "\\DH" "Ð" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DH" "Ð" nil :system t)
    (puthash "\\times" "×" table)
    (define-abbrev julia-latexsub-abbrev-table "\\times" "×" nil :system t)
    (puthash "\\O" "Ø" table)
    (define-abbrev julia-latexsub-abbrev-table "\\O" "Ø" nil :system t)
    (puthash "\\TH" "Þ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\TH" "Þ" nil :system t)
    (puthash "\\ss" "ß" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ss" "ß" nil :system t)
    (puthash "\\aa" "å" table)
    (define-abbrev julia-latexsub-abbrev-table "\\aa" "å" nil :system t)
    (puthash "\\ae" "æ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ae" "æ" nil :system t)
    (puthash "\\eth" "ð" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eth" "ð" nil :system t)
    (puthash "\\dh" "ð" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dh" "ð" nil :system t)
    (puthash "\\div" "÷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\div" "÷" nil :system t)
    (puthash "\\o" "ø" table)
    (define-abbrev julia-latexsub-abbrev-table "\\o" "ø" nil :system t)
    (puthash "\\th" "þ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\th" "þ" nil :system t)
    (puthash "\\DJ" "Đ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DJ" "Đ" nil :system t)
    (puthash "\\dj" "đ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dj" "đ" nil :system t)
    (puthash "\\hbar" "ħ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hbar" "ħ" nil :system t)
    (puthash "\\imath" "ı" table)
    (define-abbrev julia-latexsub-abbrev-table "\\imath" "ı" nil :system t)
    (puthash "\\L" "Ł" table)
    (define-abbrev julia-latexsub-abbrev-table "\\L" "Ł" nil :system t)
    (puthash "\\l" "ł" table)
    (define-abbrev julia-latexsub-abbrev-table "\\l" "ł" nil :system t)
    (puthash "\\NG" "Ŋ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NG" "Ŋ" nil :system t)
    (puthash "\\ng" "ŋ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ng" "ŋ" nil :system t)
    (puthash "\\OE" "Œ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\OE" "Œ" nil :system t)
    (puthash "\\oe" "œ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oe" "œ" nil :system t)
    (puthash "\\hvlig" "ƕ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hvlig" "ƕ" nil :system t)
    (puthash "\\nrleg" "ƞ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nrleg" "ƞ" nil :system t)
    (puthash "\\Zbar" "Ƶ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Zbar" "Ƶ" nil :system t)
    (puthash "\\doublepipe" "ǂ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\doublepipe" "ǂ" nil :system t)
    (puthash "\\jmath" "ȷ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\jmath" "ȷ" nil :system t)
    (puthash "\\trna" "ɐ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trna" "ɐ" nil :system t)
    (puthash "\\trnsa" "ɒ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnsa" "ɒ" nil :system t)
    (puthash "\\openo" "ɔ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\openo" "ɔ" nil :system t)
    (puthash "\\rtld" "ɖ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtld" "ɖ" nil :system t)
    (puthash "\\schwa" "ə" table)
    (define-abbrev julia-latexsub-abbrev-table "\\schwa" "ə" nil :system t)
    (puthash "\\pgamma" "ɣ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pgamma" "ɣ" nil :system t)
    (puthash "\\pbgam" "ɤ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pbgam" "ɤ" nil :system t)
    (puthash "\\trnh" "ɥ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnh" "ɥ" nil :system t)
    (puthash "\\btdl" "ɬ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\btdl" "ɬ" nil :system t)
    (puthash "\\rtll" "ɭ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtll" "ɭ" nil :system t)
    (puthash "\\trnm" "ɯ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnm" "ɯ" nil :system t)
    (puthash "\\trnmlr" "ɰ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnmlr" "ɰ" nil :system t)
    (puthash "\\ltlmr" "ɱ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltlmr" "ɱ" nil :system t)
    (puthash "\\ltln" "ɲ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltln" "ɲ" nil :system t)
    (puthash "\\rtln" "ɳ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtln" "ɳ" nil :system t)
    (puthash "\\clomeg" "ɷ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\clomeg" "ɷ" nil :system t)
    (puthash "\\ltphi" "ɸ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltphi" "ɸ" nil :system t)
    (puthash "\\trnr" "ɹ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnr" "ɹ" nil :system t)
    (puthash "\\trnrl" "ɺ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnrl" "ɺ" nil :system t)
    (puthash "\\rttrnr" "ɻ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rttrnr" "ɻ" nil :system t)
    (puthash "\\rl" "ɼ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rl" "ɼ" nil :system t)
    (puthash "\\rtlr" "ɽ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtlr" "ɽ" nil :system t)
    (puthash "\\fhr" "ɾ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fhr" "ɾ" nil :system t)
    (puthash "\\rtls" "ʂ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtls" "ʂ" nil :system t)
    (puthash "\\esh" "ʃ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\esh" "ʃ" nil :system t)
    (puthash "\\trnt" "ʇ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trnt" "ʇ" nil :system t)
    (puthash "\\rtlt" "ʈ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtlt" "ʈ" nil :system t)
    (puthash "\\pupsil" "ʊ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pupsil" "ʊ" nil :system t)
    (puthash "\\pscrv" "ʋ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pscrv" "ʋ" nil :system t)
    (puthash "\\invv" "ʌ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\invv" "ʌ" nil :system t)
    (puthash "\\invw" "ʍ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\invw" "ʍ" nil :system t)
    (puthash "\\trny" "ʎ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trny" "ʎ" nil :system t)
    (puthash "\\rtlz" "ʐ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtlz" "ʐ" nil :system t)
    (puthash "\\yogh" "ʒ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\yogh" "ʒ" nil :system t)
    (puthash "\\glst" "ʔ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\glst" "ʔ" nil :system t)
    (puthash "\\reglst" "ʕ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\reglst" "ʕ" nil :system t)
    (puthash "\\inglst" "ʖ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\inglst" "ʖ" nil :system t)
    (puthash "\\turnk" "ʞ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\turnk" "ʞ" nil :system t)
    (puthash "\\dyogh" "ʤ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dyogh" "ʤ" nil :system t)
    (puthash "\\tesh" "ʧ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tesh" "ʧ" nil :system t)
    (puthash "\\^h" "ʰ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^h" "ʰ" nil :system t)
    (puthash "\\^j" "ʲ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^j" "ʲ" nil :system t)
    (puthash "\\^r" "ʳ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^r" "ʳ" nil :system t)
    (puthash "\\^w" "ʷ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^w" "ʷ" nil :system t)
    (puthash "\\^y" "ʸ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^y" "ʸ" nil :system t)
    (puthash "\\rasp" "ʼ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rasp" "ʼ" nil :system t)
    (puthash "\\verts" "ˈ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\verts" "ˈ" nil :system t)
    (puthash "\\verti" "ˌ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\verti" "ˌ" nil :system t)
    (puthash "\\lmrk" "ː" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lmrk" "ː" nil :system t)
    (puthash "\\hlmrk" "ˑ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hlmrk" "ˑ" nil :system t)
    (puthash "\\sbrhr" "˒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sbrhr" "˒" nil :system t)
    (puthash "\\sblhr" "˓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sblhr" "˓" nil :system t)
    (puthash "\\rais" "˔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rais" "˔" nil :system t)
    (puthash "\\low" "˕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\low" "˕" nil :system t)
    (puthash "\\u" "˘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\u" "˘" nil :system t)
    (puthash "\\tildelow" "˜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tildelow" "˜" nil :system t)
    (puthash "\\^l" "ˡ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^l" "ˡ" nil :system t)
    (puthash "\\^s" "ˢ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^s" "ˢ" nil :system t)
    (puthash "\\^x" "ˣ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^x" "ˣ" nil :system t)
    (puthash "\\grave" "̀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\grave" "̀" nil :system t)
    (puthash "\\acute" "́" table)
    (define-abbrev julia-latexsub-abbrev-table "\\acute" "́" nil :system t)
    (puthash "\\hat" "̂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hat" "̂" nil :system t)
    (puthash "\\tilde" "̃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tilde" "̃" nil :system t)
    (puthash "\\bar" "̄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bar" "̄" nil :system t)
    (puthash "\\overbar" "̅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\overbar" "̅" nil :system t)
    (puthash "\\breve" "̆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\breve" "̆" nil :system t)
    (puthash "\\dot" "̇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dot" "̇" nil :system t)
    (puthash "\\ddot" "̈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ddot" "̈" nil :system t)
    (puthash "\\ovhook" "̉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ovhook" "̉" nil :system t)
    (puthash "\\ocirc" "̊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ocirc" "̊" nil :system t)
    (puthash "\\H" "̋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\H" "̋" nil :system t)
    (puthash "\\check" "̌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\check" "̌" nil :system t)
    (puthash "\\candra" "̐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\candra" "̐" nil :system t)
    (puthash "\\oturnedcomma" "̒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oturnedcomma" "̒" nil :system t)
    (puthash "\\ocommatopright" "̕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ocommatopright" "̕" nil :system t)
    (puthash "\\droang" "̚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\droang" "̚" nil :system t)
    (puthash "\\palh" "̡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\palh" "̡" nil :system t)
    (puthash "\\rh" "̢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rh" "̢" nil :system t)
    (puthash "\\c" "̧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\c" "̧" nil :system t)
    (puthash "\\k" "̨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\k" "̨" nil :system t)
    (puthash "\\sbbrg" "̪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sbbrg" "̪" nil :system t)
    (puthash "\\wideutilde" "̰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wideutilde" "̰" nil :system t)
    (puthash "\\underbar" "̲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underbar" "̲" nil :system t)
    (puthash "\\strike" "̶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\strike" "̶" nil :system t)
    (puthash "\\sout" "̶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sout" "̶" nil :system t)
    (puthash "\\not" "̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\not" "̸" nil :system t)
    (puthash "\\underleftrightarrow" "͍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underleftrightarrow" "͍" nil :system t)
    (puthash "\\Alpha" "Α" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Alpha" "Α" nil :system t)
    (puthash "\\Beta" "Β" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Beta" "Β" nil :system t)
    (puthash "\\Gamma" "Γ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Gamma" "Γ" nil :system t)
    (puthash "\\Delta" "Δ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Delta" "Δ" nil :system t)
    (puthash "\\Epsilon" "Ε" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Epsilon" "Ε" nil :system t)
    (puthash "\\Zeta" "Ζ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Zeta" "Ζ" nil :system t)
    (puthash "\\Eta" "Η" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Eta" "Η" nil :system t)
    (puthash "\\Theta" "Θ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Theta" "Θ" nil :system t)
    (puthash "\\Iota" "Ι" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Iota" "Ι" nil :system t)
    (puthash "\\Kappa" "Κ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Kappa" "Κ" nil :system t)
    (puthash "\\Lambda" "Λ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Lambda" "Λ" nil :system t)
    (puthash "\\Mu" "Μ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Mu" "Μ" nil :system t)
    (puthash "\\Nu" "Ν" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Nu" "Ν" nil :system t)
    (puthash "\\Xi" "Ξ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Xi" "Ξ" nil :system t)
    (puthash "\\Omicron" "Ο" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Omicron" "Ο" nil :system t)
    (puthash "\\Pi" "Π" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Pi" "Π" nil :system t)
    (puthash "\\Rho" "Ρ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Rho" "Ρ" nil :system t)
    (puthash "\\Sigma" "Σ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Sigma" "Σ" nil :system t)
    (puthash "\\Tau" "Τ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Tau" "Τ" nil :system t)
    (puthash "\\Upsilon" "Υ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Upsilon" "Υ" nil :system t)
    (puthash "\\Phi" "Φ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Phi" "Φ" nil :system t)
    (puthash "\\Chi" "Χ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Chi" "Χ" nil :system t)
    (puthash "\\Psi" "Ψ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Psi" "Ψ" nil :system t)
    (puthash "\\Omega" "Ω" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Omega" "Ω" nil :system t)
    (puthash "\\alpha" "α" table)
    (define-abbrev julia-latexsub-abbrev-table "\\alpha" "α" nil :system t)
    (puthash "\\beta" "β" table)
    (define-abbrev julia-latexsub-abbrev-table "\\beta" "β" nil :system t)
    (puthash "\\gamma" "γ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gamma" "γ" nil :system t)
    (puthash "\\delta" "δ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\delta" "δ" nil :system t)
    (puthash "\\varepsilon" "ε" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varepsilon" "ε" nil :system t)
    (puthash "\\zeta" "ζ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\zeta" "ζ" nil :system t)
    (puthash "\\eta" "η" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eta" "η" nil :system t)
    (puthash "\\theta" "θ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\theta" "θ" nil :system t)
    (puthash "\\iota" "ι" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iota" "ι" nil :system t)
    (puthash "\\kappa" "κ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\kappa" "κ" nil :system t)
    (puthash "\\lambda" "λ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lambda" "λ" nil :system t)
    (puthash "\\mu" "μ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mu" "μ" nil :system t)
    (puthash "\\nu" "ν" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nu" "ν" nil :system t)
    (puthash "\\xi" "ξ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\xi" "ξ" nil :system t)
    (puthash "\\omicron" "ο" table)
    (define-abbrev julia-latexsub-abbrev-table "\\omicron" "ο" nil :system t)
    (puthash "\\pi" "π" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pi" "π" nil :system t)
    (puthash "\\rho" "ρ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rho" "ρ" nil :system t)
    (puthash "\\varsigma" "ς" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varsigma" "ς" nil :system t)
    (puthash "\\sigma" "σ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sigma" "σ" nil :system t)
    (puthash "\\tau" "τ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tau" "τ" nil :system t)
    (puthash "\\upsilon" "υ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upsilon" "υ" nil :system t)
    (puthash "\\varphi" "φ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varphi" "φ" nil :system t)
    (puthash "\\chi" "χ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\chi" "χ" nil :system t)
    (puthash "\\psi" "ψ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\psi" "ψ" nil :system t)
    (puthash "\\omega" "ω" table)
    (define-abbrev julia-latexsub-abbrev-table "\\omega" "ω" nil :system t)
    (puthash "\\varbeta" "ϐ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varbeta" "ϐ" nil :system t)
    (puthash "\\vartheta" "ϑ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vartheta" "ϑ" nil :system t)
    (puthash "\\phi" "ϕ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\phi" "ϕ" nil :system t)
    (puthash "\\varpi" "ϖ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varpi" "ϖ" nil :system t)
    (puthash "\\oldKoppa" "Ϙ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oldKoppa" "Ϙ" nil :system t)
    (puthash "\\oldkoppa" "ϙ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oldkoppa" "ϙ" nil :system t)
    (puthash "\\Stigma" "Ϛ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Stigma" "Ϛ" nil :system t)
    (puthash "\\stigma" "ϛ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\stigma" "ϛ" nil :system t)
    (puthash "\\Digamma" "Ϝ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Digamma" "Ϝ" nil :system t)
    (puthash "\\digamma" "ϝ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\digamma" "ϝ" nil :system t)
    (puthash "\\Koppa" "Ϟ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Koppa" "Ϟ" nil :system t)
    (puthash "\\koppa" "ϟ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\koppa" "ϟ" nil :system t)
    (puthash "\\Sampi" "Ϡ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Sampi" "Ϡ" nil :system t)
    (puthash "\\sampi" "ϡ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sampi" "ϡ" nil :system t)
    (puthash "\\varkappa" "ϰ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varkappa" "ϰ" nil :system t)
    (puthash "\\varrho" "ϱ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varrho" "ϱ" nil :system t)
    (puthash "\\varTheta" "ϴ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varTheta" "ϴ" nil :system t)
    (puthash "\\epsilon" "ϵ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\epsilon" "ϵ" nil :system t)
    (puthash "\\backepsilon" "϶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\backepsilon" "϶" nil :system t)
    (puthash "\\^A" "ᴬ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^A" "ᴬ" nil :system t)
    (puthash "\\^B" "ᴮ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^B" "ᴮ" nil :system t)
    (puthash "\\^D" "ᴰ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^D" "ᴰ" nil :system t)
    (puthash "\\^E" "ᴱ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^E" "ᴱ" nil :system t)
    (puthash "\\^G" "ᴳ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^G" "ᴳ" nil :system t)
    (puthash "\\^H" "ᴴ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^H" "ᴴ" nil :system t)
    (puthash "\\^I" "ᴵ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^I" "ᴵ" nil :system t)
    (puthash "\\^J" "ᴶ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^J" "ᴶ" nil :system t)
    (puthash "\\^K" "ᴷ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^K" "ᴷ" nil :system t)
    (puthash "\\^L" "ᴸ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^L" "ᴸ" nil :system t)
    (puthash "\\^M" "ᴹ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^M" "ᴹ" nil :system t)
    (puthash "\\^N" "ᴺ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^N" "ᴺ" nil :system t)
    (puthash "\\^O" "ᴼ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^O" "ᴼ" nil :system t)
    (puthash "\\^P" "ᴾ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^P" "ᴾ" nil :system t)
    (puthash "\\^R" "ᴿ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^R" "ᴿ" nil :system t)
    (puthash "\\^T" "ᵀ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^T" "ᵀ" nil :system t)
    (puthash "\\^U" "ᵁ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^U" "ᵁ" nil :system t)
    (puthash "\\^W" "ᵂ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^W" "ᵂ" nil :system t)
    (puthash "\\^a" "ᵃ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^a" "ᵃ" nil :system t)
    (puthash "\\^alpha" "ᵅ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^alpha" "ᵅ" nil :system t)
    (puthash "\\^b" "ᵇ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^b" "ᵇ" nil :system t)
    (puthash "\\^d" "ᵈ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^d" "ᵈ" nil :system t)
    (puthash "\\^e" "ᵉ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^e" "ᵉ" nil :system t)
    (puthash "\\^epsilon" "ᵋ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^epsilon" "ᵋ" nil :system t)
    (puthash "\\^g" "ᵍ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^g" "ᵍ" nil :system t)
    (puthash "\\^k" "ᵏ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^k" "ᵏ" nil :system t)
    (puthash "\\^m" "ᵐ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^m" "ᵐ" nil :system t)
    (puthash "\\^o" "ᵒ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^o" "ᵒ" nil :system t)
    (puthash "\\^p" "ᵖ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^p" "ᵖ" nil :system t)
    (puthash "\\^t" "ᵗ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^t" "ᵗ" nil :system t)
    (puthash "\\^u" "ᵘ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^u" "ᵘ" nil :system t)
    (puthash "\\^v" "ᵛ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^v" "ᵛ" nil :system t)
    (puthash "\\^beta" "ᵝ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^beta" "ᵝ" nil :system t)
    (puthash "\\^gamma" "ᵞ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^gamma" "ᵞ" nil :system t)
    (puthash "\\^delta" "ᵟ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^delta" "ᵟ" nil :system t)
    (puthash "\\^phi" "ᵠ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^phi" "ᵠ" nil :system t)
    (puthash "\\^chi" "ᵡ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^chi" "ᵡ" nil :system t)
    (puthash "\\_i" "ᵢ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_i" "ᵢ" nil :system t)
    (puthash "\\_r" "ᵣ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_r" "ᵣ" nil :system t)
    (puthash "\\_u" "ᵤ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_u" "ᵤ" nil :system t)
    (puthash "\\_v" "ᵥ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_v" "ᵥ" nil :system t)
    (puthash "\\_beta" "ᵦ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_beta" "ᵦ" nil :system t)
    (puthash "\\_gamma" "ᵧ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_gamma" "ᵧ" nil :system t)
    (puthash "\\_rho" "ᵨ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_rho" "ᵨ" nil :system t)
    (puthash "\\_phi" "ᵩ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_phi" "ᵩ" nil :system t)
    (puthash "\\_chi" "ᵪ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_chi" "ᵪ" nil :system t)
    (puthash "\\^c" "ᶜ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^c" "ᶜ" nil :system t)
    (puthash "\\^f" "ᶠ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^f" "ᶠ" nil :system t)
    (puthash "\\^iota" "ᶥ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^iota" "ᶥ" nil :system t)
    (puthash "\\^ltphi" "ᶲ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^ltphi" "ᶲ" nil :system t)
    (puthash "\\^z" "ᶻ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^z" "ᶻ" nil :system t)
    (puthash "\\^theta" "ᶿ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^theta" "ᶿ" nil :system t)
    (puthash "\\enspace" " " table)
    (define-abbrev julia-latexsub-abbrev-table "\\enspace" " " nil :system t)
    (puthash "\\quad" " " table)
    (define-abbrev julia-latexsub-abbrev-table "\\quad" " " nil :system t)
    (puthash "\\thickspace" " " table)
    (define-abbrev julia-latexsub-abbrev-table "\\thickspace" " " nil :system t)
    (puthash "\\thinspace" " " table)
    (define-abbrev julia-latexsub-abbrev-table "\\thinspace" " " nil :system t)
    (puthash "\\hspace" " " table)
    (define-abbrev julia-latexsub-abbrev-table "\\hspace" " " nil :system t)
    (puthash "\\endash" "–" table)
    (define-abbrev julia-latexsub-abbrev-table "\\endash" "–" nil :system t)
    (puthash "\\emdash" "—" table)
    (define-abbrev julia-latexsub-abbrev-table "\\emdash" "—" nil :system t)
    (puthash "\\Vert" "‖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Vert" "‖" nil :system t)
    (puthash "\\lq" "‘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lq" "‘" nil :system t)
    (puthash "\\rq" "’" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rq" "’" nil :system t)
    (puthash "\\reapos" "‛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\reapos" "‛" nil :system t)
    (puthash "\\ldq" "“" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ldq" "“" nil :system t)
    (puthash "\\rdq" "”" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rdq" "”" nil :system t)
    (puthash "\\dagger" "†" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dagger" "†" nil :system t)
    (puthash "\\ddagger" "‡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ddagger" "‡" nil :system t)
    (puthash "\\bullet" "•" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bullet" "•" nil :system t)
    (puthash "\\dots" "…" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dots" "…" nil :system t)
    (puthash "\\ldots" "…" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ldots" "…" nil :system t)
    (puthash "\\perthousand" "‰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\perthousand" "‰" nil :system t)
    (puthash "\\pertenthousand" "‱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pertenthousand" "‱" nil :system t)
    (puthash "\\prime" "′" table)
    (define-abbrev julia-latexsub-abbrev-table "\\prime" "′" nil :system t)
    (puthash "\\pprime" "″" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pprime" "″" nil :system t)
    (puthash "\\ppprime" "‴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ppprime" "‴" nil :system t)
    (puthash "\\backprime" "‵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\backprime" "‵" nil :system t)
    (puthash "\\backpprime" "‶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\backpprime" "‶" nil :system t)
    (puthash "\\backppprime" "‷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\backppprime" "‷" nil :system t)
    (puthash "\\guilsinglleft" "‹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\guilsinglleft" "‹" nil :system t)
    (puthash "\\guilsinglright" "›" table)
    (define-abbrev julia-latexsub-abbrev-table "\\guilsinglright" "›" nil :system t)
    (puthash "\\:bangbang:" "‼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bangbang:" "‼" nil :system t)
    (puthash "\\tieconcat" "⁀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tieconcat" "⁀" nil :system t)
    (puthash "\\:interrobang:" "⁉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:interrobang:" "⁉" nil :system t)
    (puthash "\\pppprime" "⁗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pppprime" "⁗" nil :system t)
    (puthash "\\tricolon" "⁝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tricolon" "⁝" nil :system t)
    (puthash "\\nolinebreak" "\u2060" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nolinebreak" "\u2060" nil :system t)
    (puthash "\\^0" "⁰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^0" "⁰" nil :system t)
    (puthash "\\^i" "ⁱ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^i" "ⁱ" nil :system t)
    (puthash "\\^4" "⁴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^4" "⁴" nil :system t)
    (puthash "\\^5" "⁵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^5" "⁵" nil :system t)
    (puthash "\\^6" "⁶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^6" "⁶" nil :system t)
    (puthash "\\^7" "⁷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^7" "⁷" nil :system t)
    (puthash "\\^8" "⁸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^8" "⁸" nil :system t)
    (puthash "\\^9" "⁹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^9" "⁹" nil :system t)
    (puthash "\\^+" "⁺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^+" "⁺" nil :system t)
    (puthash "\\^-" "⁻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^-" "⁻" nil :system t)
    (puthash "\\^=" "⁼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^=" "⁼" nil :system t)
    (puthash "\\^(" "⁽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^(" "⁽" nil :system t)
    (puthash "\\^)" "⁾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^)" "⁾" nil :system t)
    (puthash "\\^n" "ⁿ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^n" "ⁿ" nil :system t)
    (puthash "\\_0" "₀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_0" "₀" nil :system t)
    (puthash "\\_1" "₁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_1" "₁" nil :system t)
    (puthash "\\_2" "₂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_2" "₂" nil :system t)
    (puthash "\\_3" "₃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_3" "₃" nil :system t)
    (puthash "\\_4" "₄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_4" "₄" nil :system t)
    (puthash "\\_5" "₅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_5" "₅" nil :system t)
    (puthash "\\_6" "₆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_6" "₆" nil :system t)
    (puthash "\\_7" "₇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_7" "₇" nil :system t)
    (puthash "\\_8" "₈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_8" "₈" nil :system t)
    (puthash "\\_9" "₉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_9" "₉" nil :system t)
    (puthash "\\_+" "₊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_+" "₊" nil :system t)
    (puthash "\\_-" "₋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_-" "₋" nil :system t)
    (puthash "\\_=" "₌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_=" "₌" nil :system t)
    (puthash "\\_(" "₍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_(" "₍" nil :system t)
    (puthash "\\_)" "₎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_)" "₎" nil :system t)
    (puthash "\\_a" "ₐ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_a" "ₐ" nil :system t)
    (puthash "\\_e" "ₑ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_e" "ₑ" nil :system t)
    (puthash "\\_o" "ₒ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_o" "ₒ" nil :system t)
    (puthash "\\_x" "ₓ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_x" "ₓ" nil :system t)
    (puthash "\\_schwa" "ₔ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_schwa" "ₔ" nil :system t)
    (puthash "\\_h" "ₕ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_h" "ₕ" nil :system t)
    (puthash "\\_k" "ₖ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_k" "ₖ" nil :system t)
    (puthash "\\_l" "ₗ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_l" "ₗ" nil :system t)
    (puthash "\\_m" "ₘ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_m" "ₘ" nil :system t)
    (puthash "\\_n" "ₙ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_n" "ₙ" nil :system t)
    (puthash "\\_p" "ₚ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_p" "ₚ" nil :system t)
    (puthash "\\_s" "ₛ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_s" "ₛ" nil :system t)
    (puthash "\\_t" "ₜ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_t" "ₜ" nil :system t)
    (puthash "\\pes" "₧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pes" "₧" nil :system t)
    (puthash "\\euro" "€" table)
    (define-abbrev julia-latexsub-abbrev-table "\\euro" "€" nil :system t)
    (puthash "\\leftharpoonaccent" "⃐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftharpoonaccent" "⃐" nil :system t)
    (puthash "\\rightharpoonaccent" "⃑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightharpoonaccent" "⃑" nil :system t)
    (puthash "\\vertoverlay" "⃒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vertoverlay" "⃒" nil :system t)
    (puthash "\\overleftarrow" "⃖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\overleftarrow" "⃖" nil :system t)
    (puthash "\\vec" "⃗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vec" "⃗" nil :system t)
    (puthash "\\dddot" "⃛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dddot" "⃛" nil :system t)
    (puthash "\\ddddot" "⃜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ddddot" "⃜" nil :system t)
    (puthash "\\enclosecircle" "⃝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\enclosecircle" "⃝" nil :system t)
    (puthash "\\enclosesquare" "⃞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\enclosesquare" "⃞" nil :system t)
    (puthash "\\enclosediamond" "⃟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\enclosediamond" "⃟" nil :system t)
    (puthash "\\overleftrightarrow" "⃡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\overleftrightarrow" "⃡" nil :system t)
    (puthash "\\enclosetriangle" "⃤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\enclosetriangle" "⃤" nil :system t)
    (puthash "\\annuity" "⃧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\annuity" "⃧" nil :system t)
    (puthash "\\threeunderdot" "⃨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\threeunderdot" "⃨" nil :system t)
    (puthash "\\widebridgeabove" "⃩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\widebridgeabove" "⃩" nil :system t)
    (puthash "\\underrightharpoondown" "⃬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underrightharpoondown" "⃬" nil :system t)
    (puthash "\\underleftharpoondown" "⃭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underleftharpoondown" "⃭" nil :system t)
    (puthash "\\underleftarrow" "⃮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underleftarrow" "⃮" nil :system t)
    (puthash "\\underrightarrow" "⃯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underrightarrow" "⃯" nil :system t)
    (puthash "\\asteraccent" "⃰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\asteraccent" "⃰" nil :system t)
    (puthash "\\bbC" "ℂ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbC" "ℂ" nil :system t)
    (puthash "\\eulermascheroni" "ℇ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eulermascheroni" "ℇ" nil :system t)
    (puthash "\\scrg" "ℊ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrg" "ℊ" nil :system t)
    (puthash "\\scrH" "ℋ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrH" "ℋ" nil :system t)
    (puthash "\\frakH" "ℌ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakH" "ℌ" nil :system t)
    (puthash "\\bbH" "ℍ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbH" "ℍ" nil :system t)
    (puthash "\\planck" "ℎ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\planck" "ℎ" nil :system t)
    (puthash "\\ith" "ℎ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ith" "ℎ" nil :system t)
    (puthash "\\hslash" "ℏ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hslash" "ℏ" nil :system t)
    (puthash "\\scrI" "ℐ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrI" "ℐ" nil :system t)
    (puthash "\\frakI" "ℑ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakI" "ℑ" nil :system t)
    (puthash "\\Im" "ℑ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Im" "ℑ" nil :system t)
    (puthash "\\scrL" "ℒ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrL" "ℒ" nil :system t)
    (puthash "\\ell" "ℓ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ell" "ℓ" nil :system t)
    (puthash "\\bbN" "ℕ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbN" "ℕ" nil :system t)
    (puthash "\\numero" "№" table)
    (define-abbrev julia-latexsub-abbrev-table "\\numero" "№" nil :system t)
    (puthash "\\wp" "℘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wp" "℘" nil :system t)
    (puthash "\\bbP" "ℙ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbP" "ℙ" nil :system t)
    (puthash "\\bbQ" "ℚ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbQ" "ℚ" nil :system t)
    (puthash "\\scrR" "ℛ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrR" "ℛ" nil :system t)
    (puthash "\\Re" "ℜ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Re" "ℜ" nil :system t)
    (puthash "\\frakR" "ℜ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakR" "ℜ" nil :system t)
    (puthash "\\bbR" "ℝ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbR" "ℝ" nil :system t)
    (puthash "\\xrat" "℞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\xrat" "℞" nil :system t)
    (puthash "\\trademark" "™" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trademark" "™" nil :system t)
    (puthash "\\:tm:" "™" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tm:" "™" nil :system t)
    (puthash "\\bbZ" "ℤ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbZ" "ℤ" nil :system t)
    (puthash "\\ohm" "Ω" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ohm" "Ω" nil :system t)
    (puthash "\\mho" "℧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mho" "℧" nil :system t)
    (puthash "\\frakZ" "ℨ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakZ" "ℨ" nil :system t)
    (puthash "\\turnediota" "℩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\turnediota" "℩" nil :system t)
    (puthash "\\Angstrom" "Å" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Angstrom" "Å" nil :system t)
    (puthash "\\scrB" "ℬ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrB" "ℬ" nil :system t)
    (puthash "\\frakC" "ℭ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakC" "ℭ" nil :system t)
    (puthash "\\scre" "ℯ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scre" "ℯ" nil :system t)
    (puthash "\\euler" "ℯ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\euler" "ℯ" nil :system t)
    (puthash "\\scrE" "ℰ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrE" "ℰ" nil :system t)
    (puthash "\\scrF" "ℱ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrF" "ℱ" nil :system t)
    (puthash "\\Finv" "Ⅎ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Finv" "Ⅎ" nil :system t)
    (puthash "\\scrM" "ℳ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrM" "ℳ" nil :system t)
    (puthash "\\scro" "ℴ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scro" "ℴ" nil :system t)
    (puthash "\\aleph" "ℵ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\aleph" "ℵ" nil :system t)
    (puthash "\\beth" "ℶ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\beth" "ℶ" nil :system t)
    (puthash "\\gimel" "ℷ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gimel" "ℷ" nil :system t)
    (puthash "\\daleth" "ℸ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\daleth" "ℸ" nil :system t)
    (puthash "\\:information_source:" "ℹ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:information_source:" "ℹ" nil :system t)
    (puthash "\\bbpi" "ℼ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbpi" "ℼ" nil :system t)
    (puthash "\\bbgamma" "ℽ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbgamma" "ℽ" nil :system t)
    (puthash "\\bbGamma" "ℾ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbGamma" "ℾ" nil :system t)
    (puthash "\\bbPi" "ℿ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbPi" "ℿ" nil :system t)
    (puthash "\\bbsum" "⅀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbsum" "⅀" nil :system t)
    (puthash "\\Game" "⅁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Game" "⅁" nil :system t)
    (puthash "\\sansLturned" "⅂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansLturned" "⅂" nil :system t)
    (puthash "\\sansLmirrored" "⅃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansLmirrored" "⅃" nil :system t)
    (puthash "\\Yup" "⅄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Yup" "⅄" nil :system t)
    (puthash "\\bbiD" "ⅅ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbiD" "ⅅ" nil :system t)
    (puthash "\\bbid" "ⅆ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbid" "ⅆ" nil :system t)
    (puthash "\\bbie" "ⅇ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbie" "ⅇ" nil :system t)
    (puthash "\\bbii" "ⅈ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbii" "ⅈ" nil :system t)
    (puthash "\\bbij" "ⅉ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbij" "ⅉ" nil :system t)
    (puthash "\\PropertyLine" "⅊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\PropertyLine" "⅊" nil :system t)
    (puthash "\\upand" "⅋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upand" "⅋" nil :system t)
    (puthash "\\1/7" "⅐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/7" "⅐" nil :system t)
    (puthash "\\1/9" "⅑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/9" "⅑" nil :system t)
    (puthash "\\1/10" "⅒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/10" "⅒" nil :system t)
    (puthash "\\1/3" "⅓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/3" "⅓" nil :system t)
    (puthash "\\2/3" "⅔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\2/3" "⅔" nil :system t)
    (puthash "\\1/5" "⅕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/5" "⅕" nil :system t)
    (puthash "\\2/5" "⅖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\2/5" "⅖" nil :system t)
    (puthash "\\3/5" "⅗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\3/5" "⅗" nil :system t)
    (puthash "\\4/5" "⅘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\4/5" "⅘" nil :system t)
    (puthash "\\1/6" "⅙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/6" "⅙" nil :system t)
    (puthash "\\5/6" "⅚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\5/6" "⅚" nil :system t)
    (puthash "\\1/8" "⅛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/8" "⅛" nil :system t)
    (puthash "\\3/8" "⅜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\3/8" "⅜" nil :system t)
    (puthash "\\5/8" "⅝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\5/8" "⅝" nil :system t)
    (puthash "\\7/8" "⅞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\7/8" "⅞" nil :system t)
    (puthash "\\1/" "⅟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\1/" "⅟" nil :system t)
    (puthash "\\0/3" "↉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\0/3" "↉" nil :system t)
    (puthash "\\leftarrow" "←" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrow" "←" nil :system t)
    (puthash "\\uparrow" "↑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\uparrow" "↑" nil :system t)
    (puthash "\\rightarrow" "→" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrow" "→" nil :system t)
    (puthash "\\to" "→" table)
    (define-abbrev julia-latexsub-abbrev-table "\\to" "→" nil :system t)
    (puthash "\\downarrow" "↓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downarrow" "↓" nil :system t)
    (puthash "\\:left_right_arrow:" "↔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:left_right_arrow:" "↔" nil :system t)
    (puthash "\\leftrightarrow" "↔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightarrow" "↔" nil :system t)
    (puthash "\\:arrow_up_down:" "↕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_up_down:" "↕" nil :system t)
    (puthash "\\updownarrow" "↕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\updownarrow" "↕" nil :system t)
    (puthash "\\:arrow_upper_left:" "↖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_upper_left:" "↖" nil :system t)
    (puthash "\\nwarrow" "↖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nwarrow" "↖" nil :system t)
    (puthash "\\:arrow_upper_right:" "↗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_upper_right:" "↗" nil :system t)
    (puthash "\\nearrow" "↗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nearrow" "↗" nil :system t)
    (puthash "\\:arrow_lower_right:" "↘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_lower_right:" "↘" nil :system t)
    (puthash "\\searrow" "↘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\searrow" "↘" nil :system t)
    (puthash "\\swarrow" "↙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\swarrow" "↙" nil :system t)
    (puthash "\\:arrow_lower_left:" "↙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_lower_left:" "↙" nil :system t)
    (puthash "\\nleftarrow" "↚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nleftarrow" "↚" nil :system t)
    (puthash "\\nrightarrow" "↛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nrightarrow" "↛" nil :system t)
    (puthash "\\leftwavearrow" "↜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftwavearrow" "↜" nil :system t)
    (puthash "\\rightwavearrow" "↝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightwavearrow" "↝" nil :system t)
    (puthash "\\twoheadleftarrow" "↞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadleftarrow" "↞" nil :system t)
    (puthash "\\twoheaduparrow" "↟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheaduparrow" "↟" nil :system t)
    (puthash "\\twoheadrightarrow" "↠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadrightarrow" "↠" nil :system t)
    (puthash "\\twoheaddownarrow" "↡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheaddownarrow" "↡" nil :system t)
    (puthash "\\leftarrowtail" "↢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowtail" "↢" nil :system t)
    (puthash "\\rightarrowtail" "↣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowtail" "↣" nil :system t)
    (puthash "\\mapsfrom" "↤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mapsfrom" "↤" nil :system t)
    (puthash "\\mapsup" "↥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mapsup" "↥" nil :system t)
    (puthash "\\mapsto" "↦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mapsto" "↦" nil :system t)
    (puthash "\\mapsdown" "↧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mapsdown" "↧" nil :system t)
    (puthash "\\updownarrowbar" "↨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\updownarrowbar" "↨" nil :system t)
    (puthash "\\hookleftarrow" "↩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hookleftarrow" "↩" nil :system t)
    (puthash "\\:leftwards_arrow_with_hook:" "↩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:leftwards_arrow_with_hook:" "↩" nil :system t)
    (puthash "\\:arrow_right_hook:" "↪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_right_hook:" "↪" nil :system t)
    (puthash "\\hookrightarrow" "↪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hookrightarrow" "↪" nil :system t)
    (puthash "\\looparrowleft" "↫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\looparrowleft" "↫" nil :system t)
    (puthash "\\looparrowright" "↬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\looparrowright" "↬" nil :system t)
    (puthash "\\leftrightsquigarrow" "↭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightsquigarrow" "↭" nil :system t)
    (puthash "\\nleftrightarrow" "↮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nleftrightarrow" "↮" nil :system t)
    (puthash "\\downzigzagarrow" "↯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downzigzagarrow" "↯" nil :system t)
    (puthash "\\Lsh" "↰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Lsh" "↰" nil :system t)
    (puthash "\\Rsh" "↱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Rsh" "↱" nil :system t)
    (puthash "\\Ldsh" "↲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Ldsh" "↲" nil :system t)
    (puthash "\\Rdsh" "↳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Rdsh" "↳" nil :system t)
    (puthash "\\linefeed" "↴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\linefeed" "↴" nil :system t)
    (puthash "\\carriagereturn" "↵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\carriagereturn" "↵" nil :system t)
    (puthash "\\curvearrowleft" "↶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\curvearrowleft" "↶" nil :system t)
    (puthash "\\curvearrowright" "↷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\curvearrowright" "↷" nil :system t)
    (puthash "\\barovernorthwestarrow" "↸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barovernorthwestarrow" "↸" nil :system t)
    (puthash "\\barleftarrowrightarrowbar" "↹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barleftarrowrightarrowbar" "↹" nil :system t)
    (puthash "\\circlearrowleft" "↺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circlearrowleft" "↺" nil :system t)
    (puthash "\\circlearrowright" "↻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circlearrowright" "↻" nil :system t)
    (puthash "\\leftharpoonup" "↼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftharpoonup" "↼" nil :system t)
    (puthash "\\leftharpoondown" "↽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftharpoondown" "↽" nil :system t)
    (puthash "\\upharpoonright" "↾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upharpoonright" "↾" nil :system t)
    (puthash "\\upharpoonleft" "↿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upharpoonleft" "↿" nil :system t)
    (puthash "\\rightharpoonup" "⇀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightharpoonup" "⇀" nil :system t)
    (puthash "\\rightharpoondown" "⇁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightharpoondown" "⇁" nil :system t)
    (puthash "\\downharpoonright" "⇂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downharpoonright" "⇂" nil :system t)
    (puthash "\\downharpoonleft" "⇃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downharpoonleft" "⇃" nil :system t)
    (puthash "\\rightleftarrows" "⇄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightleftarrows" "⇄" nil :system t)
    (puthash "\\dblarrowupdown" "⇅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dblarrowupdown" "⇅" nil :system t)
    (puthash "\\leftrightarrows" "⇆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightarrows" "⇆" nil :system t)
    (puthash "\\leftleftarrows" "⇇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftleftarrows" "⇇" nil :system t)
    (puthash "\\upuparrows" "⇈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upuparrows" "⇈" nil :system t)
    (puthash "\\rightrightarrows" "⇉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightrightarrows" "⇉" nil :system t)
    (puthash "\\downdownarrows" "⇊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downdownarrows" "⇊" nil :system t)
    (puthash "\\leftrightharpoons" "⇋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightharpoons" "⇋" nil :system t)
    (puthash "\\rightleftharpoons" "⇌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightleftharpoons" "⇌" nil :system t)
    (puthash "\\nLeftarrow" "⇍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nLeftarrow" "⇍" nil :system t)
    (puthash "\\nLeftrightarrow" "⇎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nLeftrightarrow" "⇎" nil :system t)
    (puthash "\\nRightarrow" "⇏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nRightarrow" "⇏" nil :system t)
    (puthash "\\Leftarrow" "⇐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Leftarrow" "⇐" nil :system t)
    (puthash "\\Uparrow" "⇑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Uparrow" "⇑" nil :system t)
    (puthash "\\Rightarrow" "⇒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Rightarrow" "⇒" nil :system t)
    (puthash "\\Downarrow" "⇓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Downarrow" "⇓" nil :system t)
    (puthash "\\Leftrightarrow" "⇔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Leftrightarrow" "⇔" nil :system t)
    (puthash "\\Updownarrow" "⇕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Updownarrow" "⇕" nil :system t)
    (puthash "\\Nwarrow" "⇖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Nwarrow" "⇖" nil :system t)
    (puthash "\\Nearrow" "⇗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Nearrow" "⇗" nil :system t)
    (puthash "\\Searrow" "⇘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Searrow" "⇘" nil :system t)
    (puthash "\\Swarrow" "⇙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Swarrow" "⇙" nil :system t)
    (puthash "\\Lleftarrow" "⇚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Lleftarrow" "⇚" nil :system t)
    (puthash "\\Rrightarrow" "⇛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Rrightarrow" "⇛" nil :system t)
    (puthash "\\leftsquigarrow" "⇜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftsquigarrow" "⇜" nil :system t)
    (puthash "\\rightsquigarrow" "⇝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightsquigarrow" "⇝" nil :system t)
    (puthash "\\nHuparrow" "⇞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nHuparrow" "⇞" nil :system t)
    (puthash "\\nHdownarrow" "⇟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nHdownarrow" "⇟" nil :system t)
    (puthash "\\leftdasharrow" "⇠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftdasharrow" "⇠" nil :system t)
    (puthash "\\updasharrow" "⇡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\updasharrow" "⇡" nil :system t)
    (puthash "\\rightdasharrow" "⇢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightdasharrow" "⇢" nil :system t)
    (puthash "\\downdasharrow" "⇣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downdasharrow" "⇣" nil :system t)
    (puthash "\\barleftarrow" "⇤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barleftarrow" "⇤" nil :system t)
    (puthash "\\rightarrowbar" "⇥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowbar" "⇥" nil :system t)
    (puthash "\\leftwhitearrow" "⇦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftwhitearrow" "⇦" nil :system t)
    (puthash "\\upwhitearrow" "⇧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upwhitearrow" "⇧" nil :system t)
    (puthash "\\rightwhitearrow" "⇨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightwhitearrow" "⇨" nil :system t)
    (puthash "\\downwhitearrow" "⇩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downwhitearrow" "⇩" nil :system t)
    (puthash "\\whitearrowupfrombar" "⇪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\whitearrowupfrombar" "⇪" nil :system t)
    (puthash "\\circleonrightarrow" "⇴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circleonrightarrow" "⇴" nil :system t)
    (puthash "\\DownArrowUpArrow" "⇵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownArrowUpArrow" "⇵" nil :system t)
    (puthash "\\rightthreearrows" "⇶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightthreearrows" "⇶" nil :system t)
    (puthash "\\nvleftarrow" "⇷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvleftarrow" "⇷" nil :system t)
    (puthash "\\nvrightarrow" "⇸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvrightarrow" "⇸" nil :system t)
    (puthash "\\nvleftrightarrow" "⇹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvleftrightarrow" "⇹" nil :system t)
    (puthash "\\nVleftarrow" "⇺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVleftarrow" "⇺" nil :system t)
    (puthash "\\nVrightarrow" "⇻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVrightarrow" "⇻" nil :system t)
    (puthash "\\nVleftrightarrow" "⇼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVleftrightarrow" "⇼" nil :system t)
    (puthash "\\leftarrowtriangle" "⇽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowtriangle" "⇽" nil :system t)
    (puthash "\\rightarrowtriangle" "⇾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowtriangle" "⇾" nil :system t)
    (puthash "\\leftrightarrowtriangle" "⇿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightarrowtriangle" "⇿" nil :system t)
    (puthash "\\forall" "∀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\forall" "∀" nil :system t)
    (puthash "\\complement" "∁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\complement" "∁" nil :system t)
    (puthash "\\partial" "∂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\partial" "∂" nil :system t)
    (puthash "\\exists" "∃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\exists" "∃" nil :system t)
    (puthash "\\nexists" "∄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nexists" "∄" nil :system t)
    (puthash "\\varnothing" "∅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varnothing" "∅" nil :system t)
    (puthash "\\emptyset" "∅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\emptyset" "∅" nil :system t)
    (puthash "\\increment" "∆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\increment" "∆" nil :system t)
    (puthash "\\del" "∇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\del" "∇" nil :system t)
    (puthash "\\nabla" "∇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nabla" "∇" nil :system t)
    (puthash "\\in" "∈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\in" "∈" nil :system t)
    (puthash "\\notin" "∉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\notin" "∉" nil :system t)
    (puthash "\\smallin" "∊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smallin" "∊" nil :system t)
    (puthash "\\ni" "∋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ni" "∋" nil :system t)
    (puthash "\\nni" "∌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nni" "∌" nil :system t)
    (puthash "\\smallni" "∍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smallni" "∍" nil :system t)
    (puthash "\\QED" "∎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\QED" "∎" nil :system t)
    (puthash "\\prod" "∏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\prod" "∏" nil :system t)
    (puthash "\\coprod" "∐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\coprod" "∐" nil :system t)
    (puthash "\\sum" "∑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sum" "∑" nil :system t)
    (puthash "\\minus" "−" table)
    (define-abbrev julia-latexsub-abbrev-table "\\minus" "−" nil :system t)
    (puthash "\\mp" "∓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mp" "∓" nil :system t)
    (puthash "\\dotplus" "∔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dotplus" "∔" nil :system t)
    (puthash "\\setminus" "∖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\setminus" "∖" nil :system t)
    (puthash "\\ast" "∗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ast" "∗" nil :system t)
    (puthash "\\circ" "∘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circ" "∘" nil :system t)
    (puthash "\\vysmblkcircle" "∙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vysmblkcircle" "∙" nil :system t)
    (puthash "\\surd" "√" table)
    (define-abbrev julia-latexsub-abbrev-table "\\surd" "√" nil :system t)
    (puthash "\\sqrt" "√" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqrt" "√" nil :system t)
    (puthash "\\cbrt" "∛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cbrt" "∛" nil :system t)
    (puthash "\\fourthroot" "∜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fourthroot" "∜" nil :system t)
    (puthash "\\propto" "∝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\propto" "∝" nil :system t)
    (puthash "\\infty" "∞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\infty" "∞" nil :system t)
    (puthash "\\rightangle" "∟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightangle" "∟" nil :system t)
    (puthash "\\angle" "∠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\angle" "∠" nil :system t)
    (puthash "\\measuredangle" "∡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measuredangle" "∡" nil :system t)
    (puthash "\\sphericalangle" "∢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sphericalangle" "∢" nil :system t)
    (puthash "\\mid" "∣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mid" "∣" nil :system t)
    (puthash "\\nmid" "∤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nmid" "∤" nil :system t)
    (puthash "\\parallel" "∥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\parallel" "∥" nil :system t)
    (puthash "\\nparallel" "∦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nparallel" "∦" nil :system t)
    (puthash "\\wedge" "∧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedge" "∧" nil :system t)
    (puthash "\\vee" "∨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vee" "∨" nil :system t)
    (puthash "\\cap" "∩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cap" "∩" nil :system t)
    (puthash "\\cup" "∪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cup" "∪" nil :system t)
    (puthash "\\int" "∫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\int" "∫" nil :system t)
    (puthash "\\iint" "∬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iint" "∬" nil :system t)
    (puthash "\\iiint" "∭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iiint" "∭" nil :system t)
    (puthash "\\oint" "∮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oint" "∮" nil :system t)
    (puthash "\\oiint" "∯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oiint" "∯" nil :system t)
    (puthash "\\oiiint" "∰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oiiint" "∰" nil :system t)
    (puthash "\\clwintegral" "∱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\clwintegral" "∱" nil :system t)
    (puthash "\\varointclockwise" "∲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varointclockwise" "∲" nil :system t)
    (puthash "\\ointctrclockwise" "∳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ointctrclockwise" "∳" nil :system t)
    (puthash "\\therefore" "∴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\therefore" "∴" nil :system t)
    (puthash "\\because" "∵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\because" "∵" nil :system t)
    (puthash "\\Colon" "∷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Colon" "∷" nil :system t)
    (puthash "\\dotminus" "∸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dotminus" "∸" nil :system t)
    (puthash "\\dotsminusdots" "∺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dotsminusdots" "∺" nil :system t)
    (puthash "\\kernelcontraction" "∻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\kernelcontraction" "∻" nil :system t)
    (puthash "\\sim" "∼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sim" "∼" nil :system t)
    (puthash "\\backsim" "∽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\backsim" "∽" nil :system t)
    (puthash "\\lazysinv" "∾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lazysinv" "∾" nil :system t)
    (puthash "\\sinewave" "∿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sinewave" "∿" nil :system t)
    (puthash "\\wr" "≀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wr" "≀" nil :system t)
    (puthash "\\nsim" "≁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsim" "≁" nil :system t)
    (puthash "\\eqsim" "≂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqsim" "≂" nil :system t)
    (puthash "\\neqsim" "≂̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neqsim" "≂̸" nil :system t)
    (puthash "\\simeq" "≃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simeq" "≃" nil :system t)
    (puthash "\\nsime" "≄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsime" "≄" nil :system t)
    (puthash "\\cong" "≅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cong" "≅" nil :system t)
    (puthash "\\approxnotequal" "≆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\approxnotequal" "≆" nil :system t)
    (puthash "\\ncong" "≇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ncong" "≇" nil :system t)
    (puthash "\\approx" "≈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\approx" "≈" nil :system t)
    (puthash "\\napprox" "≉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\napprox" "≉" nil :system t)
    (puthash "\\approxeq" "≊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\approxeq" "≊" nil :system t)
    (puthash "\\tildetrpl" "≋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tildetrpl" "≋" nil :system t)
    (puthash "\\allequal" "≌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\allequal" "≌" nil :system t)
    (puthash "\\asymp" "≍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\asymp" "≍" nil :system t)
    (puthash "\\Bumpeq" "≎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Bumpeq" "≎" nil :system t)
    (puthash "\\nBumpeq" "≎̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nBumpeq" "≎̸" nil :system t)
    (puthash "\\bumpeq" "≏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bumpeq" "≏" nil :system t)
    (puthash "\\nbumpeq" "≏̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nbumpeq" "≏̸" nil :system t)
    (puthash "\\doteq" "≐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\doteq" "≐" nil :system t)
    (puthash "\\Doteq" "≑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Doteq" "≑" nil :system t)
    (puthash "\\fallingdotseq" "≒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fallingdotseq" "≒" nil :system t)
    (puthash "\\risingdotseq" "≓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\risingdotseq" "≓" nil :system t)
    (puthash "\\coloneq" "≔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\coloneq" "≔" nil :system t)
    (puthash "\\eqcolon" "≕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqcolon" "≕" nil :system t)
    (puthash "\\eqcirc" "≖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqcirc" "≖" nil :system t)
    (puthash "\\circeq" "≗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circeq" "≗" nil :system t)
    (puthash "\\arceq" "≘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\arceq" "≘" nil :system t)
    (puthash "\\wedgeq" "≙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedgeq" "≙" nil :system t)
    (puthash "\\veeeq" "≚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\veeeq" "≚" nil :system t)
    (puthash "\\starequal" "≛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\starequal" "≛" nil :system t)
    (puthash "\\triangleq" "≜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangleq" "≜" nil :system t)
    (puthash "\\eqdef" "≝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqdef" "≝" nil :system t)
    (puthash "\\measeq" "≞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measeq" "≞" nil :system t)
    (puthash "\\questeq" "≟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\questeq" "≟" nil :system t)
    (puthash "\\ne" "≠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ne" "≠" nil :system t)
    (puthash "\\neq" "≠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neq" "≠" nil :system t)
    (puthash "\\equiv" "≡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\equiv" "≡" nil :system t)
    (puthash "\\nequiv" "≢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nequiv" "≢" nil :system t)
    (puthash "\\Equiv" "≣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Equiv" "≣" nil :system t)
    (puthash "\\le" "≤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\le" "≤" nil :system t)
    (puthash "\\leq" "≤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leq" "≤" nil :system t)
    (puthash "\\ge" "≥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ge" "≥" nil :system t)
    (puthash "\\geq" "≥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\geq" "≥" nil :system t)
    (puthash "\\leqq" "≦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leqq" "≦" nil :system t)
    (puthash "\\geqq" "≧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\geqq" "≧" nil :system t)
    (puthash "\\lneqq" "≨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lneqq" "≨" nil :system t)
    (puthash "\\lvertneqq" "≨︀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lvertneqq" "≨︀" nil :system t)
    (puthash "\\gneqq" "≩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gneqq" "≩" nil :system t)
    (puthash "\\gvertneqq" "≩︀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gvertneqq" "≩︀" nil :system t)
    (puthash "\\ll" "≪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ll" "≪" nil :system t)
    (puthash "\\NotLessLess" "≪̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotLessLess" "≪̸" nil :system t)
    (puthash "\\gg" "≫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gg" "≫" nil :system t)
    (puthash "\\NotGreaterGreater" "≫̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotGreaterGreater" "≫̸" nil :system t)
    (puthash "\\between" "≬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\between" "≬" nil :system t)
    (puthash "\\nasymp" "≭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nasymp" "≭" nil :system t)
    (puthash "\\nless" "≮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nless" "≮" nil :system t)
    (puthash "\\ngtr" "≯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ngtr" "≯" nil :system t)
    (puthash "\\nleq" "≰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nleq" "≰" nil :system t)
    (puthash "\\ngeq" "≱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ngeq" "≱" nil :system t)
    (puthash "\\lesssim" "≲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesssim" "≲" nil :system t)
    (puthash "\\gtrsim" "≳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtrsim" "≳" nil :system t)
    (puthash "\\nlesssim" "≴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nlesssim" "≴" nil :system t)
    (puthash "\\ngtrsim" "≵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ngtrsim" "≵" nil :system t)
    (puthash "\\lessgtr" "≶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lessgtr" "≶" nil :system t)
    (puthash "\\gtrless" "≷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtrless" "≷" nil :system t)
    (puthash "\\notlessgreater" "≸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\notlessgreater" "≸" nil :system t)
    (puthash "\\notgreaterless" "≹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\notgreaterless" "≹" nil :system t)
    (puthash "\\prec" "≺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\prec" "≺" nil :system t)
    (puthash "\\succ" "≻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succ" "≻" nil :system t)
    (puthash "\\preccurlyeq" "≼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\preccurlyeq" "≼" nil :system t)
    (puthash "\\succcurlyeq" "≽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succcurlyeq" "≽" nil :system t)
    (puthash "\\precsim" "≾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\precsim" "≾" nil :system t)
    (puthash "\\nprecsim" "≾̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nprecsim" "≾̸" nil :system t)
    (puthash "\\succsim" "≿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succsim" "≿" nil :system t)
    (puthash "\\nsuccsim" "≿̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsuccsim" "≿̸" nil :system t)
    (puthash "\\nprec" "⊀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nprec" "⊀" nil :system t)
    (puthash "\\nsucc" "⊁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsucc" "⊁" nil :system t)
    (puthash "\\subset" "⊂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subset" "⊂" nil :system t)
    (puthash "\\supset" "⊃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supset" "⊃" nil :system t)
    (puthash "\\nsubset" "⊄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsubset" "⊄" nil :system t)
    (puthash "\\nsupset" "⊅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsupset" "⊅" nil :system t)
    (puthash "\\subseteq" "⊆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subseteq" "⊆" nil :system t)
    (puthash "\\supseteq" "⊇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supseteq" "⊇" nil :system t)
    (puthash "\\nsubseteq" "⊈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsubseteq" "⊈" nil :system t)
    (puthash "\\nsupseteq" "⊉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsupseteq" "⊉" nil :system t)
    (puthash "\\subsetneq" "⊊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsetneq" "⊊" nil :system t)
    (puthash "\\varsubsetneqq" "⊊︀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varsubsetneqq" "⊊︀" nil :system t)
    (puthash "\\supsetneq" "⊋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsetneq" "⊋" nil :system t)
    (puthash "\\varsupsetneq" "⊋︀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varsupsetneq" "⊋︀" nil :system t)
    (puthash "\\cupdot" "⊍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cupdot" "⊍" nil :system t)
    (puthash "\\uplus" "⊎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\uplus" "⊎" nil :system t)
    (puthash "\\sqsubset" "⊏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqsubset" "⊏" nil :system t)
    (puthash "\\NotSquareSubset" "⊏̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotSquareSubset" "⊏̸" nil :system t)
    (puthash "\\sqsupset" "⊐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqsupset" "⊐" nil :system t)
    (puthash "\\NotSquareSuperset" "⊐̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotSquareSuperset" "⊐̸" nil :system t)
    (puthash "\\sqsubseteq" "⊑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqsubseteq" "⊑" nil :system t)
    (puthash "\\sqsupseteq" "⊒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqsupseteq" "⊒" nil :system t)
    (puthash "\\sqcap" "⊓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqcap" "⊓" nil :system t)
    (puthash "\\sqcup" "⊔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqcup" "⊔" nil :system t)
    (puthash "\\oplus" "⊕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oplus" "⊕" nil :system t)
    (puthash "\\ominus" "⊖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ominus" "⊖" nil :system t)
    (puthash "\\otimes" "⊗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\otimes" "⊗" nil :system t)
    (puthash "\\oslash" "⊘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oslash" "⊘" nil :system t)
    (puthash "\\odot" "⊙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\odot" "⊙" nil :system t)
    (puthash "\\circledcirc" "⊚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledcirc" "⊚" nil :system t)
    (puthash "\\circledast" "⊛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledast" "⊛" nil :system t)
    (puthash "\\circledequal" "⊜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledequal" "⊜" nil :system t)
    (puthash "\\circleddash" "⊝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circleddash" "⊝" nil :system t)
    (puthash "\\boxplus" "⊞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxplus" "⊞" nil :system t)
    (puthash "\\boxminus" "⊟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxminus" "⊟" nil :system t)
    (puthash "\\boxtimes" "⊠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxtimes" "⊠" nil :system t)
    (puthash "\\boxdot" "⊡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxdot" "⊡" nil :system t)
    (puthash "\\vdash" "⊢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vdash" "⊢" nil :system t)
    (puthash "\\dashv" "⊣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dashv" "⊣" nil :system t)
    (puthash "\\top" "⊤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\top" "⊤" nil :system t)
    (puthash "\\bot" "⊥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bot" "⊥" nil :system t)
    (puthash "\\models" "⊧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\models" "⊧" nil :system t)
    (puthash "\\vDash" "⊨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vDash" "⊨" nil :system t)
    (puthash "\\Vdash" "⊩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Vdash" "⊩" nil :system t)
    (puthash "\\Vvdash" "⊪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Vvdash" "⊪" nil :system t)
    (puthash "\\VDash" "⊫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\VDash" "⊫" nil :system t)
    (puthash "\\nvdash" "⊬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvdash" "⊬" nil :system t)
    (puthash "\\nvDash" "⊭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvDash" "⊭" nil :system t)
    (puthash "\\nVdash" "⊮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVdash" "⊮" nil :system t)
    (puthash "\\nVDash" "⊯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVDash" "⊯" nil :system t)
    (puthash "\\prurel" "⊰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\prurel" "⊰" nil :system t)
    (puthash "\\scurel" "⊱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scurel" "⊱" nil :system t)
    (puthash "\\vartriangleleft" "⊲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vartriangleleft" "⊲" nil :system t)
    (puthash "\\vartriangleright" "⊳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vartriangleright" "⊳" nil :system t)
    (puthash "\\trianglelefteq" "⊴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trianglelefteq" "⊴" nil :system t)
    (puthash "\\trianglerighteq" "⊵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trianglerighteq" "⊵" nil :system t)
    (puthash "\\original" "⊶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\original" "⊶" nil :system t)
    (puthash "\\image" "⊷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\image" "⊷" nil :system t)
    (puthash "\\multimap" "⊸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\multimap" "⊸" nil :system t)
    (puthash "\\hermitconjmatrix" "⊹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hermitconjmatrix" "⊹" nil :system t)
    (puthash "\\intercal" "⊺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intercal" "⊺" nil :system t)
    (puthash "\\xor" "⊻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\xor" "⊻" nil :system t)
    (puthash "\\veebar" "⊻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\veebar" "⊻" nil :system t)
    (puthash "\\barwedge" "⊼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barwedge" "⊼" nil :system t)
    (puthash "\\nand" "⊼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nand" "⊼" nil :system t)
    (puthash "\\barvee" "⊽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barvee" "⊽" nil :system t)
    (puthash "\\nor" "⊽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nor" "⊽" nil :system t)
    (puthash "\\rightanglearc" "⊾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightanglearc" "⊾" nil :system t)
    (puthash "\\varlrtriangle" "⊿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varlrtriangle" "⊿" nil :system t)
    (puthash "\\bigwedge" "⋀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigwedge" "⋀" nil :system t)
    (puthash "\\bigvee" "⋁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigvee" "⋁" nil :system t)
    (puthash "\\bigcap" "⋂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigcap" "⋂" nil :system t)
    (puthash "\\bigcup" "⋃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigcup" "⋃" nil :system t)
    (puthash "\\diamond" "⋄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamond" "⋄" nil :system t)
    (puthash "\\cdot" "⋅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cdot" "⋅" nil :system t)
    (puthash "\\star" "⋆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\star" "⋆" nil :system t)
    (puthash "\\divideontimes" "⋇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\divideontimes" "⋇" nil :system t)
    (puthash "\\bowtie" "⋈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bowtie" "⋈" nil :system t)
    (puthash "\\ltimes" "⋉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltimes" "⋉" nil :system t)
    (puthash "\\rtimes" "⋊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rtimes" "⋊" nil :system t)
    (puthash "\\leftthreetimes" "⋋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftthreetimes" "⋋" nil :system t)
    (puthash "\\rightthreetimes" "⋌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightthreetimes" "⋌" nil :system t)
    (puthash "\\backsimeq" "⋍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\backsimeq" "⋍" nil :system t)
    (puthash "\\curlyvee" "⋎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\curlyvee" "⋎" nil :system t)
    (puthash "\\curlywedge" "⋏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\curlywedge" "⋏" nil :system t)
    (puthash "\\Subset" "⋐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Subset" "⋐" nil :system t)
    (puthash "\\Supset" "⋑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Supset" "⋑" nil :system t)
    (puthash "\\Cap" "⋒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Cap" "⋒" nil :system t)
    (puthash "\\Cup" "⋓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Cup" "⋓" nil :system t)
    (puthash "\\pitchfork" "⋔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pitchfork" "⋔" nil :system t)
    (puthash "\\equalparallel" "⋕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\equalparallel" "⋕" nil :system t)
    (puthash "\\lessdot" "⋖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lessdot" "⋖" nil :system t)
    (puthash "\\gtrdot" "⋗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtrdot" "⋗" nil :system t)
    (puthash "\\verymuchless" "⋘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\verymuchless" "⋘" nil :system t)
    (puthash "\\ggg" "⋙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ggg" "⋙" nil :system t)
    (puthash "\\lesseqgtr" "⋚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesseqgtr" "⋚" nil :system t)
    (puthash "\\gtreqless" "⋛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtreqless" "⋛" nil :system t)
    (puthash "\\eqless" "⋜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqless" "⋜" nil :system t)
    (puthash "\\eqgtr" "⋝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqgtr" "⋝" nil :system t)
    (puthash "\\curlyeqprec" "⋞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\curlyeqprec" "⋞" nil :system t)
    (puthash "\\curlyeqsucc" "⋟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\curlyeqsucc" "⋟" nil :system t)
    (puthash "\\npreccurlyeq" "⋠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\npreccurlyeq" "⋠" nil :system t)
    (puthash "\\nsucccurlyeq" "⋡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsucccurlyeq" "⋡" nil :system t)
    (puthash "\\nsqsubseteq" "⋢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsqsubseteq" "⋢" nil :system t)
    (puthash "\\nsqsupseteq" "⋣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsqsupseteq" "⋣" nil :system t)
    (puthash "\\sqsubsetneq" "⋤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqsubsetneq" "⋤" nil :system t)
    (puthash "\\sqsupsetneq" "⋥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqsupsetneq" "⋥" nil :system t)
    (puthash "\\lnsim" "⋦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lnsim" "⋦" nil :system t)
    (puthash "\\gnsim" "⋧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gnsim" "⋧" nil :system t)
    (puthash "\\precnsim" "⋨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\precnsim" "⋨" nil :system t)
    (puthash "\\succnsim" "⋩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succnsim" "⋩" nil :system t)
    (puthash "\\ntriangleleft" "⋪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ntriangleleft" "⋪" nil :system t)
    (puthash "\\ntriangleright" "⋫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ntriangleright" "⋫" nil :system t)
    (puthash "\\ntrianglelefteq" "⋬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ntrianglelefteq" "⋬" nil :system t)
    (puthash "\\ntrianglerighteq" "⋭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ntrianglerighteq" "⋭" nil :system t)
    (puthash "\\vdots" "⋮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vdots" "⋮" nil :system t)
    (puthash "\\cdots" "⋯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cdots" "⋯" nil :system t)
    (puthash "\\adots" "⋰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\adots" "⋰" nil :system t)
    (puthash "\\ddots" "⋱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ddots" "⋱" nil :system t)
    (puthash "\\disin" "⋲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\disin" "⋲" nil :system t)
    (puthash "\\varisins" "⋳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varisins" "⋳" nil :system t)
    (puthash "\\isins" "⋴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isins" "⋴" nil :system t)
    (puthash "\\isindot" "⋵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isindot" "⋵" nil :system t)
    (puthash "\\varisinobar" "⋶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varisinobar" "⋶" nil :system t)
    (puthash "\\isinobar" "⋷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isinobar" "⋷" nil :system t)
    (puthash "\\isinvb" "⋸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isinvb" "⋸" nil :system t)
    (puthash "\\isinE" "⋹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isinE" "⋹" nil :system t)
    (puthash "\\nisd" "⋺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nisd" "⋺" nil :system t)
    (puthash "\\varnis" "⋻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varnis" "⋻" nil :system t)
    (puthash "\\nis" "⋼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nis" "⋼" nil :system t)
    (puthash "\\varniobar" "⋽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varniobar" "⋽" nil :system t)
    (puthash "\\niobar" "⋾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\niobar" "⋾" nil :system t)
    (puthash "\\bagmember" "⋿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bagmember" "⋿" nil :system t)
    (puthash "\\diameter" "⌀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diameter" "⌀" nil :system t)
    (puthash "\\house" "⌂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\house" "⌂" nil :system t)
    (puthash "\\varbarwedge" "⌅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varbarwedge" "⌅" nil :system t)
    (puthash "\\vardoublebarwedge" "⌆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vardoublebarwedge" "⌆" nil :system t)
    (puthash "\\lceil" "⌈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lceil" "⌈" nil :system t)
    (puthash "\\rceil" "⌉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rceil" "⌉" nil :system t)
    (puthash "\\lfloor" "⌊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lfloor" "⌊" nil :system t)
    (puthash "\\rfloor" "⌋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rfloor" "⌋" nil :system t)
    (puthash "\\invnot" "⌐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\invnot" "⌐" nil :system t)
    (puthash "\\sqlozenge" "⌑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqlozenge" "⌑" nil :system t)
    (puthash "\\profline" "⌒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\profline" "⌒" nil :system t)
    (puthash "\\profsurf" "⌓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\profsurf" "⌓" nil :system t)
    (puthash "\\recorder" "⌕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\recorder" "⌕" nil :system t)
    (puthash "\\viewdata" "⌗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\viewdata" "⌗" nil :system t)
    (puthash "\\turnednot" "⌙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\turnednot" "⌙" nil :system t)
    (puthash "\\:watch:" "⌚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:watch:" "⌚" nil :system t)
    (puthash "\\:hourglass:" "⌛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hourglass:" "⌛" nil :system t)
    (puthash "\\ulcorner" "⌜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ulcorner" "⌜" nil :system t)
    (puthash "\\urcorner" "⌝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\urcorner" "⌝" nil :system t)
    (puthash "\\llcorner" "⌞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\llcorner" "⌞" nil :system t)
    (puthash "\\lrcorner" "⌟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lrcorner" "⌟" nil :system t)
    (puthash "\\frown" "⌢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frown" "⌢" nil :system t)
    (puthash "\\smile" "⌣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smile" "⌣" nil :system t)
    (puthash "\\varhexagonlrbonds" "⌬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varhexagonlrbonds" "⌬" nil :system t)
    (puthash "\\conictaper" "⌲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\conictaper" "⌲" nil :system t)
    (puthash "\\topbot" "⌶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\topbot" "⌶" nil :system t)
    (puthash "\\obar" "⌽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\obar" "⌽" nil :system t)
    (puthash "\\notslash" "⌿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\notslash" "⌿" nil :system t)
    (puthash "\\notbackslash" "⍀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\notbackslash" "⍀" nil :system t)
    (puthash "\\boxupcaret" "⍓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxupcaret" "⍓" nil :system t)
    (puthash "\\boxquestion" "⍰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxquestion" "⍰" nil :system t)
    (puthash "\\hexagon" "⎔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hexagon" "⎔" nil :system t)
    (puthash "\\dlcorn" "⎣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dlcorn" "⎣" nil :system t)
    (puthash "\\lmoustache" "⎰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lmoustache" "⎰" nil :system t)
    (puthash "\\rmoustache" "⎱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rmoustache" "⎱" nil :system t)
    (puthash "\\overbracket" "⎴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\overbracket" "⎴" nil :system t)
    (puthash "\\underbracket" "⎵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underbracket" "⎵" nil :system t)
    (puthash "\\bbrktbrk" "⎶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbrktbrk" "⎶" nil :system t)
    (puthash "\\sqrtbottom" "⎷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqrtbottom" "⎷" nil :system t)
    (puthash "\\lvboxline" "⎸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lvboxline" "⎸" nil :system t)
    (puthash "\\rvboxline" "⎹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rvboxline" "⎹" nil :system t)
    (puthash "\\varcarriagereturn" "⏎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varcarriagereturn" "⏎" nil :system t)
    (puthash "\\overbrace" "⏞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\overbrace" "⏞" nil :system t)
    (puthash "\\underbrace" "⏟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\underbrace" "⏟" nil :system t)
    (puthash "\\trapezium" "⏢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trapezium" "⏢" nil :system t)
    (puthash "\\benzenr" "⏣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\benzenr" "⏣" nil :system t)
    (puthash "\\strns" "⏤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\strns" "⏤" nil :system t)
    (puthash "\\fltns" "⏥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fltns" "⏥" nil :system t)
    (puthash "\\accurrent" "⏦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\accurrent" "⏦" nil :system t)
    (puthash "\\elinters" "⏧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\elinters" "⏧" nil :system t)
    (puthash "\\:fast_forward:" "⏩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fast_forward:" "⏩" nil :system t)
    (puthash "\\:rewind:" "⏪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rewind:" "⏪" nil :system t)
    (puthash "\\:arrow_double_up:" "⏫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_double_up:" "⏫" nil :system t)
    (puthash "\\:arrow_double_down:" "⏬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_double_down:" "⏬" nil :system t)
    (puthash "\\:alarm_clock:" "⏰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:alarm_clock:" "⏰" nil :system t)
    (puthash "\\:hourglass_flowing_sand:" "⏳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hourglass_flowing_sand:" "⏳" nil :system t)
    (puthash "\\blanksymbol" "␢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blanksymbol" "␢" nil :system t)
    (puthash "\\visiblespace" "␣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\visiblespace" "␣" nil :system t)
    (puthash "\\:m:" "Ⓜ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:m:" "Ⓜ" nil :system t)
    (puthash "\\circledS" "Ⓢ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledS" "Ⓢ" nil :system t)
    (puthash "\\dshfnc" "┆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dshfnc" "┆" nil :system t)
    (puthash "\\sqfnw" "┙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqfnw" "┙" nil :system t)
    (puthash "\\diagup" "╱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diagup" "╱" nil :system t)
    (puthash "\\diagdown" "╲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diagdown" "╲" nil :system t)
    (puthash "\\blockuphalf" "▀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blockuphalf" "▀" nil :system t)
    (puthash "\\blocklowhalf" "▄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blocklowhalf" "▄" nil :system t)
    (puthash "\\blockfull" "█" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blockfull" "█" nil :system t)
    (puthash "\\blocklefthalf" "▌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blocklefthalf" "▌" nil :system t)
    (puthash "\\blockrighthalf" "▐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blockrighthalf" "▐" nil :system t)
    (puthash "\\blockqtrshaded" "░" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blockqtrshaded" "░" nil :system t)
    (puthash "\\blockhalfshaded" "▒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blockhalfshaded" "▒" nil :system t)
    (puthash "\\blockthreeqtrshaded" "▓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blockthreeqtrshaded" "▓" nil :system t)
    (puthash "\\blacksquare" "■" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacksquare" "■" nil :system t)
    (puthash "\\square" "□" table)
    (define-abbrev julia-latexsub-abbrev-table "\\square" "□" nil :system t)
    (puthash "\\squoval" "▢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squoval" "▢" nil :system t)
    (puthash "\\blackinwhitesquare" "▣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackinwhitesquare" "▣" nil :system t)
    (puthash "\\squarehfill" "▤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarehfill" "▤" nil :system t)
    (puthash "\\squarevfill" "▥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarevfill" "▥" nil :system t)
    (puthash "\\squarehvfill" "▦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarehvfill" "▦" nil :system t)
    (puthash "\\squarenwsefill" "▧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarenwsefill" "▧" nil :system t)
    (puthash "\\squareneswfill" "▨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squareneswfill" "▨" nil :system t)
    (puthash "\\squarecrossfill" "▩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarecrossfill" "▩" nil :system t)
    (puthash "\\smblksquare" "▪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smblksquare" "▪" nil :system t)
    (puthash "\\:black_small_square:" "▪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_small_square:" "▪" nil :system t)
    (puthash "\\:white_small_square:" "▫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_small_square:" "▫" nil :system t)
    (puthash "\\smwhtsquare" "▫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smwhtsquare" "▫" nil :system t)
    (puthash "\\hrectangleblack" "▬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hrectangleblack" "▬" nil :system t)
    (puthash "\\hrectangle" "▭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hrectangle" "▭" nil :system t)
    (puthash "\\vrectangleblack" "▮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vrectangleblack" "▮" nil :system t)
    (puthash "\\vrecto" "▯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vrecto" "▯" nil :system t)
    (puthash "\\parallelogramblack" "▰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\parallelogramblack" "▰" nil :system t)
    (puthash "\\parallelogram" "▱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\parallelogram" "▱" nil :system t)
    (puthash "\\bigblacktriangleup" "▲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigblacktriangleup" "▲" nil :system t)
    (puthash "\\bigtriangleup" "△" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigtriangleup" "△" nil :system t)
    (puthash "\\blacktriangle" "▴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacktriangle" "▴" nil :system t)
    (puthash "\\vartriangle" "▵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vartriangle" "▵" nil :system t)
    (puthash "\\:arrow_forward:" "▶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_forward:" "▶" nil :system t)
    (puthash "\\blacktriangleright" "▶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacktriangleright" "▶" nil :system t)
    (puthash "\\triangleright" "▷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangleright" "▷" nil :system t)
    (puthash "\\smallblacktriangleright" "▸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smallblacktriangleright" "▸" nil :system t)
    (puthash "\\smalltriangleright" "▹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smalltriangleright" "▹" nil :system t)
    (puthash "\\blackpointerright" "►" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackpointerright" "►" nil :system t)
    (puthash "\\whitepointerright" "▻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\whitepointerright" "▻" nil :system t)
    (puthash "\\bigblacktriangledown" "▼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigblacktriangledown" "▼" nil :system t)
    (puthash "\\bigtriangledown" "▽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigtriangledown" "▽" nil :system t)
    (puthash "\\blacktriangledown" "▾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacktriangledown" "▾" nil :system t)
    (puthash "\\triangledown" "▿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangledown" "▿" nil :system t)
    (puthash "\\blacktriangleleft" "◀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacktriangleleft" "◀" nil :system t)
    (puthash "\\:arrow_backward:" "◀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_backward:" "◀" nil :system t)
    (puthash "\\triangleleft" "◁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangleleft" "◁" nil :system t)
    (puthash "\\smallblacktriangleleft" "◂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smallblacktriangleleft" "◂" nil :system t)
    (puthash "\\smalltriangleleft" "◃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smalltriangleleft" "◃" nil :system t)
    (puthash "\\blackpointerleft" "◄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackpointerleft" "◄" nil :system t)
    (puthash "\\whitepointerleft" "◅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\whitepointerleft" "◅" nil :system t)
    (puthash "\\mdlgblkdiamond" "◆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdlgblkdiamond" "◆" nil :system t)
    (puthash "\\mdlgwhtdiamond" "◇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdlgwhtdiamond" "◇" nil :system t)
    (puthash "\\blackinwhitediamond" "◈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackinwhitediamond" "◈" nil :system t)
    (puthash "\\fisheye" "◉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fisheye" "◉" nil :system t)
    (puthash "\\lozenge" "◊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lozenge" "◊" nil :system t)
    (puthash "\\bigcirc" "○" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigcirc" "○" nil :system t)
    (puthash "\\dottedcircle" "◌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dottedcircle" "◌" nil :system t)
    (puthash "\\circlevertfill" "◍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circlevertfill" "◍" nil :system t)
    (puthash "\\bullseye" "◎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bullseye" "◎" nil :system t)
    (puthash "\\mdlgblkcircle" "●" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdlgblkcircle" "●" nil :system t)
    (puthash "\\cirfl" "◐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cirfl" "◐" nil :system t)
    (puthash "\\cirfr" "◑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cirfr" "◑" nil :system t)
    (puthash "\\cirfb" "◒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cirfb" "◒" nil :system t)
    (puthash "\\circletophalfblack" "◓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circletophalfblack" "◓" nil :system t)
    (puthash "\\circleurquadblack" "◔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circleurquadblack" "◔" nil :system t)
    (puthash "\\blackcircleulquadwhite" "◕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackcircleulquadwhite" "◕" nil :system t)
    (puthash "\\blacklefthalfcircle" "◖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacklefthalfcircle" "◖" nil :system t)
    (puthash "\\blackrighthalfcircle" "◗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackrighthalfcircle" "◗" nil :system t)
    (puthash "\\rvbull" "◘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rvbull" "◘" nil :system t)
    (puthash "\\inversewhitecircle" "◙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\inversewhitecircle" "◙" nil :system t)
    (puthash "\\invwhiteupperhalfcircle" "◚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\invwhiteupperhalfcircle" "◚" nil :system t)
    (puthash "\\invwhitelowerhalfcircle" "◛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\invwhitelowerhalfcircle" "◛" nil :system t)
    (puthash "\\ularc" "◜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ularc" "◜" nil :system t)
    (puthash "\\urarc" "◝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\urarc" "◝" nil :system t)
    (puthash "\\lrarc" "◞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lrarc" "◞" nil :system t)
    (puthash "\\llarc" "◟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\llarc" "◟" nil :system t)
    (puthash "\\topsemicircle" "◠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\topsemicircle" "◠" nil :system t)
    (puthash "\\botsemicircle" "◡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\botsemicircle" "◡" nil :system t)
    (puthash "\\lrblacktriangle" "◢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lrblacktriangle" "◢" nil :system t)
    (puthash "\\llblacktriangle" "◣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\llblacktriangle" "◣" nil :system t)
    (puthash "\\ulblacktriangle" "◤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ulblacktriangle" "◤" nil :system t)
    (puthash "\\urblacktriangle" "◥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\urblacktriangle" "◥" nil :system t)
    (puthash "\\smwhtcircle" "◦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smwhtcircle" "◦" nil :system t)
    (puthash "\\sqfl" "◧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqfl" "◧" nil :system t)
    (puthash "\\sqfr" "◨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqfr" "◨" nil :system t)
    (puthash "\\squareulblack" "◩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squareulblack" "◩" nil :system t)
    (puthash "\\sqfse" "◪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqfse" "◪" nil :system t)
    (puthash "\\boxbar" "◫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxbar" "◫" nil :system t)
    (puthash "\\trianglecdot" "◬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trianglecdot" "◬" nil :system t)
    (puthash "\\triangleleftblack" "◭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangleleftblack" "◭" nil :system t)
    (puthash "\\trianglerightblack" "◮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\trianglerightblack" "◮" nil :system t)
    (puthash "\\lgwhtcircle" "◯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lgwhtcircle" "◯" nil :system t)
    (puthash "\\squareulquad" "◰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squareulquad" "◰" nil :system t)
    (puthash "\\squarellquad" "◱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarellquad" "◱" nil :system t)
    (puthash "\\squarelrquad" "◲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarelrquad" "◲" nil :system t)
    (puthash "\\squareurquad" "◳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squareurquad" "◳" nil :system t)
    (puthash "\\circleulquad" "◴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circleulquad" "◴" nil :system t)
    (puthash "\\circlellquad" "◵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circlellquad" "◵" nil :system t)
    (puthash "\\circlelrquad" "◶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circlelrquad" "◶" nil :system t)
    (puthash "\\circleurquad" "◷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circleurquad" "◷" nil :system t)
    (puthash "\\ultriangle" "◸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ultriangle" "◸" nil :system t)
    (puthash "\\urtriangle" "◹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\urtriangle" "◹" nil :system t)
    (puthash "\\lltriangle" "◺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lltriangle" "◺" nil :system t)
    (puthash "\\mdwhtsquare" "◻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdwhtsquare" "◻" nil :system t)
    (puthash "\\:white_medium_square:" "◻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_medium_square:" "◻" nil :system t)
    (puthash "\\:black_medium_square:" "◼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_medium_square:" "◼" nil :system t)
    (puthash "\\mdblksquare" "◼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdblksquare" "◼" nil :system t)
    (puthash "\\:white_medium_small_square:" "◽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_medium_small_square:" "◽" nil :system t)
    (puthash "\\mdsmwhtsquare" "◽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdsmwhtsquare" "◽" nil :system t)
    (puthash "\\mdsmblksquare" "◾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdsmblksquare" "◾" nil :system t)
    (puthash "\\:black_medium_small_square:" "◾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_medium_small_square:" "◾" nil :system t)
    (puthash "\\lrtriangle" "◿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lrtriangle" "◿" nil :system t)
    (puthash "\\:sunny:" "☀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sunny:" "☀" nil :system t)
    (puthash "\\:cloud:" "☁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cloud:" "☁" nil :system t)
    (puthash "\\bigstar" "★" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigstar" "★" nil :system t)
    (puthash "\\bigwhitestar" "☆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigwhitestar" "☆" nil :system t)
    (puthash "\\astrosun" "☉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\astrosun" "☉" nil :system t)
    (puthash "\\:phone:" "☎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:phone:" "☎" nil :system t)
    (puthash "\\:ballot_box_with_check:" "☑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ballot_box_with_check:" "☑" nil :system t)
    (puthash "\\:umbrella_with_rain_drops:" "☔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:umbrella_with_rain_drops:" "☔" nil :system t)
    (puthash "\\:umbrella:" "☔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:umbrella:" "☔" nil :system t)
    (puthash "\\:coffee:" "☕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:coffee:" "☕" nil :system t)
    (puthash "\\:point_up:" "☝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:point_up:" "☝" nil :system t)
    (puthash "\\danger" "☡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\danger" "☡" nil :system t)
    (puthash "\\:relaxed:" "☺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:relaxed:" "☺" nil :system t)
    (puthash "\\blacksmiley" "☻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacksmiley" "☻" nil :system t)
    (puthash "\\sun" "☼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sun" "☼" nil :system t)
    (puthash "\\rightmoon" "☽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightmoon" "☽" nil :system t)
    (puthash "\\leftmoon" "☾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftmoon" "☾" nil :system t)
    (puthash "\\mercury" "☿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mercury" "☿" nil :system t)
    (puthash "\\venus" "♀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\venus" "♀" nil :system t)
    (puthash "\\female" "♀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\female" "♀" nil :system t)
    (puthash "\\mars" "♂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mars" "♂" nil :system t)
    (puthash "\\male" "♂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\male" "♂" nil :system t)
    (puthash "\\jupiter" "♃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\jupiter" "♃" nil :system t)
    (puthash "\\saturn" "♄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\saturn" "♄" nil :system t)
    (puthash "\\uranus" "♅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\uranus" "♅" nil :system t)
    (puthash "\\neptune" "♆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neptune" "♆" nil :system t)
    (puthash "\\pluto" "♇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pluto" "♇" nil :system t)
    (puthash "\\aries" "♈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\aries" "♈" nil :system t)
    (puthash "\\:aries:" "♈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:aries:" "♈" nil :system t)
    (puthash "\\:taurus:" "♉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:taurus:" "♉" nil :system t)
    (puthash "\\taurus" "♉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\taurus" "♉" nil :system t)
    (puthash "\\gemini" "♊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gemini" "♊" nil :system t)
    (puthash "\\:gemini:" "♊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gemini:" "♊" nil :system t)
    (puthash "\\:cancer:" "♋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cancer:" "♋" nil :system t)
    (puthash "\\cancer" "♋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cancer" "♋" nil :system t)
    (puthash "\\leo" "♌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leo" "♌" nil :system t)
    (puthash "\\:leo:" "♌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:leo:" "♌" nil :system t)
    (puthash "\\:virgo:" "♍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:virgo:" "♍" nil :system t)
    (puthash "\\virgo" "♍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\virgo" "♍" nil :system t)
    (puthash "\\libra" "♎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\libra" "♎" nil :system t)
    (puthash "\\:libra:" "♎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:libra:" "♎" nil :system t)
    (puthash "\\scorpio" "♏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scorpio" "♏" nil :system t)
    (puthash "\\:scorpius:" "♏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scorpius:" "♏" nil :system t)
    (puthash "\\sagittarius" "♐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sagittarius" "♐" nil :system t)
    (puthash "\\:sagittarius:" "♐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sagittarius:" "♐" nil :system t)
    (puthash "\\capricornus" "♑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\capricornus" "♑" nil :system t)
    (puthash "\\:capricorn:" "♑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:capricorn:" "♑" nil :system t)
    (puthash "\\:aquarius:" "♒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:aquarius:" "♒" nil :system t)
    (puthash "\\aquarius" "♒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\aquarius" "♒" nil :system t)
    (puthash "\\pisces" "♓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pisces" "♓" nil :system t)
    (puthash "\\:pisces:" "♓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pisces:" "♓" nil :system t)
    (puthash "\\spadesuit" "♠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\spadesuit" "♠" nil :system t)
    (puthash "\\:spades:" "♠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:spades:" "♠" nil :system t)
    (puthash "\\heartsuit" "♡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\heartsuit" "♡" nil :system t)
    (puthash "\\diamondsuit" "♢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondsuit" "♢" nil :system t)
    (puthash "\\:clubs:" "♣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clubs:" "♣" nil :system t)
    (puthash "\\clubsuit" "♣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\clubsuit" "♣" nil :system t)
    (puthash "\\varspadesuit" "♤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varspadesuit" "♤" nil :system t)
    (puthash "\\varheartsuit" "♥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varheartsuit" "♥" nil :system t)
    (puthash "\\:hearts:" "♥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hearts:" "♥" nil :system t)
    (puthash "\\vardiamondsuit" "♦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vardiamondsuit" "♦" nil :system t)
    (puthash "\\:diamonds:" "♦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:diamonds:" "♦" nil :system t)
    (puthash "\\varclubsuit" "♧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varclubsuit" "♧" nil :system t)
    (puthash "\\:hotsprings:" "♨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hotsprings:" "♨" nil :system t)
    (puthash "\\quarternote" "♩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\quarternote" "♩" nil :system t)
    (puthash "\\eighthnote" "♪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eighthnote" "♪" nil :system t)
    (puthash "\\twonotes" "♫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twonotes" "♫" nil :system t)
    (puthash "\\flat" "♭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\flat" "♭" nil :system t)
    (puthash "\\natural" "♮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\natural" "♮" nil :system t)
    (puthash "\\sharp" "♯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sharp" "♯" nil :system t)
    (puthash "\\:recycle:" "♻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:recycle:" "♻" nil :system t)
    (puthash "\\acidfree" "♾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\acidfree" "♾" nil :system t)
    (puthash "\\:wheelchair:" "♿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wheelchair:" "♿" nil :system t)
    (puthash "\\dicei" "⚀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dicei" "⚀" nil :system t)
    (puthash "\\diceii" "⚁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diceii" "⚁" nil :system t)
    (puthash "\\diceiii" "⚂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diceiii" "⚂" nil :system t)
    (puthash "\\diceiv" "⚃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diceiv" "⚃" nil :system t)
    (puthash "\\dicev" "⚄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dicev" "⚄" nil :system t)
    (puthash "\\dicevi" "⚅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dicevi" "⚅" nil :system t)
    (puthash "\\circledrightdot" "⚆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledrightdot" "⚆" nil :system t)
    (puthash "\\circledtwodots" "⚇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledtwodots" "⚇" nil :system t)
    (puthash "\\blackcircledrightdot" "⚈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackcircledrightdot" "⚈" nil :system t)
    (puthash "\\blackcircledtwodots" "⚉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blackcircledtwodots" "⚉" nil :system t)
    (puthash "\\:anchor:" "⚓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:anchor:" "⚓" nil :system t)
    (puthash "\\:warning:" "⚠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:warning:" "⚠" nil :system t)
    (puthash "\\:zap:" "⚡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:zap:" "⚡" nil :system t)
    (puthash "\\hermaphrodite" "⚥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hermaphrodite" "⚥" nil :system t)
    (puthash "\\:white_circle:" "⚪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_circle:" "⚪" nil :system t)
    (puthash "\\mdwhtcircle" "⚪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdwhtcircle" "⚪" nil :system t)
    (puthash "\\mdblkcircle" "⚫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdblkcircle" "⚫" nil :system t)
    (puthash "\\:black_circle:" "⚫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_circle:" "⚫" nil :system t)
    (puthash "\\mdsmwhtcircle" "⚬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdsmwhtcircle" "⚬" nil :system t)
    (puthash "\\neuter" "⚲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neuter" "⚲" nil :system t)
    (puthash "\\:soccer:" "⚽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:soccer:" "⚽" nil :system t)
    (puthash "\\:baseball:" "⚾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baseball:" "⚾" nil :system t)
    (puthash "\\:snowman_without_snow:" "⛄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:snowman_without_snow:" "⛄" nil :system t)
    (puthash "\\:snowman:" "⛄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:snowman:" "⛄" nil :system t)
    (puthash "\\:partly_sunny:" "⛅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:partly_sunny:" "⛅" nil :system t)
    (puthash "\\:ophiuchus:" "⛎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ophiuchus:" "⛎" nil :system t)
    (puthash "\\:no_entry:" "⛔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_entry:" "⛔" nil :system t)
    (puthash "\\:church:" "⛪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:church:" "⛪" nil :system t)
    (puthash "\\:fountain:" "⛲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fountain:" "⛲" nil :system t)
    (puthash "\\:golf:" "⛳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:golf:" "⛳" nil :system t)
    (puthash "\\:boat:" "⛵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boat:" "⛵" nil :system t)
    (puthash "\\:tent:" "⛺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tent:" "⛺" nil :system t)
    (puthash "\\:fuelpump:" "⛽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fuelpump:" "⛽" nil :system t)
    (puthash "\\:scissors:" "✂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scissors:" "✂" nil :system t)
    (puthash "\\:white_check_mark:" "✅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_check_mark:" "✅" nil :system t)
    (puthash "\\:airplane:" "✈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:airplane:" "✈" nil :system t)
    (puthash "\\:email:" "✉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:email:" "✉" nil :system t)
    (puthash "\\:fist:" "✊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fist:" "✊" nil :system t)
    (puthash "\\:hand:" "✋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hand:" "✋" nil :system t)
    (puthash "\\:v:" "✌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:v:" "✌" nil :system t)
    (puthash "\\:pencil2:" "✏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pencil2:" "✏" nil :system t)
    (puthash "\\:black_nib:" "✒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_nib:" "✒" nil :system t)
    (puthash "\\checkmark" "✓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\checkmark" "✓" nil :system t)
    (puthash "\\:heavy_check_mark:" "✔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heavy_check_mark:" "✔" nil :system t)
    (puthash "\\:heavy_multiplication_x:" "✖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heavy_multiplication_x:" "✖" nil :system t)
    (puthash "\\maltese" "✠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\maltese" "✠" nil :system t)
    (puthash "\\:sparkles:" "✨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sparkles:" "✨" nil :system t)
    (puthash "\\circledstar" "✪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledstar" "✪" nil :system t)
    (puthash "\\:eight_spoked_asterisk:" "✳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:eight_spoked_asterisk:" "✳" nil :system t)
    (puthash "\\:eight_pointed_black_star:" "✴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:eight_pointed_black_star:" "✴" nil :system t)
    (puthash "\\varstar" "✶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varstar" "✶" nil :system t)
    (puthash "\\dingasterisk" "✽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dingasterisk" "✽" nil :system t)
    (puthash "\\:snowflake:" "❄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:snowflake:" "❄" nil :system t)
    (puthash "\\:sparkle:" "❇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sparkle:" "❇" nil :system t)
    (puthash "\\:x:" "❌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:x:" "❌" nil :system t)
    (puthash "\\:negative_squared_cross_mark:" "❎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:negative_squared_cross_mark:" "❎" nil :system t)
    (puthash "\\:question:" "❓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:question:" "❓" nil :system t)
    (puthash "\\:grey_question:" "❔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:grey_question:" "❔" nil :system t)
    (puthash "\\:grey_exclamation:" "❕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:grey_exclamation:" "❕" nil :system t)
    (puthash "\\:exclamation:" "❗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:exclamation:" "❗" nil :system t)
    (puthash "\\:heart:" "❤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heart:" "❤" nil :system t)
    (puthash "\\:heavy_plus_sign:" "➕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heavy_plus_sign:" "➕" nil :system t)
    (puthash "\\:heavy_minus_sign:" "➖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heavy_minus_sign:" "➖" nil :system t)
    (puthash "\\:heavy_division_sign:" "➗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heavy_division_sign:" "➗" nil :system t)
    (puthash "\\draftingarrow" "➛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\draftingarrow" "➛" nil :system t)
    (puthash "\\:arrow_right:" "➡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_right:" "➡" nil :system t)
    (puthash "\\:curly_loop:" "➰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:curly_loop:" "➰" nil :system t)
    (puthash "\\:loop:" "➿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:loop:" "➿" nil :system t)
    (puthash "\\threedangle" "⟀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\threedangle" "⟀" nil :system t)
    (puthash "\\whiteinwhitetriangle" "⟁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\whiteinwhitetriangle" "⟁" nil :system t)
    (puthash "\\perp" "⟂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\perp" "⟂" nil :system t)
    (puthash "\\veedot" "⟇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\veedot" "⟇" nil :system t)
    (puthash "\\bsolhsub" "⟈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsolhsub" "⟈" nil :system t)
    (puthash "\\suphsol" "⟉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\suphsol" "⟉" nil :system t)
    (puthash "\\wedgedot" "⟑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedgedot" "⟑" nil :system t)
    (puthash "\\upin" "⟒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upin" "⟒" nil :system t)
    (puthash "\\leftouterjoin" "⟕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftouterjoin" "⟕" nil :system t)
    (puthash "\\rightouterjoin" "⟖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightouterjoin" "⟖" nil :system t)
    (puthash "\\fullouterjoin" "⟗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fullouterjoin" "⟗" nil :system t)
    (puthash "\\bigbot" "⟘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigbot" "⟘" nil :system t)
    (puthash "\\bigtop" "⟙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigtop" "⟙" nil :system t)
    (puthash "\\llbracket" "⟦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\llbracket" "⟦" nil :system t)
    (puthash "\\openbracketleft" "⟦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\openbracketleft" "⟦" nil :system t)
    (puthash "\\openbracketright" "⟧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\openbracketright" "⟧" nil :system t)
    (puthash "\\rrbracket" "⟧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rrbracket" "⟧" nil :system t)
    (puthash "\\langle" "⟨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\langle" "⟨" nil :system t)
    (puthash "\\rangle" "⟩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rangle" "⟩" nil :system t)
    (puthash "\\UUparrow" "⟰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\UUparrow" "⟰" nil :system t)
    (puthash "\\DDownarrow" "⟱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DDownarrow" "⟱" nil :system t)
    (puthash "\\longleftarrow" "⟵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longleftarrow" "⟵" nil :system t)
    (puthash "\\longrightarrow" "⟶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longrightarrow" "⟶" nil :system t)
    (puthash "\\longleftrightarrow" "⟷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longleftrightarrow" "⟷" nil :system t)
    (puthash "\\impliedby" "⟸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\impliedby" "⟸" nil :system t)
    (puthash "\\Longleftarrow" "⟸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Longleftarrow" "⟸" nil :system t)
    (puthash "\\implies" "⟹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\implies" "⟹" nil :system t)
    (puthash "\\Longrightarrow" "⟹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Longrightarrow" "⟹" nil :system t)
    (puthash "\\iff" "⟺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iff" "⟺" nil :system t)
    (puthash "\\Longleftrightarrow" "⟺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Longleftrightarrow" "⟺" nil :system t)
    (puthash "\\longmapsfrom" "⟻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longmapsfrom" "⟻" nil :system t)
    (puthash "\\longmapsto" "⟼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longmapsto" "⟼" nil :system t)
    (puthash "\\Longmapsfrom" "⟽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Longmapsfrom" "⟽" nil :system t)
    (puthash "\\Longmapsto" "⟾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Longmapsto" "⟾" nil :system t)
    (puthash "\\longrightsquigarrow" "⟿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longrightsquigarrow" "⟿" nil :system t)
    (puthash "\\nvtwoheadrightarrow" "⤀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvtwoheadrightarrow" "⤀" nil :system t)
    (puthash "\\nVtwoheadrightarrow" "⤁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVtwoheadrightarrow" "⤁" nil :system t)
    (puthash "\\nvLeftarrow" "⤂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvLeftarrow" "⤂" nil :system t)
    (puthash "\\nvRightarrow" "⤃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvRightarrow" "⤃" nil :system t)
    (puthash "\\nvLeftrightarrow" "⤄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvLeftrightarrow" "⤄" nil :system t)
    (puthash "\\twoheadmapsto" "⤅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadmapsto" "⤅" nil :system t)
    (puthash "\\Mapsfrom" "⤆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Mapsfrom" "⤆" nil :system t)
    (puthash "\\Mapsto" "⤇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Mapsto" "⤇" nil :system t)
    (puthash "\\downarrowbarred" "⤈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downarrowbarred" "⤈" nil :system t)
    (puthash "\\uparrowbarred" "⤉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\uparrowbarred" "⤉" nil :system t)
    (puthash "\\Uuparrow" "⤊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Uuparrow" "⤊" nil :system t)
    (puthash "\\Ddownarrow" "⤋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Ddownarrow" "⤋" nil :system t)
    (puthash "\\leftbkarrow" "⤌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftbkarrow" "⤌" nil :system t)
    (puthash "\\bkarow" "⤍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bkarow" "⤍" nil :system t)
    (puthash "\\leftdbkarrow" "⤎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftdbkarrow" "⤎" nil :system t)
    (puthash "\\dbkarow" "⤏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dbkarow" "⤏" nil :system t)
    (puthash "\\drbkarrow" "⤐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\drbkarrow" "⤐" nil :system t)
    (puthash "\\rightdotarrow" "⤑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightdotarrow" "⤑" nil :system t)
    (puthash "\\UpArrowBar" "⤒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\UpArrowBar" "⤒" nil :system t)
    (puthash "\\DownArrowBar" "⤓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownArrowBar" "⤓" nil :system t)
    (puthash "\\nvrightarrowtail" "⤔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvrightarrowtail" "⤔" nil :system t)
    (puthash "\\nVrightarrowtail" "⤕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVrightarrowtail" "⤕" nil :system t)
    (puthash "\\twoheadrightarrowtail" "⤖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadrightarrowtail" "⤖" nil :system t)
    (puthash "\\nvtwoheadrightarrowtail" "⤗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvtwoheadrightarrowtail" "⤗" nil :system t)
    (puthash "\\nVtwoheadrightarrowtail" "⤘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVtwoheadrightarrowtail" "⤘" nil :system t)
    (puthash "\\diamondleftarrow" "⤝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondleftarrow" "⤝" nil :system t)
    (puthash "\\rightarrowdiamond" "⤞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowdiamond" "⤞" nil :system t)
    (puthash "\\diamondleftarrowbar" "⤟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondleftarrowbar" "⤟" nil :system t)
    (puthash "\\barrightarrowdiamond" "⤠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barrightarrowdiamond" "⤠" nil :system t)
    (puthash "\\hksearow" "⤥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hksearow" "⤥" nil :system t)
    (puthash "\\hkswarow" "⤦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hkswarow" "⤦" nil :system t)
    (puthash "\\tona" "⤧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tona" "⤧" nil :system t)
    (puthash "\\toea" "⤨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\toea" "⤨" nil :system t)
    (puthash "\\tosa" "⤩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tosa" "⤩" nil :system t)
    (puthash "\\towa" "⤪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\towa" "⤪" nil :system t)
    (puthash "\\rdiagovfdiag" "⤫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rdiagovfdiag" "⤫" nil :system t)
    (puthash "\\fdiagovrdiag" "⤬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fdiagovrdiag" "⤬" nil :system t)
    (puthash "\\seovnearrow" "⤭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\seovnearrow" "⤭" nil :system t)
    (puthash "\\neovsearrow" "⤮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neovsearrow" "⤮" nil :system t)
    (puthash "\\fdiagovnearrow" "⤯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fdiagovnearrow" "⤯" nil :system t)
    (puthash "\\rdiagovsearrow" "⤰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rdiagovsearrow" "⤰" nil :system t)
    (puthash "\\neovnwarrow" "⤱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\neovnwarrow" "⤱" nil :system t)
    (puthash "\\nwovnearrow" "⤲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nwovnearrow" "⤲" nil :system t)
    (puthash "\\:arrow_heading_up:" "⤴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_heading_up:" "⤴" nil :system t)
    (puthash "\\:arrow_heading_down:" "⤵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_heading_down:" "⤵" nil :system t)
    (puthash "\\Rlarr" "⥂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Rlarr" "⥂" nil :system t)
    (puthash "\\rLarr" "⥄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rLarr" "⥄" nil :system t)
    (puthash "\\rightarrowplus" "⥅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowplus" "⥅" nil :system t)
    (puthash "\\leftarrowplus" "⥆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowplus" "⥆" nil :system t)
    (puthash "\\rarrx" "⥇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rarrx" "⥇" nil :system t)
    (puthash "\\leftrightarrowcircle" "⥈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightarrowcircle" "⥈" nil :system t)
    (puthash "\\twoheaduparrowcircle" "⥉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheaduparrowcircle" "⥉" nil :system t)
    (puthash "\\leftrightharpoonupdown" "⥊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightharpoonupdown" "⥊" nil :system t)
    (puthash "\\leftrightharpoondownup" "⥋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightharpoondownup" "⥋" nil :system t)
    (puthash "\\updownharpoonrightleft" "⥌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\updownharpoonrightleft" "⥌" nil :system t)
    (puthash "\\updownharpoonleftright" "⥍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\updownharpoonleftright" "⥍" nil :system t)
    (puthash "\\LeftRightVector" "⥎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftRightVector" "⥎" nil :system t)
    (puthash "\\RightUpDownVector" "⥏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightUpDownVector" "⥏" nil :system t)
    (puthash "\\DownLeftRightVector" "⥐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownLeftRightVector" "⥐" nil :system t)
    (puthash "\\LeftUpDownVector" "⥑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftUpDownVector" "⥑" nil :system t)
    (puthash "\\LeftVectorBar" "⥒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftVectorBar" "⥒" nil :system t)
    (puthash "\\RightVectorBar" "⥓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightVectorBar" "⥓" nil :system t)
    (puthash "\\RightUpVectorBar" "⥔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightUpVectorBar" "⥔" nil :system t)
    (puthash "\\RightDownVectorBar" "⥕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightDownVectorBar" "⥕" nil :system t)
    (puthash "\\DownLeftVectorBar" "⥖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownLeftVectorBar" "⥖" nil :system t)
    (puthash "\\DownRightVectorBar" "⥗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownRightVectorBar" "⥗" nil :system t)
    (puthash "\\LeftUpVectorBar" "⥘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftUpVectorBar" "⥘" nil :system t)
    (puthash "\\LeftDownVectorBar" "⥙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftDownVectorBar" "⥙" nil :system t)
    (puthash "\\LeftTeeVector" "⥚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftTeeVector" "⥚" nil :system t)
    (puthash "\\RightTeeVector" "⥛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightTeeVector" "⥛" nil :system t)
    (puthash "\\RightUpTeeVector" "⥜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightUpTeeVector" "⥜" nil :system t)
    (puthash "\\RightDownTeeVector" "⥝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightDownTeeVector" "⥝" nil :system t)
    (puthash "\\DownLeftTeeVector" "⥞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownLeftTeeVector" "⥞" nil :system t)
    (puthash "\\DownRightTeeVector" "⥟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\DownRightTeeVector" "⥟" nil :system t)
    (puthash "\\LeftUpTeeVector" "⥠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftUpTeeVector" "⥠" nil :system t)
    (puthash "\\LeftDownTeeVector" "⥡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftDownTeeVector" "⥡" nil :system t)
    (puthash "\\leftharpoonsupdown" "⥢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftharpoonsupdown" "⥢" nil :system t)
    (puthash "\\upharpoonsleftright" "⥣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upharpoonsleftright" "⥣" nil :system t)
    (puthash "\\rightharpoonsupdown" "⥤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightharpoonsupdown" "⥤" nil :system t)
    (puthash "\\downharpoonsleftright" "⥥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downharpoonsleftright" "⥥" nil :system t)
    (puthash "\\leftrightharpoonsup" "⥦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightharpoonsup" "⥦" nil :system t)
    (puthash "\\leftrightharpoonsdown" "⥧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftrightharpoonsdown" "⥧" nil :system t)
    (puthash "\\rightleftharpoonsup" "⥨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightleftharpoonsup" "⥨" nil :system t)
    (puthash "\\rightleftharpoonsdown" "⥩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightleftharpoonsdown" "⥩" nil :system t)
    (puthash "\\leftharpoonupdash" "⥪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftharpoonupdash" "⥪" nil :system t)
    (puthash "\\dashleftharpoondown" "⥫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dashleftharpoondown" "⥫" nil :system t)
    (puthash "\\rightharpoonupdash" "⥬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightharpoonupdash" "⥬" nil :system t)
    (puthash "\\dashrightharpoondown" "⥭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dashrightharpoondown" "⥭" nil :system t)
    (puthash "\\UpEquilibrium" "⥮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\UpEquilibrium" "⥮" nil :system t)
    (puthash "\\ReverseUpEquilibrium" "⥯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ReverseUpEquilibrium" "⥯" nil :system t)
    (puthash "\\RoundImplies" "⥰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RoundImplies" "⥰" nil :system t)
    (puthash "\\leftarrowless" "⥷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowless" "⥷" nil :system t)
    (puthash "\\leftarrowsubset" "⥺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowsubset" "⥺" nil :system t)
    (puthash "\\Vvert" "⦀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Vvert" "⦀" nil :system t)
    (puthash "\\Elroang" "⦆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Elroang" "⦆" nil :system t)
    (puthash "\\ddfnc" "⦙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ddfnc" "⦙" nil :system t)
    (puthash "\\measuredangleleft" "⦛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measuredangleleft" "⦛" nil :system t)
    (puthash "\\Angle" "⦜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Angle" "⦜" nil :system t)
    (puthash "\\rightanglemdot" "⦝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightanglemdot" "⦝" nil :system t)
    (puthash "\\angles" "⦞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\angles" "⦞" nil :system t)
    (puthash "\\angdnr" "⦟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\angdnr" "⦟" nil :system t)
    (puthash "\\lpargt" "⦠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lpargt" "⦠" nil :system t)
    (puthash "\\sphericalangleup" "⦡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sphericalangleup" "⦡" nil :system t)
    (puthash "\\turnangle" "⦢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\turnangle" "⦢" nil :system t)
    (puthash "\\revangle" "⦣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\revangle" "⦣" nil :system t)
    (puthash "\\angleubar" "⦤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\angleubar" "⦤" nil :system t)
    (puthash "\\revangleubar" "⦥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\revangleubar" "⦥" nil :system t)
    (puthash "\\wideangledown" "⦦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wideangledown" "⦦" nil :system t)
    (puthash "\\wideangleup" "⦧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wideangleup" "⦧" nil :system t)
    (puthash "\\measanglerutone" "⦨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measanglerutone" "⦨" nil :system t)
    (puthash "\\measanglelutonw" "⦩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measanglelutonw" "⦩" nil :system t)
    (puthash "\\measanglerdtose" "⦪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measanglerdtose" "⦪" nil :system t)
    (puthash "\\measangleldtosw" "⦫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measangleldtosw" "⦫" nil :system t)
    (puthash "\\measangleurtone" "⦬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measangleurtone" "⦬" nil :system t)
    (puthash "\\measangleultonw" "⦭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measangleultonw" "⦭" nil :system t)
    (puthash "\\measangledrtose" "⦮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measangledrtose" "⦮" nil :system t)
    (puthash "\\measangledltosw" "⦯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\measangledltosw" "⦯" nil :system t)
    (puthash "\\revemptyset" "⦰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\revemptyset" "⦰" nil :system t)
    (puthash "\\emptysetobar" "⦱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\emptysetobar" "⦱" nil :system t)
    (puthash "\\emptysetocirc" "⦲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\emptysetocirc" "⦲" nil :system t)
    (puthash "\\emptysetoarr" "⦳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\emptysetoarr" "⦳" nil :system t)
    (puthash "\\emptysetoarrl" "⦴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\emptysetoarrl" "⦴" nil :system t)
    (puthash "\\circledparallel" "⦷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledparallel" "⦷" nil :system t)
    (puthash "\\obslash" "⦸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\obslash" "⦸" nil :system t)
    (puthash "\\odotslashdot" "⦼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\odotslashdot" "⦼" nil :system t)
    (puthash "\\circledwhitebullet" "⦾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledwhitebullet" "⦾" nil :system t)
    (puthash "\\circledbullet" "⦿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circledbullet" "⦿" nil :system t)
    (puthash "\\olessthan" "⧀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\olessthan" "⧀" nil :system t)
    (puthash "\\ogreaterthan" "⧁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ogreaterthan" "⧁" nil :system t)
    (puthash "\\boxdiag" "⧄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxdiag" "⧄" nil :system t)
    (puthash "\\boxbslash" "⧅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxbslash" "⧅" nil :system t)
    (puthash "\\boxast" "⧆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxast" "⧆" nil :system t)
    (puthash "\\boxcircle" "⧇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\boxcircle" "⧇" nil :system t)
    (puthash "\\Lap" "⧊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Lap" "⧊" nil :system t)
    (puthash "\\defas" "⧋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\defas" "⧋" nil :system t)
    (puthash "\\LeftTriangleBar" "⧏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LeftTriangleBar" "⧏" nil :system t)
    (puthash "\\NotLeftTriangleBar" "⧏̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotLeftTriangleBar" "⧏̸" nil :system t)
    (puthash "\\RightTriangleBar" "⧐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RightTriangleBar" "⧐" nil :system t)
    (puthash "\\NotRightTriangleBar" "⧐̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotRightTriangleBar" "⧐̸" nil :system t)
    (puthash "\\dualmap" "⧟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dualmap" "⧟" nil :system t)
    (puthash "\\lrtriangleeq" "⧡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lrtriangleeq" "⧡" nil :system t)
    (puthash "\\shuffle" "⧢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\shuffle" "⧢" nil :system t)
    (puthash "\\eparsl" "⧣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eparsl" "⧣" nil :system t)
    (puthash "\\smeparsl" "⧤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smeparsl" "⧤" nil :system t)
    (puthash "\\eqvparsl" "⧥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqvparsl" "⧥" nil :system t)
    (puthash "\\blacklozenge" "⧫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blacklozenge" "⧫" nil :system t)
    (puthash "\\RuleDelayed" "⧴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RuleDelayed" "⧴" nil :system t)
    (puthash "\\dsol" "⧶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dsol" "⧶" nil :system t)
    (puthash "\\rsolbar" "⧷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rsolbar" "⧷" nil :system t)
    (puthash "\\doubleplus" "⧺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\doubleplus" "⧺" nil :system t)
    (puthash "\\tripleplus" "⧻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tripleplus" "⧻" nil :system t)
    (puthash "\\bigodot" "⨀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigodot" "⨀" nil :system t)
    (puthash "\\bigoplus" "⨁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigoplus" "⨁" nil :system t)
    (puthash "\\bigotimes" "⨂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigotimes" "⨂" nil :system t)
    (puthash "\\bigcupdot" "⨃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigcupdot" "⨃" nil :system t)
    (puthash "\\biguplus" "⨄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biguplus" "⨄" nil :system t)
    (puthash "\\bigsqcap" "⨅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigsqcap" "⨅" nil :system t)
    (puthash "\\bigsqcup" "⨆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigsqcup" "⨆" nil :system t)
    (puthash "\\conjquant" "⨇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\conjquant" "⨇" nil :system t)
    (puthash "\\disjquant" "⨈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\disjquant" "⨈" nil :system t)
    (puthash "\\bigtimes" "⨉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigtimes" "⨉" nil :system t)
    (puthash "\\modtwosum" "⨊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\modtwosum" "⨊" nil :system t)
    (puthash "\\sumint" "⨋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sumint" "⨋" nil :system t)
    (puthash "\\iiiint" "⨌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iiiint" "⨌" nil :system t)
    (puthash "\\intbar" "⨍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intbar" "⨍" nil :system t)
    (puthash "\\intBar" "⨎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intBar" "⨎" nil :system t)
    (puthash "\\clockoint" "⨏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\clockoint" "⨏" nil :system t)
    (puthash "\\cirfnint" "⨐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cirfnint" "⨐" nil :system t)
    (puthash "\\awint" "⨑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\awint" "⨑" nil :system t)
    (puthash "\\rppolint" "⨒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rppolint" "⨒" nil :system t)
    (puthash "\\scpolint" "⨓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scpolint" "⨓" nil :system t)
    (puthash "\\npolint" "⨔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\npolint" "⨔" nil :system t)
    (puthash "\\pointint" "⨕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pointint" "⨕" nil :system t)
    (puthash "\\sqrint" "⨖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sqrint" "⨖" nil :system t)
    (puthash "\\intx" "⨘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intx" "⨘" nil :system t)
    (puthash "\\intcap" "⨙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intcap" "⨙" nil :system t)
    (puthash "\\intcup" "⨚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intcup" "⨚" nil :system t)
    (puthash "\\upint" "⨛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upint" "⨛" nil :system t)
    (puthash "\\lowint" "⨜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lowint" "⨜" nil :system t)
    (puthash "\\join" "⨝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\join" "⨝" nil :system t)
    (puthash "\\bbsemi" "⨟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbsemi" "⨟" nil :system t)
    (puthash "\\ringplus" "⨢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ringplus" "⨢" nil :system t)
    (puthash "\\plushat" "⨣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\plushat" "⨣" nil :system t)
    (puthash "\\simplus" "⨤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simplus" "⨤" nil :system t)
    (puthash "\\plusdot" "⨥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\plusdot" "⨥" nil :system t)
    (puthash "\\plussim" "⨦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\plussim" "⨦" nil :system t)
    (puthash "\\plussubtwo" "⨧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\plussubtwo" "⨧" nil :system t)
    (puthash "\\plustrif" "⨨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\plustrif" "⨨" nil :system t)
    (puthash "\\commaminus" "⨩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\commaminus" "⨩" nil :system t)
    (puthash "\\minusdot" "⨪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\minusdot" "⨪" nil :system t)
    (puthash "\\minusfdots" "⨫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\minusfdots" "⨫" nil :system t)
    (puthash "\\minusrdots" "⨬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\minusrdots" "⨬" nil :system t)
    (puthash "\\opluslhrim" "⨭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\opluslhrim" "⨭" nil :system t)
    (puthash "\\oplusrhrim" "⨮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\oplusrhrim" "⨮" nil :system t)
    (puthash "\\Times" "⨯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Times" "⨯" nil :system t)
    (puthash "\\dottimes" "⨰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dottimes" "⨰" nil :system t)
    (puthash "\\timesbar" "⨱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\timesbar" "⨱" nil :system t)
    (puthash "\\btimes" "⨲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\btimes" "⨲" nil :system t)
    (puthash "\\smashtimes" "⨳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smashtimes" "⨳" nil :system t)
    (puthash "\\otimeslhrim" "⨴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\otimeslhrim" "⨴" nil :system t)
    (puthash "\\otimesrhrim" "⨵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\otimesrhrim" "⨵" nil :system t)
    (puthash "\\otimeshat" "⨶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\otimeshat" "⨶" nil :system t)
    (puthash "\\Otimes" "⨷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Otimes" "⨷" nil :system t)
    (puthash "\\odiv" "⨸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\odiv" "⨸" nil :system t)
    (puthash "\\triangleplus" "⨹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangleplus" "⨹" nil :system t)
    (puthash "\\triangleminus" "⨺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangleminus" "⨺" nil :system t)
    (puthash "\\triangletimes" "⨻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\triangletimes" "⨻" nil :system t)
    (puthash "\\intprod" "⨼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intprod" "⨼" nil :system t)
    (puthash "\\intprodr" "⨽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\intprodr" "⨽" nil :system t)
    (puthash "\\amalg" "⨿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\amalg" "⨿" nil :system t)
    (puthash "\\capdot" "⩀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\capdot" "⩀" nil :system t)
    (puthash "\\uminus" "⩁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\uminus" "⩁" nil :system t)
    (puthash "\\barcup" "⩂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barcup" "⩂" nil :system t)
    (puthash "\\barcap" "⩃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\barcap" "⩃" nil :system t)
    (puthash "\\capwedge" "⩄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\capwedge" "⩄" nil :system t)
    (puthash "\\cupvee" "⩅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\cupvee" "⩅" nil :system t)
    (puthash "\\twocups" "⩊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twocups" "⩊" nil :system t)
    (puthash "\\twocaps" "⩋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twocaps" "⩋" nil :system t)
    (puthash "\\closedvarcup" "⩌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\closedvarcup" "⩌" nil :system t)
    (puthash "\\closedvarcap" "⩍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\closedvarcap" "⩍" nil :system t)
    (puthash "\\Sqcap" "⩎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Sqcap" "⩎" nil :system t)
    (puthash "\\Sqcup" "⩏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Sqcup" "⩏" nil :system t)
    (puthash "\\closedvarcupsmashprod" "⩐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\closedvarcupsmashprod" "⩐" nil :system t)
    (puthash "\\wedgeodot" "⩑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedgeodot" "⩑" nil :system t)
    (puthash "\\veeodot" "⩒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\veeodot" "⩒" nil :system t)
    (puthash "\\And" "⩓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\And" "⩓" nil :system t)
    (puthash "\\Or" "⩔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Or" "⩔" nil :system t)
    (puthash "\\wedgeonwedge" "⩕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedgeonwedge" "⩕" nil :system t)
    (puthash "\\ElOr" "⩖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ElOr" "⩖" nil :system t)
    (puthash "\\bigslopedvee" "⩗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigslopedvee" "⩗" nil :system t)
    (puthash "\\bigslopedwedge" "⩘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigslopedwedge" "⩘" nil :system t)
    (puthash "\\wedgemidvert" "⩚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedgemidvert" "⩚" nil :system t)
    (puthash "\\veemidvert" "⩛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\veemidvert" "⩛" nil :system t)
    (puthash "\\midbarwedge" "⩜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\midbarwedge" "⩜" nil :system t)
    (puthash "\\midbarvee" "⩝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\midbarvee" "⩝" nil :system t)
    (puthash "\\perspcorrespond" "⩞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\perspcorrespond" "⩞" nil :system t)
    (puthash "\\minhat" "⩟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\minhat" "⩟" nil :system t)
    (puthash "\\wedgedoublebar" "⩠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\wedgedoublebar" "⩠" nil :system t)
    (puthash "\\varveebar" "⩡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varveebar" "⩡" nil :system t)
    (puthash "\\doublebarvee" "⩢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\doublebarvee" "⩢" nil :system t)
    (puthash "\\veedoublebar" "⩣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\veedoublebar" "⩣" nil :system t)
    (puthash "\\eqdot" "⩦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqdot" "⩦" nil :system t)
    (puthash "\\dotequiv" "⩧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dotequiv" "⩧" nil :system t)
    (puthash "\\dotsim" "⩪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dotsim" "⩪" nil :system t)
    (puthash "\\simrdots" "⩫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simrdots" "⩫" nil :system t)
    (puthash "\\simminussim" "⩬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simminussim" "⩬" nil :system t)
    (puthash "\\congdot" "⩭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\congdot" "⩭" nil :system t)
    (puthash "\\asteq" "⩮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\asteq" "⩮" nil :system t)
    (puthash "\\hatapprox" "⩯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hatapprox" "⩯" nil :system t)
    (puthash "\\approxeqq" "⩰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\approxeqq" "⩰" nil :system t)
    (puthash "\\eqqplus" "⩱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqqplus" "⩱" nil :system t)
    (puthash "\\pluseqq" "⩲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pluseqq" "⩲" nil :system t)
    (puthash "\\eqqsim" "⩳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqqsim" "⩳" nil :system t)
    (puthash "\\Coloneq" "⩴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Coloneq" "⩴" nil :system t)
    (puthash "\\Equal" "⩵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Equal" "⩵" nil :system t)
    (puthash "\\eqeqeq" "⩶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqeqeq" "⩶" nil :system t)
    (puthash "\\ddotseq" "⩷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ddotseq" "⩷" nil :system t)
    (puthash "\\equivDD" "⩸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\equivDD" "⩸" nil :system t)
    (puthash "\\ltcir" "⩹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltcir" "⩹" nil :system t)
    (puthash "\\gtcir" "⩺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtcir" "⩺" nil :system t)
    (puthash "\\ltquest" "⩻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltquest" "⩻" nil :system t)
    (puthash "\\gtquest" "⩼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtquest" "⩼" nil :system t)
    (puthash "\\leqslant" "⩽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leqslant" "⩽" nil :system t)
    (puthash "\\nleqslant" "⩽̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nleqslant" "⩽̸" nil :system t)
    (puthash "\\geqslant" "⩾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\geqslant" "⩾" nil :system t)
    (puthash "\\ngeqslant" "⩾̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ngeqslant" "⩾̸" nil :system t)
    (puthash "\\lesdot" "⩿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesdot" "⩿" nil :system t)
    (puthash "\\gesdot" "⪀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gesdot" "⪀" nil :system t)
    (puthash "\\lesdoto" "⪁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesdoto" "⪁" nil :system t)
    (puthash "\\gesdoto" "⪂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gesdoto" "⪂" nil :system t)
    (puthash "\\lesdotor" "⪃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesdotor" "⪃" nil :system t)
    (puthash "\\gesdotol" "⪄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gesdotol" "⪄" nil :system t)
    (puthash "\\lessapprox" "⪅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lessapprox" "⪅" nil :system t)
    (puthash "\\gtrapprox" "⪆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtrapprox" "⪆" nil :system t)
    (puthash "\\lneq" "⪇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lneq" "⪇" nil :system t)
    (puthash "\\gneq" "⪈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gneq" "⪈" nil :system t)
    (puthash "\\lnapprox" "⪉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lnapprox" "⪉" nil :system t)
    (puthash "\\gnapprox" "⪊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gnapprox" "⪊" nil :system t)
    (puthash "\\lesseqqgtr" "⪋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesseqqgtr" "⪋" nil :system t)
    (puthash "\\gtreqqless" "⪌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtreqqless" "⪌" nil :system t)
    (puthash "\\lsime" "⪍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lsime" "⪍" nil :system t)
    (puthash "\\gsime" "⪎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gsime" "⪎" nil :system t)
    (puthash "\\lsimg" "⪏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lsimg" "⪏" nil :system t)
    (puthash "\\gsiml" "⪐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gsiml" "⪐" nil :system t)
    (puthash "\\lgE" "⪑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lgE" "⪑" nil :system t)
    (puthash "\\glE" "⪒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\glE" "⪒" nil :system t)
    (puthash "\\lesges" "⪓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lesges" "⪓" nil :system t)
    (puthash "\\gesles" "⪔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gesles" "⪔" nil :system t)
    (puthash "\\eqslantless" "⪕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqslantless" "⪕" nil :system t)
    (puthash "\\eqslantgtr" "⪖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqslantgtr" "⪖" nil :system t)
    (puthash "\\elsdot" "⪗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\elsdot" "⪗" nil :system t)
    (puthash "\\egsdot" "⪘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\egsdot" "⪘" nil :system t)
    (puthash "\\eqqless" "⪙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqqless" "⪙" nil :system t)
    (puthash "\\eqqgtr" "⪚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqqgtr" "⪚" nil :system t)
    (puthash "\\eqqslantless" "⪛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqqslantless" "⪛" nil :system t)
    (puthash "\\eqqslantgtr" "⪜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\eqqslantgtr" "⪜" nil :system t)
    (puthash "\\simless" "⪝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simless" "⪝" nil :system t)
    (puthash "\\simgtr" "⪞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simgtr" "⪞" nil :system t)
    (puthash "\\simlE" "⪟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simlE" "⪟" nil :system t)
    (puthash "\\simgE" "⪠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\simgE" "⪠" nil :system t)
    (puthash "\\NestedLessLess" "⪡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NestedLessLess" "⪡" nil :system t)
    (puthash "\\NotNestedLessLess" "⪡̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotNestedLessLess" "⪡̸" nil :system t)
    (puthash "\\NestedGreaterGreater" "⪢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NestedGreaterGreater" "⪢" nil :system t)
    (puthash "\\NotNestedGreaterGreater" "⪢̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\NotNestedGreaterGreater" "⪢̸" nil :system t)
    (puthash "\\partialmeetcontraction" "⪣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\partialmeetcontraction" "⪣" nil :system t)
    (puthash "\\glj" "⪤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\glj" "⪤" nil :system t)
    (puthash "\\gla" "⪥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gla" "⪥" nil :system t)
    (puthash "\\ltcc" "⪦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ltcc" "⪦" nil :system t)
    (puthash "\\gtcc" "⪧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gtcc" "⪧" nil :system t)
    (puthash "\\lescc" "⪨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lescc" "⪨" nil :system t)
    (puthash "\\gescc" "⪩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gescc" "⪩" nil :system t)
    (puthash "\\smt" "⪪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smt" "⪪" nil :system t)
    (puthash "\\lat" "⪫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lat" "⪫" nil :system t)
    (puthash "\\smte" "⪬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smte" "⪬" nil :system t)
    (puthash "\\late" "⪭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\late" "⪭" nil :system t)
    (puthash "\\bumpeqq" "⪮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bumpeqq" "⪮" nil :system t)
    (puthash "\\preceq" "⪯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\preceq" "⪯" nil :system t)
    (puthash "\\npreceq" "⪯̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\npreceq" "⪯̸" nil :system t)
    (puthash "\\succeq" "⪰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succeq" "⪰" nil :system t)
    (puthash "\\nsucceq" "⪰̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsucceq" "⪰̸" nil :system t)
    (puthash "\\precneq" "⪱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\precneq" "⪱" nil :system t)
    (puthash "\\succneq" "⪲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succneq" "⪲" nil :system t)
    (puthash "\\preceqq" "⪳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\preceqq" "⪳" nil :system t)
    (puthash "\\succeqq" "⪴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succeqq" "⪴" nil :system t)
    (puthash "\\precneqq" "⪵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\precneqq" "⪵" nil :system t)
    (puthash "\\succneqq" "⪶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succneqq" "⪶" nil :system t)
    (puthash "\\precapprox" "⪷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\precapprox" "⪷" nil :system t)
    (puthash "\\succapprox" "⪸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succapprox" "⪸" nil :system t)
    (puthash "\\precnapprox" "⪹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\precnapprox" "⪹" nil :system t)
    (puthash "\\succnapprox" "⪺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\succnapprox" "⪺" nil :system t)
    (puthash "\\Prec" "⪻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Prec" "⪻" nil :system t)
    (puthash "\\Succ" "⪼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Succ" "⪼" nil :system t)
    (puthash "\\subsetdot" "⪽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsetdot" "⪽" nil :system t)
    (puthash "\\supsetdot" "⪾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsetdot" "⪾" nil :system t)
    (puthash "\\subsetplus" "⪿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsetplus" "⪿" nil :system t)
    (puthash "\\supsetplus" "⫀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsetplus" "⫀" nil :system t)
    (puthash "\\submult" "⫁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\submult" "⫁" nil :system t)
    (puthash "\\supmult" "⫂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supmult" "⫂" nil :system t)
    (puthash "\\subedot" "⫃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subedot" "⫃" nil :system t)
    (puthash "\\supedot" "⫄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supedot" "⫄" nil :system t)
    (puthash "\\subseteqq" "⫅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subseteqq" "⫅" nil :system t)
    (puthash "\\nsubseteqq" "⫅̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsubseteqq" "⫅̸" nil :system t)
    (puthash "\\supseteqq" "⫆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supseteqq" "⫆" nil :system t)
    (puthash "\\nsupseteqq" "⫆̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nsupseteqq" "⫆̸" nil :system t)
    (puthash "\\subsim" "⫇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsim" "⫇" nil :system t)
    (puthash "\\supsim" "⫈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsim" "⫈" nil :system t)
    (puthash "\\subsetapprox" "⫉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsetapprox" "⫉" nil :system t)
    (puthash "\\supsetapprox" "⫊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsetapprox" "⫊" nil :system t)
    (puthash "\\subsetneqq" "⫋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsetneqq" "⫋" nil :system t)
    (puthash "\\supsetneqq" "⫌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsetneqq" "⫌" nil :system t)
    (puthash "\\lsqhook" "⫍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lsqhook" "⫍" nil :system t)
    (puthash "\\rsqhook" "⫎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rsqhook" "⫎" nil :system t)
    (puthash "\\csub" "⫏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\csub" "⫏" nil :system t)
    (puthash "\\csup" "⫐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\csup" "⫐" nil :system t)
    (puthash "\\csube" "⫑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\csube" "⫑" nil :system t)
    (puthash "\\csupe" "⫒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\csupe" "⫒" nil :system t)
    (puthash "\\subsup" "⫓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsup" "⫓" nil :system t)
    (puthash "\\supsub" "⫔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsub" "⫔" nil :system t)
    (puthash "\\subsub" "⫕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\subsub" "⫕" nil :system t)
    (puthash "\\supsup" "⫖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supsup" "⫖" nil :system t)
    (puthash "\\suphsub" "⫗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\suphsub" "⫗" nil :system t)
    (puthash "\\supdsub" "⫘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\supdsub" "⫘" nil :system t)
    (puthash "\\forkv" "⫙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\forkv" "⫙" nil :system t)
    (puthash "\\mlcp" "⫛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mlcp" "⫛" nil :system t)
    (puthash "\\forks" "⫝̸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\forks" "⫝̸" nil :system t)
    (puthash "\\forksnot" "⫝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\forksnot" "⫝" nil :system t)
    (puthash "\\dashV" "⫣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dashV" "⫣" nil :system t)
    (puthash "\\Dashv" "⫤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Dashv" "⫤" nil :system t)
    (puthash "\\downvDash" "⫪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\downvDash" "⫪" nil :system t)
    (puthash "\\Top" "⫪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Top" "⫪" nil :system t)
    (puthash "\\upvDash" "⫫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\upvDash" "⫫" nil :system t)
    (puthash "\\Bot" "⫫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\Bot" "⫫" nil :system t)
    (puthash "\\indep" "⫫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\indep" "⫫" nil :system t)
    (puthash "\\interleave" "⫴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\interleave" "⫴" nil :system t)
    (puthash "\\tdcol" "⫶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tdcol" "⫶" nil :system t)
    (puthash "\\lllnest" "⫷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lllnest" "⫷" nil :system t)
    (puthash "\\gggnest" "⫸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\gggnest" "⫸" nil :system t)
    (puthash "\\leqqslant" "⫹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leqqslant" "⫹" nil :system t)
    (puthash "\\geqqslant" "⫺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\geqqslant" "⫺" nil :system t)
    (puthash "\\:arrow_left:" "⬅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_left:" "⬅" nil :system t)
    (puthash "\\:arrow_up:" "⬆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_up:" "⬆" nil :system t)
    (puthash "\\:arrow_down:" "⬇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_down:" "⬇" nil :system t)
    (puthash "\\squaretopblack" "⬒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squaretopblack" "⬒" nil :system t)
    (puthash "\\squarebotblack" "⬓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarebotblack" "⬓" nil :system t)
    (puthash "\\squareurblack" "⬔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squareurblack" "⬔" nil :system t)
    (puthash "\\squarellblack" "⬕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\squarellblack" "⬕" nil :system t)
    (puthash "\\diamondleftblack" "⬖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondleftblack" "⬖" nil :system t)
    (puthash "\\diamondrightblack" "⬗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondrightblack" "⬗" nil :system t)
    (puthash "\\diamondtopblack" "⬘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondtopblack" "⬘" nil :system t)
    (puthash "\\diamondbotblack" "⬙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\diamondbotblack" "⬙" nil :system t)
    (puthash "\\dottedsquare" "⬚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\dottedsquare" "⬚" nil :system t)
    (puthash "\\:black_large_square:" "⬛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_large_square:" "⬛" nil :system t)
    (puthash "\\lgblksquare" "⬛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lgblksquare" "⬛" nil :system t)
    (puthash "\\lgwhtsquare" "⬜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lgwhtsquare" "⬜" nil :system t)
    (puthash "\\:white_large_square:" "⬜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_large_square:" "⬜" nil :system t)
    (puthash "\\vysmblksquare" "⬝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vysmblksquare" "⬝" nil :system t)
    (puthash "\\vysmwhtsquare" "⬞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\vysmwhtsquare" "⬞" nil :system t)
    (puthash "\\pentagonblack" "⬟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pentagonblack" "⬟" nil :system t)
    (puthash "\\pentagon" "⬠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\pentagon" "⬠" nil :system t)
    (puthash "\\varhexagon" "⬡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varhexagon" "⬡" nil :system t)
    (puthash "\\varhexagonblack" "⬢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\varhexagonblack" "⬢" nil :system t)
    (puthash "\\hexagonblack" "⬣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\hexagonblack" "⬣" nil :system t)
    (puthash "\\lgblkcircle" "⬤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\lgblkcircle" "⬤" nil :system t)
    (puthash "\\mdblkdiamond" "⬥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdblkdiamond" "⬥" nil :system t)
    (puthash "\\mdwhtdiamond" "⬦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdwhtdiamond" "⬦" nil :system t)
    (puthash "\\mdblklozenge" "⬧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdblklozenge" "⬧" nil :system t)
    (puthash "\\mdwhtlozenge" "⬨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\mdwhtlozenge" "⬨" nil :system t)
    (puthash "\\smblkdiamond" "⬩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smblkdiamond" "⬩" nil :system t)
    (puthash "\\smblklozenge" "⬪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smblklozenge" "⬪" nil :system t)
    (puthash "\\smwhtlozenge" "⬫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smwhtlozenge" "⬫" nil :system t)
    (puthash "\\blkhorzoval" "⬬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blkhorzoval" "⬬" nil :system t)
    (puthash "\\whthorzoval" "⬭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\whthorzoval" "⬭" nil :system t)
    (puthash "\\blkvertoval" "⬮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\blkvertoval" "⬮" nil :system t)
    (puthash "\\whtvertoval" "⬯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\whtvertoval" "⬯" nil :system t)
    (puthash "\\circleonleftarrow" "⬰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\circleonleftarrow" "⬰" nil :system t)
    (puthash "\\leftthreearrows" "⬱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftthreearrows" "⬱" nil :system t)
    (puthash "\\leftarrowonoplus" "⬲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowonoplus" "⬲" nil :system t)
    (puthash "\\longleftsquigarrow" "⬳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\longleftsquigarrow" "⬳" nil :system t)
    (puthash "\\nvtwoheadleftarrow" "⬴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvtwoheadleftarrow" "⬴" nil :system t)
    (puthash "\\nVtwoheadleftarrow" "⬵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVtwoheadleftarrow" "⬵" nil :system t)
    (puthash "\\twoheadmapsfrom" "⬶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadmapsfrom" "⬶" nil :system t)
    (puthash "\\twoheadleftdbkarrow" "⬷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadleftdbkarrow" "⬷" nil :system t)
    (puthash "\\leftdotarrow" "⬸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftdotarrow" "⬸" nil :system t)
    (puthash "\\nvleftarrowtail" "⬹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvleftarrowtail" "⬹" nil :system t)
    (puthash "\\nVleftarrowtail" "⬺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVleftarrowtail" "⬺" nil :system t)
    (puthash "\\twoheadleftarrowtail" "⬻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\twoheadleftarrowtail" "⬻" nil :system t)
    (puthash "\\nvtwoheadleftarrowtail" "⬼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nvtwoheadleftarrowtail" "⬼" nil :system t)
    (puthash "\\nVtwoheadleftarrowtail" "⬽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\nVtwoheadleftarrowtail" "⬽" nil :system t)
    (puthash "\\leftarrowx" "⬾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowx" "⬾" nil :system t)
    (puthash "\\leftcurvedarrow" "⬿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftcurvedarrow" "⬿" nil :system t)
    (puthash "\\equalleftarrow" "⭀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\equalleftarrow" "⭀" nil :system t)
    (puthash "\\bsimilarleftarrow" "⭁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsimilarleftarrow" "⭁" nil :system t)
    (puthash "\\leftarrowbackapprox" "⭂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowbackapprox" "⭂" nil :system t)
    (puthash "\\rightarrowgtr" "⭃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowgtr" "⭃" nil :system t)
    (puthash "\\rightarrowsupset" "⭄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowsupset" "⭄" nil :system t)
    (puthash "\\LLeftarrow" "⭅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\LLeftarrow" "⭅" nil :system t)
    (puthash "\\RRightarrow" "⭆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\RRightarrow" "⭆" nil :system t)
    (puthash "\\bsimilarrightarrow" "⭇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsimilarrightarrow" "⭇" nil :system t)
    (puthash "\\rightarrowbackapprox" "⭈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowbackapprox" "⭈" nil :system t)
    (puthash "\\similarleftarrow" "⭉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\similarleftarrow" "⭉" nil :system t)
    (puthash "\\leftarrowapprox" "⭊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowapprox" "⭊" nil :system t)
    (puthash "\\leftarrowbsimilar" "⭋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\leftarrowbsimilar" "⭋" nil :system t)
    (puthash "\\rightarrowbsimilar" "⭌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightarrowbsimilar" "⭌" nil :system t)
    (puthash "\\:star:" "⭐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:star:" "⭐" nil :system t)
    (puthash "\\medwhitestar" "⭐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\medwhitestar" "⭐" nil :system t)
    (puthash "\\medblackstar" "⭑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\medblackstar" "⭑" nil :system t)
    (puthash "\\smwhitestar" "⭒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\smwhitestar" "⭒" nil :system t)
    (puthash "\\rightpentagonblack" "⭓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightpentagonblack" "⭓" nil :system t)
    (puthash "\\rightpentagon" "⭔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\rightpentagon" "⭔" nil :system t)
    (puthash "\\:o:" "⭕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:o:" "⭕" nil :system t)
    (puthash "\\_j" "ⱼ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\_j" "ⱼ" nil :system t)
    (puthash "\\^V" "ⱽ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^V" "ⱽ" nil :system t)
    (puthash "\\postalmark" "〒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\postalmark" "〒" nil :system t)
    (puthash "\\:wavy_dash:" "〰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wavy_dash:" "〰" nil :system t)
    (puthash "\\:part_alternation_mark:" "〽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:part_alternation_mark:" "〽" nil :system t)
    (puthash "\\:congratulations:" "㊗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:congratulations:" "㊗" nil :system t)
    (puthash "\\:secret:" "㊙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:secret:" "㊙" nil :system t)
    (puthash "\\^uparrow" "ꜛ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^uparrow" "ꜛ" nil :system t)
    (puthash "\\^downarrow" "ꜜ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^downarrow" "ꜜ" nil :system t)
    (puthash "\\^!" "ꜝ" table)
    (define-abbrev julia-latexsub-abbrev-table "\\^!" "ꜝ" nil :system t)
    (puthash "\\bfA" "𝐀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfA" "𝐀" nil :system t)
    (puthash "\\bfB" "𝐁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfB" "𝐁" nil :system t)
    (puthash "\\bfC" "𝐂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfC" "𝐂" nil :system t)
    (puthash "\\bfD" "𝐃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfD" "𝐃" nil :system t)
    (puthash "\\bfE" "𝐄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfE" "𝐄" nil :system t)
    (puthash "\\bfF" "𝐅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfF" "𝐅" nil :system t)
    (puthash "\\bfG" "𝐆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfG" "𝐆" nil :system t)
    (puthash "\\bfH" "𝐇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfH" "𝐇" nil :system t)
    (puthash "\\bfI" "𝐈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfI" "𝐈" nil :system t)
    (puthash "\\bfJ" "𝐉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfJ" "𝐉" nil :system t)
    (puthash "\\bfK" "𝐊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfK" "𝐊" nil :system t)
    (puthash "\\bfL" "𝐋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfL" "𝐋" nil :system t)
    (puthash "\\bfM" "𝐌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfM" "𝐌" nil :system t)
    (puthash "\\bfN" "𝐍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfN" "𝐍" nil :system t)
    (puthash "\\bfO" "𝐎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfO" "𝐎" nil :system t)
    (puthash "\\bfP" "𝐏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfP" "𝐏" nil :system t)
    (puthash "\\bfQ" "𝐐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfQ" "𝐐" nil :system t)
    (puthash "\\bfR" "𝐑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfR" "𝐑" nil :system t)
    (puthash "\\bfS" "𝐒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfS" "𝐒" nil :system t)
    (puthash "\\bfT" "𝐓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfT" "𝐓" nil :system t)
    (puthash "\\bfU" "𝐔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfU" "𝐔" nil :system t)
    (puthash "\\bfV" "𝐕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfV" "𝐕" nil :system t)
    (puthash "\\bfW" "𝐖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfW" "𝐖" nil :system t)
    (puthash "\\bfX" "𝐗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfX" "𝐗" nil :system t)
    (puthash "\\bfY" "𝐘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfY" "𝐘" nil :system t)
    (puthash "\\bfZ" "𝐙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfZ" "𝐙" nil :system t)
    (puthash "\\bfa" "𝐚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfa" "𝐚" nil :system t)
    (puthash "\\bfb" "𝐛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfb" "𝐛" nil :system t)
    (puthash "\\bfc" "𝐜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfc" "𝐜" nil :system t)
    (puthash "\\bfd" "𝐝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfd" "𝐝" nil :system t)
    (puthash "\\bfe" "𝐞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfe" "𝐞" nil :system t)
    (puthash "\\bff" "𝐟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bff" "𝐟" nil :system t)
    (puthash "\\bfg" "𝐠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfg" "𝐠" nil :system t)
    (puthash "\\bfh" "𝐡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfh" "𝐡" nil :system t)
    (puthash "\\bfi" "𝐢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfi" "𝐢" nil :system t)
    (puthash "\\bfj" "𝐣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfj" "𝐣" nil :system t)
    (puthash "\\bfk" "𝐤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfk" "𝐤" nil :system t)
    (puthash "\\bfl" "𝐥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfl" "𝐥" nil :system t)
    (puthash "\\bfm" "𝐦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfm" "𝐦" nil :system t)
    (puthash "\\bfn" "𝐧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfn" "𝐧" nil :system t)
    (puthash "\\bfo" "𝐨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfo" "𝐨" nil :system t)
    (puthash "\\bfp" "𝐩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfp" "𝐩" nil :system t)
    (puthash "\\bfq" "𝐪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfq" "𝐪" nil :system t)
    (puthash "\\bfr" "𝐫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfr" "𝐫" nil :system t)
    (puthash "\\bfs" "𝐬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfs" "𝐬" nil :system t)
    (puthash "\\bft" "𝐭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bft" "𝐭" nil :system t)
    (puthash "\\bfu" "𝐮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfu" "𝐮" nil :system t)
    (puthash "\\bfv" "𝐯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfv" "𝐯" nil :system t)
    (puthash "\\bfw" "𝐰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfw" "𝐰" nil :system t)
    (puthash "\\bfx" "𝐱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfx" "𝐱" nil :system t)
    (puthash "\\bfy" "𝐲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfy" "𝐲" nil :system t)
    (puthash "\\bfz" "𝐳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfz" "𝐳" nil :system t)
    (puthash "\\itA" "𝐴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itA" "𝐴" nil :system t)
    (puthash "\\itB" "𝐵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itB" "𝐵" nil :system t)
    (puthash "\\itC" "𝐶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itC" "𝐶" nil :system t)
    (puthash "\\itD" "𝐷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itD" "𝐷" nil :system t)
    (puthash "\\itE" "𝐸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itE" "𝐸" nil :system t)
    (puthash "\\itF" "𝐹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itF" "𝐹" nil :system t)
    (puthash "\\itG" "𝐺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itG" "𝐺" nil :system t)
    (puthash "\\itH" "𝐻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itH" "𝐻" nil :system t)
    (puthash "\\itI" "𝐼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itI" "𝐼" nil :system t)
    (puthash "\\itJ" "𝐽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itJ" "𝐽" nil :system t)
    (puthash "\\itK" "𝐾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itK" "𝐾" nil :system t)
    (puthash "\\itL" "𝐿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itL" "𝐿" nil :system t)
    (puthash "\\itM" "𝑀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itM" "𝑀" nil :system t)
    (puthash "\\itN" "𝑁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itN" "𝑁" nil :system t)
    (puthash "\\itO" "𝑂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itO" "𝑂" nil :system t)
    (puthash "\\itP" "𝑃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itP" "𝑃" nil :system t)
    (puthash "\\itQ" "𝑄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itQ" "𝑄" nil :system t)
    (puthash "\\itR" "𝑅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itR" "𝑅" nil :system t)
    (puthash "\\itS" "𝑆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itS" "𝑆" nil :system t)
    (puthash "\\itT" "𝑇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itT" "𝑇" nil :system t)
    (puthash "\\itU" "𝑈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itU" "𝑈" nil :system t)
    (puthash "\\itV" "𝑉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itV" "𝑉" nil :system t)
    (puthash "\\itW" "𝑊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itW" "𝑊" nil :system t)
    (puthash "\\itX" "𝑋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itX" "𝑋" nil :system t)
    (puthash "\\itY" "𝑌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itY" "𝑌" nil :system t)
    (puthash "\\itZ" "𝑍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itZ" "𝑍" nil :system t)
    (puthash "\\ita" "𝑎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ita" "𝑎" nil :system t)
    (puthash "\\itb" "𝑏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itb" "𝑏" nil :system t)
    (puthash "\\itc" "𝑐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itc" "𝑐" nil :system t)
    (puthash "\\itd" "𝑑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itd" "𝑑" nil :system t)
    (puthash "\\ite" "𝑒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ite" "𝑒" nil :system t)
    (puthash "\\itf" "𝑓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itf" "𝑓" nil :system t)
    (puthash "\\itg" "𝑔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itg" "𝑔" nil :system t)
    (puthash "\\iti" "𝑖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iti" "𝑖" nil :system t)
    (puthash "\\itj" "𝑗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itj" "𝑗" nil :system t)
    (puthash "\\itk" "𝑘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itk" "𝑘" nil :system t)
    (puthash "\\itl" "𝑙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itl" "𝑙" nil :system t)
    (puthash "\\itm" "𝑚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itm" "𝑚" nil :system t)
    (puthash "\\itn" "𝑛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itn" "𝑛" nil :system t)
    (puthash "\\ito" "𝑜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ito" "𝑜" nil :system t)
    (puthash "\\itp" "𝑝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itp" "𝑝" nil :system t)
    (puthash "\\itq" "𝑞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itq" "𝑞" nil :system t)
    (puthash "\\itr" "𝑟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itr" "𝑟" nil :system t)
    (puthash "\\its" "𝑠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\its" "𝑠" nil :system t)
    (puthash "\\itt" "𝑡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itt" "𝑡" nil :system t)
    (puthash "\\itu" "𝑢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itu" "𝑢" nil :system t)
    (puthash "\\itv" "𝑣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itv" "𝑣" nil :system t)
    (puthash "\\itw" "𝑤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itw" "𝑤" nil :system t)
    (puthash "\\itx" "𝑥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itx" "𝑥" nil :system t)
    (puthash "\\ity" "𝑦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ity" "𝑦" nil :system t)
    (puthash "\\itz" "𝑧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itz" "𝑧" nil :system t)
    (puthash "\\biA" "𝑨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biA" "𝑨" nil :system t)
    (puthash "\\biB" "𝑩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biB" "𝑩" nil :system t)
    (puthash "\\biC" "𝑪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biC" "𝑪" nil :system t)
    (puthash "\\biD" "𝑫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biD" "𝑫" nil :system t)
    (puthash "\\biE" "𝑬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biE" "𝑬" nil :system t)
    (puthash "\\biF" "𝑭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biF" "𝑭" nil :system t)
    (puthash "\\biG" "𝑮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biG" "𝑮" nil :system t)
    (puthash "\\biH" "𝑯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biH" "𝑯" nil :system t)
    (puthash "\\biI" "𝑰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biI" "𝑰" nil :system t)
    (puthash "\\biJ" "𝑱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biJ" "𝑱" nil :system t)
    (puthash "\\biK" "𝑲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biK" "𝑲" nil :system t)
    (puthash "\\biL" "𝑳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biL" "𝑳" nil :system t)
    (puthash "\\biM" "𝑴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biM" "𝑴" nil :system t)
    (puthash "\\biN" "𝑵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biN" "𝑵" nil :system t)
    (puthash "\\biO" "𝑶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biO" "𝑶" nil :system t)
    (puthash "\\biP" "𝑷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biP" "𝑷" nil :system t)
    (puthash "\\biQ" "𝑸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biQ" "𝑸" nil :system t)
    (puthash "\\biR" "𝑹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biR" "𝑹" nil :system t)
    (puthash "\\biS" "𝑺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biS" "𝑺" nil :system t)
    (puthash "\\biT" "𝑻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biT" "𝑻" nil :system t)
    (puthash "\\biU" "𝑼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biU" "𝑼" nil :system t)
    (puthash "\\biV" "𝑽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biV" "𝑽" nil :system t)
    (puthash "\\biW" "𝑾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biW" "𝑾" nil :system t)
    (puthash "\\biX" "𝑿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biX" "𝑿" nil :system t)
    (puthash "\\biY" "𝒀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biY" "𝒀" nil :system t)
    (puthash "\\biZ" "𝒁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biZ" "𝒁" nil :system t)
    (puthash "\\bia" "𝒂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bia" "𝒂" nil :system t)
    (puthash "\\bib" "𝒃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bib" "𝒃" nil :system t)
    (puthash "\\bic" "𝒄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bic" "𝒄" nil :system t)
    (puthash "\\bid" "𝒅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bid" "𝒅" nil :system t)
    (puthash "\\bie" "𝒆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bie" "𝒆" nil :system t)
    (puthash "\\bif" "𝒇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bif" "𝒇" nil :system t)
    (puthash "\\big" "𝒈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\big" "𝒈" nil :system t)
    (puthash "\\bih" "𝒉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bih" "𝒉" nil :system t)
    (puthash "\\bii" "𝒊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bii" "𝒊" nil :system t)
    (puthash "\\bij" "𝒋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bij" "𝒋" nil :system t)
    (puthash "\\bik" "𝒌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bik" "𝒌" nil :system t)
    (puthash "\\bil" "𝒍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bil" "𝒍" nil :system t)
    (puthash "\\bim" "𝒎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bim" "𝒎" nil :system t)
    (puthash "\\bin" "𝒏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bin" "𝒏" nil :system t)
    (puthash "\\bio" "𝒐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bio" "𝒐" nil :system t)
    (puthash "\\bip" "𝒑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bip" "𝒑" nil :system t)
    (puthash "\\biq" "𝒒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biq" "𝒒" nil :system t)
    (puthash "\\bir" "𝒓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bir" "𝒓" nil :system t)
    (puthash "\\bis" "𝒔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bis" "𝒔" nil :system t)
    (puthash "\\bit" "𝒕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bit" "𝒕" nil :system t)
    (puthash "\\biu" "𝒖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biu" "𝒖" nil :system t)
    (puthash "\\biv" "𝒗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biv" "𝒗" nil :system t)
    (puthash "\\biw" "𝒘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biw" "𝒘" nil :system t)
    (puthash "\\bix" "𝒙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bix" "𝒙" nil :system t)
    (puthash "\\biy" "𝒚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biy" "𝒚" nil :system t)
    (puthash "\\biz" "𝒛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biz" "𝒛" nil :system t)
    (puthash "\\scrA" "𝒜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrA" "𝒜" nil :system t)
    (puthash "\\scrC" "𝒞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrC" "𝒞" nil :system t)
    (puthash "\\scrD" "𝒟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrD" "𝒟" nil :system t)
    (puthash "\\scrG" "𝒢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrG" "𝒢" nil :system t)
    (puthash "\\scrJ" "𝒥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrJ" "𝒥" nil :system t)
    (puthash "\\scrK" "𝒦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrK" "𝒦" nil :system t)
    (puthash "\\scrN" "𝒩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrN" "𝒩" nil :system t)
    (puthash "\\scrO" "𝒪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrO" "𝒪" nil :system t)
    (puthash "\\scrP" "𝒫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrP" "𝒫" nil :system t)
    (puthash "\\scrQ" "𝒬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrQ" "𝒬" nil :system t)
    (puthash "\\scrS" "𝒮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrS" "𝒮" nil :system t)
    (puthash "\\scrT" "𝒯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrT" "𝒯" nil :system t)
    (puthash "\\scrU" "𝒰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrU" "𝒰" nil :system t)
    (puthash "\\scrV" "𝒱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrV" "𝒱" nil :system t)
    (puthash "\\scrW" "𝒲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrW" "𝒲" nil :system t)
    (puthash "\\scrX" "𝒳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrX" "𝒳" nil :system t)
    (puthash "\\scrY" "𝒴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrY" "𝒴" nil :system t)
    (puthash "\\scrZ" "𝒵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrZ" "𝒵" nil :system t)
    (puthash "\\scra" "𝒶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scra" "𝒶" nil :system t)
    (puthash "\\scrb" "𝒷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrb" "𝒷" nil :system t)
    (puthash "\\scrc" "𝒸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrc" "𝒸" nil :system t)
    (puthash "\\scrd" "𝒹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrd" "𝒹" nil :system t)
    (puthash "\\scrf" "𝒻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrf" "𝒻" nil :system t)
    (puthash "\\scrh" "𝒽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrh" "𝒽" nil :system t)
    (puthash "\\scri" "𝒾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scri" "𝒾" nil :system t)
    (puthash "\\scrj" "𝒿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrj" "𝒿" nil :system t)
    (puthash "\\scrk" "𝓀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrk" "𝓀" nil :system t)
    (puthash "\\scrl" "𝓁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrl" "𝓁" nil :system t)
    (puthash "\\scrm" "𝓂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrm" "𝓂" nil :system t)
    (puthash "\\scrn" "𝓃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrn" "𝓃" nil :system t)
    (puthash "\\scrp" "𝓅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrp" "𝓅" nil :system t)
    (puthash "\\scrq" "𝓆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrq" "𝓆" nil :system t)
    (puthash "\\scrr" "𝓇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrr" "𝓇" nil :system t)
    (puthash "\\scrs" "𝓈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrs" "𝓈" nil :system t)
    (puthash "\\scrt" "𝓉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrt" "𝓉" nil :system t)
    (puthash "\\scru" "𝓊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scru" "𝓊" nil :system t)
    (puthash "\\scrv" "𝓋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrv" "𝓋" nil :system t)
    (puthash "\\scrw" "𝓌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrw" "𝓌" nil :system t)
    (puthash "\\scrx" "𝓍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrx" "𝓍" nil :system t)
    (puthash "\\scry" "𝓎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scry" "𝓎" nil :system t)
    (puthash "\\scrz" "𝓏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\scrz" "𝓏" nil :system t)
    (puthash "\\bscrA" "𝓐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrA" "𝓐" nil :system t)
    (puthash "\\bscrB" "𝓑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrB" "𝓑" nil :system t)
    (puthash "\\bscrC" "𝓒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrC" "𝓒" nil :system t)
    (puthash "\\bscrD" "𝓓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrD" "𝓓" nil :system t)
    (puthash "\\bscrE" "𝓔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrE" "𝓔" nil :system t)
    (puthash "\\bscrF" "𝓕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrF" "𝓕" nil :system t)
    (puthash "\\bscrG" "𝓖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrG" "𝓖" nil :system t)
    (puthash "\\bscrH" "𝓗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrH" "𝓗" nil :system t)
    (puthash "\\bscrI" "𝓘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrI" "𝓘" nil :system t)
    (puthash "\\bscrJ" "𝓙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrJ" "𝓙" nil :system t)
    (puthash "\\bscrK" "𝓚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrK" "𝓚" nil :system t)
    (puthash "\\bscrL" "𝓛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrL" "𝓛" nil :system t)
    (puthash "\\bscrM" "𝓜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrM" "𝓜" nil :system t)
    (puthash "\\bscrN" "𝓝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrN" "𝓝" nil :system t)
    (puthash "\\bscrO" "𝓞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrO" "𝓞" nil :system t)
    (puthash "\\bscrP" "𝓟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrP" "𝓟" nil :system t)
    (puthash "\\bscrQ" "𝓠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrQ" "𝓠" nil :system t)
    (puthash "\\bscrR" "𝓡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrR" "𝓡" nil :system t)
    (puthash "\\bscrS" "𝓢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrS" "𝓢" nil :system t)
    (puthash "\\bscrT" "𝓣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrT" "𝓣" nil :system t)
    (puthash "\\bscrU" "𝓤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrU" "𝓤" nil :system t)
    (puthash "\\bscrV" "𝓥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrV" "𝓥" nil :system t)
    (puthash "\\bscrW" "𝓦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrW" "𝓦" nil :system t)
    (puthash "\\bscrX" "𝓧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrX" "𝓧" nil :system t)
    (puthash "\\bscrY" "𝓨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrY" "𝓨" nil :system t)
    (puthash "\\bscrZ" "𝓩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrZ" "𝓩" nil :system t)
    (puthash "\\bscra" "𝓪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscra" "𝓪" nil :system t)
    (puthash "\\bscrb" "𝓫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrb" "𝓫" nil :system t)
    (puthash "\\bscrc" "𝓬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrc" "𝓬" nil :system t)
    (puthash "\\bscrd" "𝓭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrd" "𝓭" nil :system t)
    (puthash "\\bscre" "𝓮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscre" "𝓮" nil :system t)
    (puthash "\\bscrf" "𝓯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrf" "𝓯" nil :system t)
    (puthash "\\bscrg" "𝓰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrg" "𝓰" nil :system t)
    (puthash "\\bscrh" "𝓱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrh" "𝓱" nil :system t)
    (puthash "\\bscri" "𝓲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscri" "𝓲" nil :system t)
    (puthash "\\bscrj" "𝓳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrj" "𝓳" nil :system t)
    (puthash "\\bscrk" "𝓴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrk" "𝓴" nil :system t)
    (puthash "\\bscrl" "𝓵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrl" "𝓵" nil :system t)
    (puthash "\\bscrm" "𝓶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrm" "𝓶" nil :system t)
    (puthash "\\bscrn" "𝓷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrn" "𝓷" nil :system t)
    (puthash "\\bscro" "𝓸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscro" "𝓸" nil :system t)
    (puthash "\\bscrp" "𝓹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrp" "𝓹" nil :system t)
    (puthash "\\bscrq" "𝓺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrq" "𝓺" nil :system t)
    (puthash "\\bscrr" "𝓻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrr" "𝓻" nil :system t)
    (puthash "\\bscrs" "𝓼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrs" "𝓼" nil :system t)
    (puthash "\\bscrt" "𝓽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrt" "𝓽" nil :system t)
    (puthash "\\bscru" "𝓾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscru" "𝓾" nil :system t)
    (puthash "\\bscrv" "𝓿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrv" "𝓿" nil :system t)
    (puthash "\\bscrw" "𝔀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrw" "𝔀" nil :system t)
    (puthash "\\bscrx" "𝔁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrx" "𝔁" nil :system t)
    (puthash "\\bscry" "𝔂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscry" "𝔂" nil :system t)
    (puthash "\\bscrz" "𝔃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bscrz" "𝔃" nil :system t)
    (puthash "\\frakA" "𝔄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakA" "𝔄" nil :system t)
    (puthash "\\frakB" "𝔅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakB" "𝔅" nil :system t)
    (puthash "\\frakD" "𝔇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakD" "𝔇" nil :system t)
    (puthash "\\frakE" "𝔈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakE" "𝔈" nil :system t)
    (puthash "\\frakF" "𝔉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakF" "𝔉" nil :system t)
    (puthash "\\frakG" "𝔊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakG" "𝔊" nil :system t)
    (puthash "\\frakJ" "𝔍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakJ" "𝔍" nil :system t)
    (puthash "\\frakK" "𝔎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakK" "𝔎" nil :system t)
    (puthash "\\frakL" "𝔏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakL" "𝔏" nil :system t)
    (puthash "\\frakM" "𝔐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakM" "𝔐" nil :system t)
    (puthash "\\frakN" "𝔑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakN" "𝔑" nil :system t)
    (puthash "\\frakO" "𝔒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakO" "𝔒" nil :system t)
    (puthash "\\frakP" "𝔓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakP" "𝔓" nil :system t)
    (puthash "\\frakQ" "𝔔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakQ" "𝔔" nil :system t)
    (puthash "\\frakS" "𝔖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakS" "𝔖" nil :system t)
    (puthash "\\frakT" "𝔗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakT" "𝔗" nil :system t)
    (puthash "\\frakU" "𝔘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakU" "𝔘" nil :system t)
    (puthash "\\frakV" "𝔙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakV" "𝔙" nil :system t)
    (puthash "\\frakW" "𝔚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakW" "𝔚" nil :system t)
    (puthash "\\frakX" "𝔛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakX" "𝔛" nil :system t)
    (puthash "\\frakY" "𝔜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakY" "𝔜" nil :system t)
    (puthash "\\fraka" "𝔞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fraka" "𝔞" nil :system t)
    (puthash "\\frakb" "𝔟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakb" "𝔟" nil :system t)
    (puthash "\\frakc" "𝔠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakc" "𝔠" nil :system t)
    (puthash "\\frakd" "𝔡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakd" "𝔡" nil :system t)
    (puthash "\\frake" "𝔢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frake" "𝔢" nil :system t)
    (puthash "\\frakf" "𝔣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakf" "𝔣" nil :system t)
    (puthash "\\frakg" "𝔤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakg" "𝔤" nil :system t)
    (puthash "\\frakh" "𝔥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakh" "𝔥" nil :system t)
    (puthash "\\fraki" "𝔦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fraki" "𝔦" nil :system t)
    (puthash "\\frakj" "𝔧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakj" "𝔧" nil :system t)
    (puthash "\\frakk" "𝔨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakk" "𝔨" nil :system t)
    (puthash "\\frakl" "𝔩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakl" "𝔩" nil :system t)
    (puthash "\\frakm" "𝔪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakm" "𝔪" nil :system t)
    (puthash "\\frakn" "𝔫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakn" "𝔫" nil :system t)
    (puthash "\\frako" "𝔬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frako" "𝔬" nil :system t)
    (puthash "\\frakp" "𝔭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakp" "𝔭" nil :system t)
    (puthash "\\frakq" "𝔮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakq" "𝔮" nil :system t)
    (puthash "\\frakr" "𝔯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakr" "𝔯" nil :system t)
    (puthash "\\fraks" "𝔰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fraks" "𝔰" nil :system t)
    (puthash "\\frakt" "𝔱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakt" "𝔱" nil :system t)
    (puthash "\\fraku" "𝔲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fraku" "𝔲" nil :system t)
    (puthash "\\frakv" "𝔳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakv" "𝔳" nil :system t)
    (puthash "\\frakw" "𝔴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakw" "𝔴" nil :system t)
    (puthash "\\frakx" "𝔵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakx" "𝔵" nil :system t)
    (puthash "\\fraky" "𝔶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\fraky" "𝔶" nil :system t)
    (puthash "\\frakz" "𝔷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\frakz" "𝔷" nil :system t)
    (puthash "\\bbA" "𝔸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbA" "𝔸" nil :system t)
    (puthash "\\bbB" "𝔹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbB" "𝔹" nil :system t)
    (puthash "\\bbD" "𝔻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbD" "𝔻" nil :system t)
    (puthash "\\bbE" "𝔼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbE" "𝔼" nil :system t)
    (puthash "\\bbF" "𝔽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbF" "𝔽" nil :system t)
    (puthash "\\bbG" "𝔾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbG" "𝔾" nil :system t)
    (puthash "\\bbI" "𝕀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbI" "𝕀" nil :system t)
    (puthash "\\bbJ" "𝕁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbJ" "𝕁" nil :system t)
    (puthash "\\bbK" "𝕂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbK" "𝕂" nil :system t)
    (puthash "\\bbL" "𝕃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbL" "𝕃" nil :system t)
    (puthash "\\bbM" "𝕄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbM" "𝕄" nil :system t)
    (puthash "\\bbO" "𝕆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbO" "𝕆" nil :system t)
    (puthash "\\bbS" "𝕊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbS" "𝕊" nil :system t)
    (puthash "\\bbT" "𝕋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbT" "𝕋" nil :system t)
    (puthash "\\bbU" "𝕌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbU" "𝕌" nil :system t)
    (puthash "\\bbV" "𝕍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbV" "𝕍" nil :system t)
    (puthash "\\bbW" "𝕎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbW" "𝕎" nil :system t)
    (puthash "\\bbX" "𝕏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbX" "𝕏" nil :system t)
    (puthash "\\bbY" "𝕐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbY" "𝕐" nil :system t)
    (puthash "\\bba" "𝕒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bba" "𝕒" nil :system t)
    (puthash "\\bbb" "𝕓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbb" "𝕓" nil :system t)
    (puthash "\\bbc" "𝕔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbc" "𝕔" nil :system t)
    (puthash "\\bbd" "𝕕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbd" "𝕕" nil :system t)
    (puthash "\\bbe" "𝕖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbe" "𝕖" nil :system t)
    (puthash "\\bbf" "𝕗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbf" "𝕗" nil :system t)
    (puthash "\\bbg" "𝕘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbg" "𝕘" nil :system t)
    (puthash "\\bbh" "𝕙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbh" "𝕙" nil :system t)
    (puthash "\\bbi" "𝕚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbi" "𝕚" nil :system t)
    (puthash "\\bbj" "𝕛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbj" "𝕛" nil :system t)
    (puthash "\\bbk" "𝕜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbk" "𝕜" nil :system t)
    (puthash "\\bbl" "𝕝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbl" "𝕝" nil :system t)
    (puthash "\\bbm" "𝕞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbm" "𝕞" nil :system t)
    (puthash "\\bbn" "𝕟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbn" "𝕟" nil :system t)
    (puthash "\\bbo" "𝕠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbo" "𝕠" nil :system t)
    (puthash "\\bbp" "𝕡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbp" "𝕡" nil :system t)
    (puthash "\\bbq" "𝕢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbq" "𝕢" nil :system t)
    (puthash "\\bbr" "𝕣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbr" "𝕣" nil :system t)
    (puthash "\\bbs" "𝕤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbs" "𝕤" nil :system t)
    (puthash "\\bbt" "𝕥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbt" "𝕥" nil :system t)
    (puthash "\\bbu" "𝕦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbu" "𝕦" nil :system t)
    (puthash "\\bbv" "𝕧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbv" "𝕧" nil :system t)
    (puthash "\\bbw" "𝕨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbw" "𝕨" nil :system t)
    (puthash "\\bbx" "𝕩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbx" "𝕩" nil :system t)
    (puthash "\\bby" "𝕪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bby" "𝕪" nil :system t)
    (puthash "\\bbz" "𝕫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbz" "𝕫" nil :system t)
    (puthash "\\bfrakA" "𝕬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakA" "𝕬" nil :system t)
    (puthash "\\bfrakB" "𝕭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakB" "𝕭" nil :system t)
    (puthash "\\bfrakC" "𝕮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakC" "𝕮" nil :system t)
    (puthash "\\bfrakD" "𝕯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakD" "𝕯" nil :system t)
    (puthash "\\bfrakE" "𝕰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakE" "𝕰" nil :system t)
    (puthash "\\bfrakF" "𝕱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakF" "𝕱" nil :system t)
    (puthash "\\bfrakG" "𝕲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakG" "𝕲" nil :system t)
    (puthash "\\bfrakH" "𝕳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakH" "𝕳" nil :system t)
    (puthash "\\bfrakI" "𝕴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakI" "𝕴" nil :system t)
    (puthash "\\bfrakJ" "𝕵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakJ" "𝕵" nil :system t)
    (puthash "\\bfrakK" "𝕶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakK" "𝕶" nil :system t)
    (puthash "\\bfrakL" "𝕷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakL" "𝕷" nil :system t)
    (puthash "\\bfrakM" "𝕸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakM" "𝕸" nil :system t)
    (puthash "\\bfrakN" "𝕹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakN" "𝕹" nil :system t)
    (puthash "\\bfrakO" "𝕺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakO" "𝕺" nil :system t)
    (puthash "\\bfrakP" "𝕻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakP" "𝕻" nil :system t)
    (puthash "\\bfrakQ" "𝕼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakQ" "𝕼" nil :system t)
    (puthash "\\bfrakR" "𝕽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakR" "𝕽" nil :system t)
    (puthash "\\bfrakS" "𝕾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakS" "𝕾" nil :system t)
    (puthash "\\bfrakT" "𝕿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakT" "𝕿" nil :system t)
    (puthash "\\bfrakU" "𝖀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakU" "𝖀" nil :system t)
    (puthash "\\bfrakV" "𝖁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakV" "𝖁" nil :system t)
    (puthash "\\bfrakW" "𝖂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakW" "𝖂" nil :system t)
    (puthash "\\bfrakX" "𝖃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakX" "𝖃" nil :system t)
    (puthash "\\bfrakY" "𝖄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakY" "𝖄" nil :system t)
    (puthash "\\bfrakZ" "𝖅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakZ" "𝖅" nil :system t)
    (puthash "\\bfraka" "𝖆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfraka" "𝖆" nil :system t)
    (puthash "\\bfrakb" "𝖇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakb" "𝖇" nil :system t)
    (puthash "\\bfrakc" "𝖈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakc" "𝖈" nil :system t)
    (puthash "\\bfrakd" "𝖉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakd" "𝖉" nil :system t)
    (puthash "\\bfrake" "𝖊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrake" "𝖊" nil :system t)
    (puthash "\\bfrakf" "𝖋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakf" "𝖋" nil :system t)
    (puthash "\\bfrakg" "𝖌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakg" "𝖌" nil :system t)
    (puthash "\\bfrakh" "𝖍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakh" "𝖍" nil :system t)
    (puthash "\\bfraki" "𝖎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfraki" "𝖎" nil :system t)
    (puthash "\\bfrakj" "𝖏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakj" "𝖏" nil :system t)
    (puthash "\\bfrakk" "𝖐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakk" "𝖐" nil :system t)
    (puthash "\\bfrakl" "𝖑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakl" "𝖑" nil :system t)
    (puthash "\\bfrakm" "𝖒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakm" "𝖒" nil :system t)
    (puthash "\\bfrakn" "𝖓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakn" "𝖓" nil :system t)
    (puthash "\\bfrako" "𝖔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrako" "𝖔" nil :system t)
    (puthash "\\bfrakp" "𝖕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakp" "𝖕" nil :system t)
    (puthash "\\bfrakq" "𝖖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakq" "𝖖" nil :system t)
    (puthash "\\bfrakr" "𝖗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakr" "𝖗" nil :system t)
    (puthash "\\bfraks" "𝖘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfraks" "𝖘" nil :system t)
    (puthash "\\bfrakt" "𝖙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakt" "𝖙" nil :system t)
    (puthash "\\bfraku" "𝖚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfraku" "𝖚" nil :system t)
    (puthash "\\bfrakv" "𝖛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakv" "𝖛" nil :system t)
    (puthash "\\bfrakw" "𝖜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakw" "𝖜" nil :system t)
    (puthash "\\bfrakx" "𝖝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakx" "𝖝" nil :system t)
    (puthash "\\bfraky" "𝖞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfraky" "𝖞" nil :system t)
    (puthash "\\bfrakz" "𝖟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrakz" "𝖟" nil :system t)
    (puthash "\\sansA" "𝖠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansA" "𝖠" nil :system t)
    (puthash "\\sansB" "𝖡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansB" "𝖡" nil :system t)
    (puthash "\\sansC" "𝖢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansC" "𝖢" nil :system t)
    (puthash "\\sansD" "𝖣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansD" "𝖣" nil :system t)
    (puthash "\\sansE" "𝖤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansE" "𝖤" nil :system t)
    (puthash "\\sansF" "𝖥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansF" "𝖥" nil :system t)
    (puthash "\\sansG" "𝖦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansG" "𝖦" nil :system t)
    (puthash "\\sansH" "𝖧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansH" "𝖧" nil :system t)
    (puthash "\\sansI" "𝖨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansI" "𝖨" nil :system t)
    (puthash "\\sansJ" "𝖩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansJ" "𝖩" nil :system t)
    (puthash "\\sansK" "𝖪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansK" "𝖪" nil :system t)
    (puthash "\\sansL" "𝖫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansL" "𝖫" nil :system t)
    (puthash "\\sansM" "𝖬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansM" "𝖬" nil :system t)
    (puthash "\\sansN" "𝖭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansN" "𝖭" nil :system t)
    (puthash "\\sansO" "𝖮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansO" "𝖮" nil :system t)
    (puthash "\\sansP" "𝖯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansP" "𝖯" nil :system t)
    (puthash "\\sansQ" "𝖰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansQ" "𝖰" nil :system t)
    (puthash "\\sansR" "𝖱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansR" "𝖱" nil :system t)
    (puthash "\\sansS" "𝖲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansS" "𝖲" nil :system t)
    (puthash "\\sansT" "𝖳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansT" "𝖳" nil :system t)
    (puthash "\\sansU" "𝖴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansU" "𝖴" nil :system t)
    (puthash "\\sansV" "𝖵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansV" "𝖵" nil :system t)
    (puthash "\\sansW" "𝖶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansW" "𝖶" nil :system t)
    (puthash "\\sansX" "𝖷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansX" "𝖷" nil :system t)
    (puthash "\\sansY" "𝖸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansY" "𝖸" nil :system t)
    (puthash "\\sansZ" "𝖹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansZ" "𝖹" nil :system t)
    (puthash "\\sansa" "𝖺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansa" "𝖺" nil :system t)
    (puthash "\\sansb" "𝖻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansb" "𝖻" nil :system t)
    (puthash "\\sansc" "𝖼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansc" "𝖼" nil :system t)
    (puthash "\\sansd" "𝖽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansd" "𝖽" nil :system t)
    (puthash "\\sanse" "𝖾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanse" "𝖾" nil :system t)
    (puthash "\\sansf" "𝖿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansf" "𝖿" nil :system t)
    (puthash "\\sansg" "𝗀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansg" "𝗀" nil :system t)
    (puthash "\\sansh" "𝗁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansh" "𝗁" nil :system t)
    (puthash "\\sansi" "𝗂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansi" "𝗂" nil :system t)
    (puthash "\\sansj" "𝗃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansj" "𝗃" nil :system t)
    (puthash "\\sansk" "𝗄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansk" "𝗄" nil :system t)
    (puthash "\\sansl" "𝗅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansl" "𝗅" nil :system t)
    (puthash "\\sansm" "𝗆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansm" "𝗆" nil :system t)
    (puthash "\\sansn" "𝗇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansn" "𝗇" nil :system t)
    (puthash "\\sanso" "𝗈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanso" "𝗈" nil :system t)
    (puthash "\\sansp" "𝗉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansp" "𝗉" nil :system t)
    (puthash "\\sansq" "𝗊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansq" "𝗊" nil :system t)
    (puthash "\\sansr" "𝗋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansr" "𝗋" nil :system t)
    (puthash "\\sanss" "𝗌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanss" "𝗌" nil :system t)
    (puthash "\\sanst" "𝗍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanst" "𝗍" nil :system t)
    (puthash "\\sansu" "𝗎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansu" "𝗎" nil :system t)
    (puthash "\\sansv" "𝗏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansv" "𝗏" nil :system t)
    (puthash "\\sansw" "𝗐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansw" "𝗐" nil :system t)
    (puthash "\\sansx" "𝗑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansx" "𝗑" nil :system t)
    (puthash "\\sansy" "𝗒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansy" "𝗒" nil :system t)
    (puthash "\\sansz" "𝗓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansz" "𝗓" nil :system t)
    (puthash "\\bsansA" "𝗔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansA" "𝗔" nil :system t)
    (puthash "\\bsansB" "𝗕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansB" "𝗕" nil :system t)
    (puthash "\\bsansC" "𝗖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansC" "𝗖" nil :system t)
    (puthash "\\bsansD" "𝗗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansD" "𝗗" nil :system t)
    (puthash "\\bsansE" "𝗘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansE" "𝗘" nil :system t)
    (puthash "\\bsansF" "𝗙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansF" "𝗙" nil :system t)
    (puthash "\\bsansG" "𝗚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansG" "𝗚" nil :system t)
    (puthash "\\bsansH" "𝗛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansH" "𝗛" nil :system t)
    (puthash "\\bsansI" "𝗜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansI" "𝗜" nil :system t)
    (puthash "\\bsansJ" "𝗝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansJ" "𝗝" nil :system t)
    (puthash "\\bsansK" "𝗞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansK" "𝗞" nil :system t)
    (puthash "\\bsansL" "𝗟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansL" "𝗟" nil :system t)
    (puthash "\\bsansM" "𝗠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansM" "𝗠" nil :system t)
    (puthash "\\bsansN" "𝗡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansN" "𝗡" nil :system t)
    (puthash "\\bsansO" "𝗢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansO" "𝗢" nil :system t)
    (puthash "\\bsansP" "𝗣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansP" "𝗣" nil :system t)
    (puthash "\\bsansQ" "𝗤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansQ" "𝗤" nil :system t)
    (puthash "\\bsansR" "𝗥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansR" "𝗥" nil :system t)
    (puthash "\\bsansS" "𝗦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansS" "𝗦" nil :system t)
    (puthash "\\bsansT" "𝗧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansT" "𝗧" nil :system t)
    (puthash "\\bsansU" "𝗨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansU" "𝗨" nil :system t)
    (puthash "\\bsansV" "𝗩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansV" "𝗩" nil :system t)
    (puthash "\\bsansW" "𝗪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansW" "𝗪" nil :system t)
    (puthash "\\bsansX" "𝗫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansX" "𝗫" nil :system t)
    (puthash "\\bsansY" "𝗬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansY" "𝗬" nil :system t)
    (puthash "\\bsansZ" "𝗭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansZ" "𝗭" nil :system t)
    (puthash "\\bsansa" "𝗮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansa" "𝗮" nil :system t)
    (puthash "\\bsansb" "𝗯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansb" "𝗯" nil :system t)
    (puthash "\\bsansc" "𝗰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansc" "𝗰" nil :system t)
    (puthash "\\bsansd" "𝗱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansd" "𝗱" nil :system t)
    (puthash "\\bsanse" "𝗲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanse" "𝗲" nil :system t)
    (puthash "\\bsansf" "𝗳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansf" "𝗳" nil :system t)
    (puthash "\\bsansg" "𝗴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansg" "𝗴" nil :system t)
    (puthash "\\bsansh" "𝗵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansh" "𝗵" nil :system t)
    (puthash "\\bsansi" "𝗶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansi" "𝗶" nil :system t)
    (puthash "\\bsansj" "𝗷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansj" "𝗷" nil :system t)
    (puthash "\\bsansk" "𝗸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansk" "𝗸" nil :system t)
    (puthash "\\bsansl" "𝗹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansl" "𝗹" nil :system t)
    (puthash "\\bsansm" "𝗺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansm" "𝗺" nil :system t)
    (puthash "\\bsansn" "𝗻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansn" "𝗻" nil :system t)
    (puthash "\\bsanso" "𝗼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanso" "𝗼" nil :system t)
    (puthash "\\bsansp" "𝗽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansp" "𝗽" nil :system t)
    (puthash "\\bsansq" "𝗾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansq" "𝗾" nil :system t)
    (puthash "\\bsansr" "𝗿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansr" "𝗿" nil :system t)
    (puthash "\\bsanss" "𝘀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanss" "𝘀" nil :system t)
    (puthash "\\bsanst" "𝘁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanst" "𝘁" nil :system t)
    (puthash "\\bsansu" "𝘂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansu" "𝘂" nil :system t)
    (puthash "\\bsansv" "𝘃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansv" "𝘃" nil :system t)
    (puthash "\\bsansw" "𝘄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansw" "𝘄" nil :system t)
    (puthash "\\bsansx" "𝘅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansx" "𝘅" nil :system t)
    (puthash "\\bsansy" "𝘆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansy" "𝘆" nil :system t)
    (puthash "\\bsansz" "𝘇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansz" "𝘇" nil :system t)
    (puthash "\\isansA" "𝘈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansA" "𝘈" nil :system t)
    (puthash "\\isansB" "𝘉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansB" "𝘉" nil :system t)
    (puthash "\\isansC" "𝘊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansC" "𝘊" nil :system t)
    (puthash "\\isansD" "𝘋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansD" "𝘋" nil :system t)
    (puthash "\\isansE" "𝘌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansE" "𝘌" nil :system t)
    (puthash "\\isansF" "𝘍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansF" "𝘍" nil :system t)
    (puthash "\\isansG" "𝘎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansG" "𝘎" nil :system t)
    (puthash "\\isansH" "𝘏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansH" "𝘏" nil :system t)
    (puthash "\\isansI" "𝘐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansI" "𝘐" nil :system t)
    (puthash "\\isansJ" "𝘑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansJ" "𝘑" nil :system t)
    (puthash "\\isansK" "𝘒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansK" "𝘒" nil :system t)
    (puthash "\\isansL" "𝘓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansL" "𝘓" nil :system t)
    (puthash "\\isansM" "𝘔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansM" "𝘔" nil :system t)
    (puthash "\\isansN" "𝘕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansN" "𝘕" nil :system t)
    (puthash "\\isansO" "𝘖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansO" "𝘖" nil :system t)
    (puthash "\\isansP" "𝘗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansP" "𝘗" nil :system t)
    (puthash "\\isansQ" "𝘘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansQ" "𝘘" nil :system t)
    (puthash "\\isansR" "𝘙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansR" "𝘙" nil :system t)
    (puthash "\\isansS" "𝘚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansS" "𝘚" nil :system t)
    (puthash "\\isansT" "𝘛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansT" "𝘛" nil :system t)
    (puthash "\\isansU" "𝘜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansU" "𝘜" nil :system t)
    (puthash "\\isansV" "𝘝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansV" "𝘝" nil :system t)
    (puthash "\\isansW" "𝘞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansW" "𝘞" nil :system t)
    (puthash "\\isansX" "𝘟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansX" "𝘟" nil :system t)
    (puthash "\\isansY" "𝘠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansY" "𝘠" nil :system t)
    (puthash "\\isansZ" "𝘡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansZ" "𝘡" nil :system t)
    (puthash "\\isansa" "𝘢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansa" "𝘢" nil :system t)
    (puthash "\\isansb" "𝘣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansb" "𝘣" nil :system t)
    (puthash "\\isansc" "𝘤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansc" "𝘤" nil :system t)
    (puthash "\\isansd" "𝘥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansd" "𝘥" nil :system t)
    (puthash "\\isanse" "𝘦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isanse" "𝘦" nil :system t)
    (puthash "\\isansf" "𝘧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansf" "𝘧" nil :system t)
    (puthash "\\isansg" "𝘨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansg" "𝘨" nil :system t)
    (puthash "\\isansh" "𝘩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansh" "𝘩" nil :system t)
    (puthash "\\isansi" "𝘪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansi" "𝘪" nil :system t)
    (puthash "\\isansj" "𝘫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansj" "𝘫" nil :system t)
    (puthash "\\isansk" "𝘬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansk" "𝘬" nil :system t)
    (puthash "\\isansl" "𝘭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansl" "𝘭" nil :system t)
    (puthash "\\isansm" "𝘮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansm" "𝘮" nil :system t)
    (puthash "\\isansn" "𝘯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansn" "𝘯" nil :system t)
    (puthash "\\isanso" "𝘰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isanso" "𝘰" nil :system t)
    (puthash "\\isansp" "𝘱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansp" "𝘱" nil :system t)
    (puthash "\\isansq" "𝘲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansq" "𝘲" nil :system t)
    (puthash "\\isansr" "𝘳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansr" "𝘳" nil :system t)
    (puthash "\\isanss" "𝘴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isanss" "𝘴" nil :system t)
    (puthash "\\isanst" "𝘵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isanst" "𝘵" nil :system t)
    (puthash "\\isansu" "𝘶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansu" "𝘶" nil :system t)
    (puthash "\\isansv" "𝘷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansv" "𝘷" nil :system t)
    (puthash "\\isansw" "𝘸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansw" "𝘸" nil :system t)
    (puthash "\\isansx" "𝘹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansx" "𝘹" nil :system t)
    (puthash "\\isansy" "𝘺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansy" "𝘺" nil :system t)
    (puthash "\\isansz" "𝘻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\isansz" "𝘻" nil :system t)
    (puthash "\\bisansA" "𝘼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansA" "𝘼" nil :system t)
    (puthash "\\bisansB" "𝘽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansB" "𝘽" nil :system t)
    (puthash "\\bisansC" "𝘾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansC" "𝘾" nil :system t)
    (puthash "\\bisansD" "𝘿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansD" "𝘿" nil :system t)
    (puthash "\\bisansE" "𝙀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansE" "𝙀" nil :system t)
    (puthash "\\bisansF" "𝙁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansF" "𝙁" nil :system t)
    (puthash "\\bisansG" "𝙂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansG" "𝙂" nil :system t)
    (puthash "\\bisansH" "𝙃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansH" "𝙃" nil :system t)
    (puthash "\\bisansI" "𝙄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansI" "𝙄" nil :system t)
    (puthash "\\bisansJ" "𝙅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansJ" "𝙅" nil :system t)
    (puthash "\\bisansK" "𝙆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansK" "𝙆" nil :system t)
    (puthash "\\bisansL" "𝙇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansL" "𝙇" nil :system t)
    (puthash "\\bisansM" "𝙈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansM" "𝙈" nil :system t)
    (puthash "\\bisansN" "𝙉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansN" "𝙉" nil :system t)
    (puthash "\\bisansO" "𝙊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansO" "𝙊" nil :system t)
    (puthash "\\bisansP" "𝙋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansP" "𝙋" nil :system t)
    (puthash "\\bisansQ" "𝙌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansQ" "𝙌" nil :system t)
    (puthash "\\bisansR" "𝙍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansR" "𝙍" nil :system t)
    (puthash "\\bisansS" "𝙎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansS" "𝙎" nil :system t)
    (puthash "\\bisansT" "𝙏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansT" "𝙏" nil :system t)
    (puthash "\\bisansU" "𝙐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansU" "𝙐" nil :system t)
    (puthash "\\bisansV" "𝙑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansV" "𝙑" nil :system t)
    (puthash "\\bisansW" "𝙒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansW" "𝙒" nil :system t)
    (puthash "\\bisansX" "𝙓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansX" "𝙓" nil :system t)
    (puthash "\\bisansY" "𝙔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansY" "𝙔" nil :system t)
    (puthash "\\bisansZ" "𝙕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansZ" "𝙕" nil :system t)
    (puthash "\\bisansa" "𝙖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansa" "𝙖" nil :system t)
    (puthash "\\bisansb" "𝙗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansb" "𝙗" nil :system t)
    (puthash "\\bisansc" "𝙘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansc" "𝙘" nil :system t)
    (puthash "\\bisansd" "𝙙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansd" "𝙙" nil :system t)
    (puthash "\\bisanse" "𝙚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanse" "𝙚" nil :system t)
    (puthash "\\bisansf" "𝙛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansf" "𝙛" nil :system t)
    (puthash "\\bisansg" "𝙜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansg" "𝙜" nil :system t)
    (puthash "\\bisansh" "𝙝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansh" "𝙝" nil :system t)
    (puthash "\\bisansi" "𝙞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansi" "𝙞" nil :system t)
    (puthash "\\bisansj" "𝙟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansj" "𝙟" nil :system t)
    (puthash "\\bisansk" "𝙠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansk" "𝙠" nil :system t)
    (puthash "\\bisansl" "𝙡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansl" "𝙡" nil :system t)
    (puthash "\\bisansm" "𝙢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansm" "𝙢" nil :system t)
    (puthash "\\bisansn" "𝙣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansn" "𝙣" nil :system t)
    (puthash "\\bisanso" "𝙤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanso" "𝙤" nil :system t)
    (puthash "\\bisansp" "𝙥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansp" "𝙥" nil :system t)
    (puthash "\\bisansq" "𝙦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansq" "𝙦" nil :system t)
    (puthash "\\bisansr" "𝙧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansr" "𝙧" nil :system t)
    (puthash "\\bisanss" "𝙨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanss" "𝙨" nil :system t)
    (puthash "\\bisanst" "𝙩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanst" "𝙩" nil :system t)
    (puthash "\\bisansu" "𝙪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansu" "𝙪" nil :system t)
    (puthash "\\bisansv" "𝙫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansv" "𝙫" nil :system t)
    (puthash "\\bisansw" "𝙬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansw" "𝙬" nil :system t)
    (puthash "\\bisansx" "𝙭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansx" "𝙭" nil :system t)
    (puthash "\\bisansy" "𝙮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansy" "𝙮" nil :system t)
    (puthash "\\bisansz" "𝙯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansz" "𝙯" nil :system t)
    (puthash "\\ttA" "𝙰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttA" "𝙰" nil :system t)
    (puthash "\\ttB" "𝙱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttB" "𝙱" nil :system t)
    (puthash "\\ttC" "𝙲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttC" "𝙲" nil :system t)
    (puthash "\\ttD" "𝙳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttD" "𝙳" nil :system t)
    (puthash "\\ttE" "𝙴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttE" "𝙴" nil :system t)
    (puthash "\\ttF" "𝙵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttF" "𝙵" nil :system t)
    (puthash "\\ttG" "𝙶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttG" "𝙶" nil :system t)
    (puthash "\\ttH" "𝙷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttH" "𝙷" nil :system t)
    (puthash "\\ttI" "𝙸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttI" "𝙸" nil :system t)
    (puthash "\\ttJ" "𝙹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttJ" "𝙹" nil :system t)
    (puthash "\\ttK" "𝙺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttK" "𝙺" nil :system t)
    (puthash "\\ttL" "𝙻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttL" "𝙻" nil :system t)
    (puthash "\\ttM" "𝙼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttM" "𝙼" nil :system t)
    (puthash "\\ttN" "𝙽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttN" "𝙽" nil :system t)
    (puthash "\\ttO" "𝙾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttO" "𝙾" nil :system t)
    (puthash "\\ttP" "𝙿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttP" "𝙿" nil :system t)
    (puthash "\\ttQ" "𝚀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttQ" "𝚀" nil :system t)
    (puthash "\\ttR" "𝚁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttR" "𝚁" nil :system t)
    (puthash "\\ttS" "𝚂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttS" "𝚂" nil :system t)
    (puthash "\\ttT" "𝚃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttT" "𝚃" nil :system t)
    (puthash "\\ttU" "𝚄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttU" "𝚄" nil :system t)
    (puthash "\\ttV" "𝚅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttV" "𝚅" nil :system t)
    (puthash "\\ttW" "𝚆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttW" "𝚆" nil :system t)
    (puthash "\\ttX" "𝚇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttX" "𝚇" nil :system t)
    (puthash "\\ttY" "𝚈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttY" "𝚈" nil :system t)
    (puthash "\\ttZ" "𝚉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttZ" "𝚉" nil :system t)
    (puthash "\\tta" "𝚊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tta" "𝚊" nil :system t)
    (puthash "\\ttb" "𝚋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttb" "𝚋" nil :system t)
    (puthash "\\ttc" "𝚌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttc" "𝚌" nil :system t)
    (puthash "\\ttd" "𝚍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttd" "𝚍" nil :system t)
    (puthash "\\tte" "𝚎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tte" "𝚎" nil :system t)
    (puthash "\\ttf" "𝚏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttf" "𝚏" nil :system t)
    (puthash "\\ttg" "𝚐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttg" "𝚐" nil :system t)
    (puthash "\\tth" "𝚑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tth" "𝚑" nil :system t)
    (puthash "\\tti" "𝚒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tti" "𝚒" nil :system t)
    (puthash "\\ttj" "𝚓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttj" "𝚓" nil :system t)
    (puthash "\\ttk" "𝚔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttk" "𝚔" nil :system t)
    (puthash "\\ttl" "𝚕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttl" "𝚕" nil :system t)
    (puthash "\\ttm" "𝚖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttm" "𝚖" nil :system t)
    (puthash "\\ttn" "𝚗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttn" "𝚗" nil :system t)
    (puthash "\\tto" "𝚘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tto" "𝚘" nil :system t)
    (puthash "\\ttp" "𝚙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttp" "𝚙" nil :system t)
    (puthash "\\ttq" "𝚚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttq" "𝚚" nil :system t)
    (puthash "\\ttr" "𝚛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttr" "𝚛" nil :system t)
    (puthash "\\tts" "𝚜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tts" "𝚜" nil :system t)
    (puthash "\\ttt" "𝚝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttt" "𝚝" nil :system t)
    (puthash "\\ttu" "𝚞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttu" "𝚞" nil :system t)
    (puthash "\\ttv" "𝚟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttv" "𝚟" nil :system t)
    (puthash "\\ttw" "𝚠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttw" "𝚠" nil :system t)
    (puthash "\\ttx" "𝚡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttx" "𝚡" nil :system t)
    (puthash "\\tty" "𝚢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tty" "𝚢" nil :system t)
    (puthash "\\ttz" "𝚣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttz" "𝚣" nil :system t)
    (puthash "\\itimath" "𝚤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itimath" "𝚤" nil :system t)
    (puthash "\\itjmath" "𝚥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itjmath" "𝚥" nil :system t)
    (puthash "\\bfAlpha" "𝚨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfAlpha" "𝚨" nil :system t)
    (puthash "\\bfBeta" "𝚩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfBeta" "𝚩" nil :system t)
    (puthash "\\bfGamma" "𝚪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfGamma" "𝚪" nil :system t)
    (puthash "\\bfDelta" "𝚫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfDelta" "𝚫" nil :system t)
    (puthash "\\bfEpsilon" "𝚬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfEpsilon" "𝚬" nil :system t)
    (puthash "\\bfZeta" "𝚭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfZeta" "𝚭" nil :system t)
    (puthash "\\bfEta" "𝚮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfEta" "𝚮" nil :system t)
    (puthash "\\bfTheta" "𝚯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfTheta" "𝚯" nil :system t)
    (puthash "\\bfIota" "𝚰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfIota" "𝚰" nil :system t)
    (puthash "\\bfKappa" "𝚱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfKappa" "𝚱" nil :system t)
    (puthash "\\bfLambda" "𝚲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfLambda" "𝚲" nil :system t)
    (puthash "\\bfMu" "𝚳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfMu" "𝚳" nil :system t)
    (puthash "\\bfNu" "𝚴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfNu" "𝚴" nil :system t)
    (puthash "\\bfXi" "𝚵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfXi" "𝚵" nil :system t)
    (puthash "\\bfOmicron" "𝚶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfOmicron" "𝚶" nil :system t)
    (puthash "\\bfPi" "𝚷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfPi" "𝚷" nil :system t)
    (puthash "\\bfRho" "𝚸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfRho" "𝚸" nil :system t)
    (puthash "\\bfvarTheta" "𝚹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarTheta" "𝚹" nil :system t)
    (puthash "\\bfSigma" "𝚺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfSigma" "𝚺" nil :system t)
    (puthash "\\bfTau" "𝚻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfTau" "𝚻" nil :system t)
    (puthash "\\bfUpsilon" "𝚼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfUpsilon" "𝚼" nil :system t)
    (puthash "\\bfPhi" "𝚽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfPhi" "𝚽" nil :system t)
    (puthash "\\bfChi" "𝚾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfChi" "𝚾" nil :system t)
    (puthash "\\bfPsi" "𝚿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfPsi" "𝚿" nil :system t)
    (puthash "\\bfOmega" "𝛀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfOmega" "𝛀" nil :system t)
    (puthash "\\bfnabla" "𝛁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfnabla" "𝛁" nil :system t)
    (puthash "\\bfalpha" "𝛂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfalpha" "𝛂" nil :system t)
    (puthash "\\bfbeta" "𝛃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfbeta" "𝛃" nil :system t)
    (puthash "\\bfgamma" "𝛄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfgamma" "𝛄" nil :system t)
    (puthash "\\bfdelta" "𝛅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfdelta" "𝛅" nil :system t)
    (puthash "\\bfvarepsilon" "𝛆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarepsilon" "𝛆" nil :system t)
    (puthash "\\bfzeta" "𝛇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfzeta" "𝛇" nil :system t)
    (puthash "\\bfeta" "𝛈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfeta" "𝛈" nil :system t)
    (puthash "\\bftheta" "𝛉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bftheta" "𝛉" nil :system t)
    (puthash "\\bfiota" "𝛊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfiota" "𝛊" nil :system t)
    (puthash "\\bfkappa" "𝛋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfkappa" "𝛋" nil :system t)
    (puthash "\\bflambda" "𝛌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bflambda" "𝛌" nil :system t)
    (puthash "\\bfmu" "𝛍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfmu" "𝛍" nil :system t)
    (puthash "\\bfnu" "𝛎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfnu" "𝛎" nil :system t)
    (puthash "\\bfxi" "𝛏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfxi" "𝛏" nil :system t)
    (puthash "\\bfomicron" "𝛐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfomicron" "𝛐" nil :system t)
    (puthash "\\bfpi" "𝛑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfpi" "𝛑" nil :system t)
    (puthash "\\bfrho" "𝛒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfrho" "𝛒" nil :system t)
    (puthash "\\bfvarsigma" "𝛓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarsigma" "𝛓" nil :system t)
    (puthash "\\bfsigma" "𝛔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfsigma" "𝛔" nil :system t)
    (puthash "\\bftau" "𝛕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bftau" "𝛕" nil :system t)
    (puthash "\\bfupsilon" "𝛖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfupsilon" "𝛖" nil :system t)
    (puthash "\\bfvarphi" "𝛗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarphi" "𝛗" nil :system t)
    (puthash "\\bfchi" "𝛘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfchi" "𝛘" nil :system t)
    (puthash "\\bfpsi" "𝛙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfpsi" "𝛙" nil :system t)
    (puthash "\\bfomega" "𝛚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfomega" "𝛚" nil :system t)
    (puthash "\\bfpartial" "𝛛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfpartial" "𝛛" nil :system t)
    (puthash "\\bfepsilon" "𝛜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfepsilon" "𝛜" nil :system t)
    (puthash "\\bfvartheta" "𝛝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvartheta" "𝛝" nil :system t)
    (puthash "\\bfvarkappa" "𝛞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarkappa" "𝛞" nil :system t)
    (puthash "\\bfphi" "𝛟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfphi" "𝛟" nil :system t)
    (puthash "\\bfvarrho" "𝛠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarrho" "𝛠" nil :system t)
    (puthash "\\bfvarpi" "𝛡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfvarpi" "𝛡" nil :system t)
    (puthash "\\itAlpha" "𝛢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itAlpha" "𝛢" nil :system t)
    (puthash "\\itBeta" "𝛣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itBeta" "𝛣" nil :system t)
    (puthash "\\itGamma" "𝛤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itGamma" "𝛤" nil :system t)
    (puthash "\\itDelta" "𝛥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itDelta" "𝛥" nil :system t)
    (puthash "\\itEpsilon" "𝛦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itEpsilon" "𝛦" nil :system t)
    (puthash "\\itZeta" "𝛧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itZeta" "𝛧" nil :system t)
    (puthash "\\itEta" "𝛨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itEta" "𝛨" nil :system t)
    (puthash "\\itTheta" "𝛩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itTheta" "𝛩" nil :system t)
    (puthash "\\itIota" "𝛪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itIota" "𝛪" nil :system t)
    (puthash "\\itKappa" "𝛫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itKappa" "𝛫" nil :system t)
    (puthash "\\itLambda" "𝛬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itLambda" "𝛬" nil :system t)
    (puthash "\\itMu" "𝛭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itMu" "𝛭" nil :system t)
    (puthash "\\itNu" "𝛮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itNu" "𝛮" nil :system t)
    (puthash "\\itXi" "𝛯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itXi" "𝛯" nil :system t)
    (puthash "\\itOmicron" "𝛰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itOmicron" "𝛰" nil :system t)
    (puthash "\\itPi" "𝛱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itPi" "𝛱" nil :system t)
    (puthash "\\itRho" "𝛲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itRho" "𝛲" nil :system t)
    (puthash "\\itvarTheta" "𝛳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarTheta" "𝛳" nil :system t)
    (puthash "\\itSigma" "𝛴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itSigma" "𝛴" nil :system t)
    (puthash "\\itTau" "𝛵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itTau" "𝛵" nil :system t)
    (puthash "\\itUpsilon" "𝛶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itUpsilon" "𝛶" nil :system t)
    (puthash "\\itPhi" "𝛷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itPhi" "𝛷" nil :system t)
    (puthash "\\itChi" "𝛸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itChi" "𝛸" nil :system t)
    (puthash "\\itPsi" "𝛹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itPsi" "𝛹" nil :system t)
    (puthash "\\itOmega" "𝛺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itOmega" "𝛺" nil :system t)
    (puthash "\\itnabla" "𝛻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itnabla" "𝛻" nil :system t)
    (puthash "\\italpha" "𝛼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\italpha" "𝛼" nil :system t)
    (puthash "\\itbeta" "𝛽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itbeta" "𝛽" nil :system t)
    (puthash "\\itgamma" "𝛾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itgamma" "𝛾" nil :system t)
    (puthash "\\itdelta" "𝛿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itdelta" "𝛿" nil :system t)
    (puthash "\\itvarepsilon" "𝜀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarepsilon" "𝜀" nil :system t)
    (puthash "\\itzeta" "𝜁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itzeta" "𝜁" nil :system t)
    (puthash "\\iteta" "𝜂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\iteta" "𝜂" nil :system t)
    (puthash "\\ittheta" "𝜃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ittheta" "𝜃" nil :system t)
    (puthash "\\itiota" "𝜄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itiota" "𝜄" nil :system t)
    (puthash "\\itkappa" "𝜅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itkappa" "𝜅" nil :system t)
    (puthash "\\itlambda" "𝜆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itlambda" "𝜆" nil :system t)
    (puthash "\\itmu" "𝜇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itmu" "𝜇" nil :system t)
    (puthash "\\itnu" "𝜈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itnu" "𝜈" nil :system t)
    (puthash "\\itxi" "𝜉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itxi" "𝜉" nil :system t)
    (puthash "\\itomicron" "𝜊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itomicron" "𝜊" nil :system t)
    (puthash "\\itpi" "𝜋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itpi" "𝜋" nil :system t)
    (puthash "\\itrho" "𝜌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itrho" "𝜌" nil :system t)
    (puthash "\\itvarsigma" "𝜍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarsigma" "𝜍" nil :system t)
    (puthash "\\itsigma" "𝜎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itsigma" "𝜎" nil :system t)
    (puthash "\\ittau" "𝜏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ittau" "𝜏" nil :system t)
    (puthash "\\itupsilon" "𝜐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itupsilon" "𝜐" nil :system t)
    (puthash "\\itvarphi" "𝜑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarphi" "𝜑" nil :system t)
    (puthash "\\itchi" "𝜒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itchi" "𝜒" nil :system t)
    (puthash "\\itpsi" "𝜓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itpsi" "𝜓" nil :system t)
    (puthash "\\itomega" "𝜔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itomega" "𝜔" nil :system t)
    (puthash "\\itpartial" "𝜕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itpartial" "𝜕" nil :system t)
    (puthash "\\itepsilon" "𝜖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itepsilon" "𝜖" nil :system t)
    (puthash "\\itvartheta" "𝜗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvartheta" "𝜗" nil :system t)
    (puthash "\\itvarkappa" "𝜘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarkappa" "𝜘" nil :system t)
    (puthash "\\itphi" "𝜙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itphi" "𝜙" nil :system t)
    (puthash "\\itvarrho" "𝜚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarrho" "𝜚" nil :system t)
    (puthash "\\itvarpi" "𝜛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\itvarpi" "𝜛" nil :system t)
    (puthash "\\biAlpha" "𝜜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biAlpha" "𝜜" nil :system t)
    (puthash "\\biBeta" "𝜝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biBeta" "𝜝" nil :system t)
    (puthash "\\biGamma" "𝜞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biGamma" "𝜞" nil :system t)
    (puthash "\\biDelta" "𝜟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biDelta" "𝜟" nil :system t)
    (puthash "\\biEpsilon" "𝜠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biEpsilon" "𝜠" nil :system t)
    (puthash "\\biZeta" "𝜡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biZeta" "𝜡" nil :system t)
    (puthash "\\biEta" "𝜢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biEta" "𝜢" nil :system t)
    (puthash "\\biTheta" "𝜣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biTheta" "𝜣" nil :system t)
    (puthash "\\biIota" "𝜤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biIota" "𝜤" nil :system t)
    (puthash "\\biKappa" "𝜥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biKappa" "𝜥" nil :system t)
    (puthash "\\biLambda" "𝜦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biLambda" "𝜦" nil :system t)
    (puthash "\\biMu" "𝜧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biMu" "𝜧" nil :system t)
    (puthash "\\biNu" "𝜨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biNu" "𝜨" nil :system t)
    (puthash "\\biXi" "𝜩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biXi" "𝜩" nil :system t)
    (puthash "\\biOmicron" "𝜪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biOmicron" "𝜪" nil :system t)
    (puthash "\\biPi" "𝜫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biPi" "𝜫" nil :system t)
    (puthash "\\biRho" "𝜬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biRho" "𝜬" nil :system t)
    (puthash "\\bivarTheta" "𝜭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarTheta" "𝜭" nil :system t)
    (puthash "\\biSigma" "𝜮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biSigma" "𝜮" nil :system t)
    (puthash "\\biTau" "𝜯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biTau" "𝜯" nil :system t)
    (puthash "\\biUpsilon" "𝜰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biUpsilon" "𝜰" nil :system t)
    (puthash "\\biPhi" "𝜱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biPhi" "𝜱" nil :system t)
    (puthash "\\biChi" "𝜲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biChi" "𝜲" nil :system t)
    (puthash "\\biPsi" "𝜳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biPsi" "𝜳" nil :system t)
    (puthash "\\biOmega" "𝜴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biOmega" "𝜴" nil :system t)
    (puthash "\\binabla" "𝜵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\binabla" "𝜵" nil :system t)
    (puthash "\\bialpha" "𝜶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bialpha" "𝜶" nil :system t)
    (puthash "\\bibeta" "𝜷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bibeta" "𝜷" nil :system t)
    (puthash "\\bigamma" "𝜸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bigamma" "𝜸" nil :system t)
    (puthash "\\bidelta" "𝜹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bidelta" "𝜹" nil :system t)
    (puthash "\\bivarepsilon" "𝜺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarepsilon" "𝜺" nil :system t)
    (puthash "\\bizeta" "𝜻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bizeta" "𝜻" nil :system t)
    (puthash "\\bieta" "𝜼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bieta" "𝜼" nil :system t)
    (puthash "\\bitheta" "𝜽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bitheta" "𝜽" nil :system t)
    (puthash "\\biiota" "𝜾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biiota" "𝜾" nil :system t)
    (puthash "\\bikappa" "𝜿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bikappa" "𝜿" nil :system t)
    (puthash "\\bilambda" "𝝀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bilambda" "𝝀" nil :system t)
    (puthash "\\bimu" "𝝁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bimu" "𝝁" nil :system t)
    (puthash "\\binu" "𝝂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\binu" "𝝂" nil :system t)
    (puthash "\\bixi" "𝝃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bixi" "𝝃" nil :system t)
    (puthash "\\biomicron" "𝝄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biomicron" "𝝄" nil :system t)
    (puthash "\\bipi" "𝝅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bipi" "𝝅" nil :system t)
    (puthash "\\birho" "𝝆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\birho" "𝝆" nil :system t)
    (puthash "\\bivarsigma" "𝝇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarsigma" "𝝇" nil :system t)
    (puthash "\\bisigma" "𝝈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisigma" "𝝈" nil :system t)
    (puthash "\\bitau" "𝝉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bitau" "𝝉" nil :system t)
    (puthash "\\biupsilon" "𝝊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biupsilon" "𝝊" nil :system t)
    (puthash "\\bivarphi" "𝝋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarphi" "𝝋" nil :system t)
    (puthash "\\bichi" "𝝌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bichi" "𝝌" nil :system t)
    (puthash "\\bipsi" "𝝍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bipsi" "𝝍" nil :system t)
    (puthash "\\biomega" "𝝎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biomega" "𝝎" nil :system t)
    (puthash "\\bipartial" "𝝏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bipartial" "𝝏" nil :system t)
    (puthash "\\biepsilon" "𝝐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biepsilon" "𝝐" nil :system t)
    (puthash "\\bivartheta" "𝝑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivartheta" "𝝑" nil :system t)
    (puthash "\\bivarkappa" "𝝒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarkappa" "𝝒" nil :system t)
    (puthash "\\biphi" "𝝓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\biphi" "𝝓" nil :system t)
    (puthash "\\bivarrho" "𝝔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarrho" "𝝔" nil :system t)
    (puthash "\\bivarpi" "𝝕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bivarpi" "𝝕" nil :system t)
    (puthash "\\bsansAlpha" "𝝖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansAlpha" "𝝖" nil :system t)
    (puthash "\\bsansBeta" "𝝗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansBeta" "𝝗" nil :system t)
    (puthash "\\bsansGamma" "𝝘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansGamma" "𝝘" nil :system t)
    (puthash "\\bsansDelta" "𝝙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansDelta" "𝝙" nil :system t)
    (puthash "\\bsansEpsilon" "𝝚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansEpsilon" "𝝚" nil :system t)
    (puthash "\\bsansZeta" "𝝛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansZeta" "𝝛" nil :system t)
    (puthash "\\bsansEta" "𝝜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansEta" "𝝜" nil :system t)
    (puthash "\\bsansTheta" "𝝝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansTheta" "𝝝" nil :system t)
    (puthash "\\bsansIota" "𝝞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansIota" "𝝞" nil :system t)
    (puthash "\\bsansKappa" "𝝟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansKappa" "𝝟" nil :system t)
    (puthash "\\bsansLambda" "𝝠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansLambda" "𝝠" nil :system t)
    (puthash "\\bsansMu" "𝝡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansMu" "𝝡" nil :system t)
    (puthash "\\bsansNu" "𝝢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansNu" "𝝢" nil :system t)
    (puthash "\\bsansXi" "𝝣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansXi" "𝝣" nil :system t)
    (puthash "\\bsansOmicron" "𝝤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansOmicron" "𝝤" nil :system t)
    (puthash "\\bsansPi" "𝝥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansPi" "𝝥" nil :system t)
    (puthash "\\bsansRho" "𝝦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansRho" "𝝦" nil :system t)
    (puthash "\\bsansvarTheta" "𝝧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarTheta" "𝝧" nil :system t)
    (puthash "\\bsansSigma" "𝝨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansSigma" "𝝨" nil :system t)
    (puthash "\\bsansTau" "𝝩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansTau" "𝝩" nil :system t)
    (puthash "\\bsansUpsilon" "𝝪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansUpsilon" "𝝪" nil :system t)
    (puthash "\\bsansPhi" "𝝫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansPhi" "𝝫" nil :system t)
    (puthash "\\bsansChi" "𝝬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansChi" "𝝬" nil :system t)
    (puthash "\\bsansPsi" "𝝭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansPsi" "𝝭" nil :system t)
    (puthash "\\bsansOmega" "𝝮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansOmega" "𝝮" nil :system t)
    (puthash "\\bsansnabla" "𝝯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansnabla" "𝝯" nil :system t)
    (puthash "\\bsansalpha" "𝝰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansalpha" "𝝰" nil :system t)
    (puthash "\\bsansbeta" "𝝱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansbeta" "𝝱" nil :system t)
    (puthash "\\bsansgamma" "𝝲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansgamma" "𝝲" nil :system t)
    (puthash "\\bsansdelta" "𝝳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansdelta" "𝝳" nil :system t)
    (puthash "\\bsansvarepsilon" "𝝴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarepsilon" "𝝴" nil :system t)
    (puthash "\\bsanszeta" "𝝵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanszeta" "𝝵" nil :system t)
    (puthash "\\bsanseta" "𝝶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanseta" "𝝶" nil :system t)
    (puthash "\\bsanstheta" "𝝷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanstheta" "𝝷" nil :system t)
    (puthash "\\bsansiota" "𝝸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansiota" "𝝸" nil :system t)
    (puthash "\\bsanskappa" "𝝹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanskappa" "𝝹" nil :system t)
    (puthash "\\bsanslambda" "𝝺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanslambda" "𝝺" nil :system t)
    (puthash "\\bsansmu" "𝝻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansmu" "𝝻" nil :system t)
    (puthash "\\bsansnu" "𝝼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansnu" "𝝼" nil :system t)
    (puthash "\\bsansxi" "𝝽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansxi" "𝝽" nil :system t)
    (puthash "\\bsansomicron" "𝝾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansomicron" "𝝾" nil :system t)
    (puthash "\\bsanspi" "𝝿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanspi" "𝝿" nil :system t)
    (puthash "\\bsansrho" "𝞀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansrho" "𝞀" nil :system t)
    (puthash "\\bsansvarsigma" "𝞁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarsigma" "𝞁" nil :system t)
    (puthash "\\bsanssigma" "𝞂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanssigma" "𝞂" nil :system t)
    (puthash "\\bsanstau" "𝞃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanstau" "𝞃" nil :system t)
    (puthash "\\bsansupsilon" "𝞄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansupsilon" "𝞄" nil :system t)
    (puthash "\\bsansvarphi" "𝞅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarphi" "𝞅" nil :system t)
    (puthash "\\bsanschi" "𝞆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanschi" "𝞆" nil :system t)
    (puthash "\\bsanspsi" "𝞇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanspsi" "𝞇" nil :system t)
    (puthash "\\bsansomega" "𝞈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansomega" "𝞈" nil :system t)
    (puthash "\\bsanspartial" "𝞉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanspartial" "𝞉" nil :system t)
    (puthash "\\bsansepsilon" "𝞊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansepsilon" "𝞊" nil :system t)
    (puthash "\\bsansvartheta" "𝞋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvartheta" "𝞋" nil :system t)
    (puthash "\\bsansvarkappa" "𝞌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarkappa" "𝞌" nil :system t)
    (puthash "\\bsansphi" "𝞍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansphi" "𝞍" nil :system t)
    (puthash "\\bsansvarrho" "𝞎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarrho" "𝞎" nil :system t)
    (puthash "\\bsansvarpi" "𝞏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansvarpi" "𝞏" nil :system t)
    (puthash "\\bisansAlpha" "𝞐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansAlpha" "𝞐" nil :system t)
    (puthash "\\bisansBeta" "𝞑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansBeta" "𝞑" nil :system t)
    (puthash "\\bisansGamma" "𝞒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansGamma" "𝞒" nil :system t)
    (puthash "\\bisansDelta" "𝞓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansDelta" "𝞓" nil :system t)
    (puthash "\\bisansEpsilon" "𝞔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansEpsilon" "𝞔" nil :system t)
    (puthash "\\bisansZeta" "𝞕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansZeta" "𝞕" nil :system t)
    (puthash "\\bisansEta" "𝞖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansEta" "𝞖" nil :system t)
    (puthash "\\bisansTheta" "𝞗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansTheta" "𝞗" nil :system t)
    (puthash "\\bisansIota" "𝞘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansIota" "𝞘" nil :system t)
    (puthash "\\bisansKappa" "𝞙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansKappa" "𝞙" nil :system t)
    (puthash "\\bisansLambda" "𝞚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansLambda" "𝞚" nil :system t)
    (puthash "\\bisansMu" "𝞛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansMu" "𝞛" nil :system t)
    (puthash "\\bisansNu" "𝞜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansNu" "𝞜" nil :system t)
    (puthash "\\bisansXi" "𝞝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansXi" "𝞝" nil :system t)
    (puthash "\\bisansOmicron" "𝞞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansOmicron" "𝞞" nil :system t)
    (puthash "\\bisansPi" "𝞟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansPi" "𝞟" nil :system t)
    (puthash "\\bisansRho" "𝞠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansRho" "𝞠" nil :system t)
    (puthash "\\bisansvarTheta" "𝞡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarTheta" "𝞡" nil :system t)
    (puthash "\\bisansSigma" "𝞢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansSigma" "𝞢" nil :system t)
    (puthash "\\bisansTau" "𝞣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansTau" "𝞣" nil :system t)
    (puthash "\\bisansUpsilon" "𝞤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansUpsilon" "𝞤" nil :system t)
    (puthash "\\bisansPhi" "𝞥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansPhi" "𝞥" nil :system t)
    (puthash "\\bisansChi" "𝞦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansChi" "𝞦" nil :system t)
    (puthash "\\bisansPsi" "𝞧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansPsi" "𝞧" nil :system t)
    (puthash "\\bisansOmega" "𝞨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansOmega" "𝞨" nil :system t)
    (puthash "\\bisansnabla" "𝞩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansnabla" "𝞩" nil :system t)
    (puthash "\\bisansalpha" "𝞪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansalpha" "𝞪" nil :system t)
    (puthash "\\bisansbeta" "𝞫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansbeta" "𝞫" nil :system t)
    (puthash "\\bisansgamma" "𝞬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansgamma" "𝞬" nil :system t)
    (puthash "\\bisansdelta" "𝞭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansdelta" "𝞭" nil :system t)
    (puthash "\\bisansvarepsilon" "𝞮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarepsilon" "𝞮" nil :system t)
    (puthash "\\bisanszeta" "𝞯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanszeta" "𝞯" nil :system t)
    (puthash "\\bisanseta" "𝞰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanseta" "𝞰" nil :system t)
    (puthash "\\bisanstheta" "𝞱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanstheta" "𝞱" nil :system t)
    (puthash "\\bisansiota" "𝞲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansiota" "𝞲" nil :system t)
    (puthash "\\bisanskappa" "𝞳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanskappa" "𝞳" nil :system t)
    (puthash "\\bisanslambda" "𝞴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanslambda" "𝞴" nil :system t)
    (puthash "\\bisansmu" "𝞵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansmu" "𝞵" nil :system t)
    (puthash "\\bisansnu" "𝞶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansnu" "𝞶" nil :system t)
    (puthash "\\bisansxi" "𝞷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansxi" "𝞷" nil :system t)
    (puthash "\\bisansomicron" "𝞸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansomicron" "𝞸" nil :system t)
    (puthash "\\bisanspi" "𝞹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanspi" "𝞹" nil :system t)
    (puthash "\\bisansrho" "𝞺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansrho" "𝞺" nil :system t)
    (puthash "\\bisansvarsigma" "𝞻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarsigma" "𝞻" nil :system t)
    (puthash "\\bisanssigma" "𝞼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanssigma" "𝞼" nil :system t)
    (puthash "\\bisanstau" "𝞽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanstau" "𝞽" nil :system t)
    (puthash "\\bisansupsilon" "𝞾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansupsilon" "𝞾" nil :system t)
    (puthash "\\bisansvarphi" "𝞿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarphi" "𝞿" nil :system t)
    (puthash "\\bisanschi" "𝟀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanschi" "𝟀" nil :system t)
    (puthash "\\bisanspsi" "𝟁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanspsi" "𝟁" nil :system t)
    (puthash "\\bisansomega" "𝟂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansomega" "𝟂" nil :system t)
    (puthash "\\bisanspartial" "𝟃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisanspartial" "𝟃" nil :system t)
    (puthash "\\bisansepsilon" "𝟄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansepsilon" "𝟄" nil :system t)
    (puthash "\\bisansvartheta" "𝟅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvartheta" "𝟅" nil :system t)
    (puthash "\\bisansvarkappa" "𝟆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarkappa" "𝟆" nil :system t)
    (puthash "\\bisansphi" "𝟇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansphi" "𝟇" nil :system t)
    (puthash "\\bisansvarrho" "𝟈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarrho" "𝟈" nil :system t)
    (puthash "\\bisansvarpi" "𝟉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bisansvarpi" "𝟉" nil :system t)
    (puthash "\\bfDigamma" "𝟊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfDigamma" "𝟊" nil :system t)
    (puthash "\\bfdigamma" "𝟋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfdigamma" "𝟋" nil :system t)
    (puthash "\\bfzero" "𝟎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfzero" "𝟎" nil :system t)
    (puthash "\\bfone" "𝟏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfone" "𝟏" nil :system t)
    (puthash "\\bftwo" "𝟐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bftwo" "𝟐" nil :system t)
    (puthash "\\bfthree" "𝟑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfthree" "𝟑" nil :system t)
    (puthash "\\bffour" "𝟒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bffour" "𝟒" nil :system t)
    (puthash "\\bffive" "𝟓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bffive" "𝟓" nil :system t)
    (puthash "\\bfsix" "𝟔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfsix" "𝟔" nil :system t)
    (puthash "\\bfseven" "𝟕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfseven" "𝟕" nil :system t)
    (puthash "\\bfeight" "𝟖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfeight" "𝟖" nil :system t)
    (puthash "\\bfnine" "𝟗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bfnine" "𝟗" nil :system t)
    (puthash "\\bbzero" "𝟘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbzero" "𝟘" nil :system t)
    (puthash "\\bbone" "𝟙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbone" "𝟙" nil :system t)
    (puthash "\\bbtwo" "𝟚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbtwo" "𝟚" nil :system t)
    (puthash "\\bbthree" "𝟛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbthree" "𝟛" nil :system t)
    (puthash "\\bbfour" "𝟜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbfour" "𝟜" nil :system t)
    (puthash "\\bbfive" "𝟝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbfive" "𝟝" nil :system t)
    (puthash "\\bbsix" "𝟞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbsix" "𝟞" nil :system t)
    (puthash "\\bbseven" "𝟟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbseven" "𝟟" nil :system t)
    (puthash "\\bbeight" "𝟠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbeight" "𝟠" nil :system t)
    (puthash "\\bbnine" "𝟡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bbnine" "𝟡" nil :system t)
    (puthash "\\sanszero" "𝟢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanszero" "𝟢" nil :system t)
    (puthash "\\sansone" "𝟣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansone" "𝟣" nil :system t)
    (puthash "\\sanstwo" "𝟤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanstwo" "𝟤" nil :system t)
    (puthash "\\sansthree" "𝟥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansthree" "𝟥" nil :system t)
    (puthash "\\sansfour" "𝟦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansfour" "𝟦" nil :system t)
    (puthash "\\sansfive" "𝟧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansfive" "𝟧" nil :system t)
    (puthash "\\sanssix" "𝟨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanssix" "𝟨" nil :system t)
    (puthash "\\sansseven" "𝟩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansseven" "𝟩" nil :system t)
    (puthash "\\sanseight" "𝟪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sanseight" "𝟪" nil :system t)
    (puthash "\\sansnine" "𝟫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\sansnine" "𝟫" nil :system t)
    (puthash "\\bsanszero" "𝟬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanszero" "𝟬" nil :system t)
    (puthash "\\bsansone" "𝟭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansone" "𝟭" nil :system t)
    (puthash "\\bsanstwo" "𝟮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanstwo" "𝟮" nil :system t)
    (puthash "\\bsansthree" "𝟯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansthree" "𝟯" nil :system t)
    (puthash "\\bsansfour" "𝟰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansfour" "𝟰" nil :system t)
    (puthash "\\bsansfive" "𝟱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansfive" "𝟱" nil :system t)
    (puthash "\\bsanssix" "𝟲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanssix" "𝟲" nil :system t)
    (puthash "\\bsansseven" "𝟳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansseven" "𝟳" nil :system t)
    (puthash "\\bsanseight" "𝟴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsanseight" "𝟴" nil :system t)
    (puthash "\\bsansnine" "𝟵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\bsansnine" "𝟵" nil :system t)
    (puthash "\\ttzero" "𝟶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttzero" "𝟶" nil :system t)
    (puthash "\\ttone" "𝟷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttone" "𝟷" nil :system t)
    (puthash "\\tttwo" "𝟸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tttwo" "𝟸" nil :system t)
    (puthash "\\ttthree" "𝟹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttthree" "𝟹" nil :system t)
    (puthash "\\ttfour" "𝟺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttfour" "𝟺" nil :system t)
    (puthash "\\ttfive" "𝟻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttfive" "𝟻" nil :system t)
    (puthash "\\ttsix" "𝟼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttsix" "𝟼" nil :system t)
    (puthash "\\ttseven" "𝟽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttseven" "𝟽" nil :system t)
    (puthash "\\tteight" "𝟾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\tteight" "𝟾" nil :system t)
    (puthash "\\ttnine" "𝟿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\ttnine" "𝟿" nil :system t)
    (puthash "\\:mahjong:" "🀄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mahjong:" "🀄" nil :system t)
    (puthash "\\:black_joker:" "🃏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_joker:" "🃏" nil :system t)
    (puthash "\\:a:" "🅰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:a:" "🅰" nil :system t)
    (puthash "\\:b:" "🅱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:b:" "🅱" nil :system t)
    (puthash "\\:o2:" "🅾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:o2:" "🅾" nil :system t)
    (puthash "\\:parking:" "🅿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:parking:" "🅿" nil :system t)
    (puthash "\\:ab:" "🆎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ab:" "🆎" nil :system t)
    (puthash "\\:cl:" "🆑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cl:" "🆑" nil :system t)
    (puthash "\\:cool:" "🆒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cool:" "🆒" nil :system t)
    (puthash "\\:free:" "🆓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:free:" "🆓" nil :system t)
    (puthash "\\:id:" "🆔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:id:" "🆔" nil :system t)
    (puthash "\\:new:" "🆕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:new:" "🆕" nil :system t)
    (puthash "\\:ng:" "🆖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ng:" "🆖" nil :system t)
    (puthash "\\:ok:" "🆗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ok:" "🆗" nil :system t)
    (puthash "\\:sos:" "🆘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sos:" "🆘" nil :system t)
    (puthash "\\:up:" "🆙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:up:" "🆙" nil :system t)
    (puthash "\\:vs:" "🆚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:vs:" "🆚" nil :system t)
    (puthash "\\:koko:" "🈁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:koko:" "🈁" nil :system t)
    (puthash "\\:sa:" "🈂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sa:" "🈂" nil :system t)
    (puthash "\\:u7121:" "🈚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u7121:" "🈚" nil :system t)
    (puthash "\\:u6307:" "🈯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u6307:" "🈯" nil :system t)
    (puthash "\\:u7981:" "🈲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u7981:" "🈲" nil :system t)
    (puthash "\\:u7a7a:" "🈳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u7a7a:" "🈳" nil :system t)
    (puthash "\\:u5408:" "🈴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u5408:" "🈴" nil :system t)
    (puthash "\\:u6e80:" "🈵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u6e80:" "🈵" nil :system t)
    (puthash "\\:u6709:" "🈶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u6709:" "🈶" nil :system t)
    (puthash "\\:u6708:" "🈷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u6708:" "🈷" nil :system t)
    (puthash "\\:u7533:" "🈸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u7533:" "🈸" nil :system t)
    (puthash "\\:u5272:" "🈹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u5272:" "🈹" nil :system t)
    (puthash "\\:u55b6:" "🈺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:u55b6:" "🈺" nil :system t)
    (puthash "\\:ideograph_advantage:" "🉐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ideograph_advantage:" "🉐" nil :system t)
    (puthash "\\:accept:" "🉑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:accept:" "🉑" nil :system t)
    (puthash "\\:cyclone:" "🌀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cyclone:" "🌀" nil :system t)
    (puthash "\\:foggy:" "🌁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:foggy:" "🌁" nil :system t)
    (puthash "\\:closed_umbrella:" "🌂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:closed_umbrella:" "🌂" nil :system t)
    (puthash "\\:night_with_stars:" "🌃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:night_with_stars:" "🌃" nil :system t)
    (puthash "\\:sunrise_over_mountains:" "🌄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sunrise_over_mountains:" "🌄" nil :system t)
    (puthash "\\:sunrise:" "🌅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sunrise:" "🌅" nil :system t)
    (puthash "\\:city_sunset:" "🌆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:city_sunset:" "🌆" nil :system t)
    (puthash "\\:city_sunrise:" "🌇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:city_sunrise:" "🌇" nil :system t)
    (puthash "\\:rainbow:" "🌈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rainbow:" "🌈" nil :system t)
    (puthash "\\:bridge_at_night:" "🌉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bridge_at_night:" "🌉" nil :system t)
    (puthash "\\:ocean:" "🌊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ocean:" "🌊" nil :system t)
    (puthash "\\:volcano:" "🌋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:volcano:" "🌋" nil :system t)
    (puthash "\\:milky_way:" "🌌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:milky_way:" "🌌" nil :system t)
    (puthash "\\:earth_africa:" "🌍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:earth_africa:" "🌍" nil :system t)
    (puthash "\\:earth_americas:" "🌎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:earth_americas:" "🌎" nil :system t)
    (puthash "\\:earth_asia:" "🌏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:earth_asia:" "🌏" nil :system t)
    (puthash "\\:globe_with_meridians:" "🌐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:globe_with_meridians:" "🌐" nil :system t)
    (puthash "\\:new_moon:" "🌑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:new_moon:" "🌑" nil :system t)
    (puthash "\\:waxing_crescent_moon:" "🌒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:waxing_crescent_moon:" "🌒" nil :system t)
    (puthash "\\:first_quarter_moon:" "🌓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:first_quarter_moon:" "🌓" nil :system t)
    (puthash "\\:moon:" "🌔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:moon:" "🌔" nil :system t)
    (puthash "\\:full_moon:" "🌕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:full_moon:" "🌕" nil :system t)
    (puthash "\\:waning_gibbous_moon:" "🌖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:waning_gibbous_moon:" "🌖" nil :system t)
    (puthash "\\:last_quarter_moon:" "🌗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:last_quarter_moon:" "🌗" nil :system t)
    (puthash "\\:waning_crescent_moon:" "🌘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:waning_crescent_moon:" "🌘" nil :system t)
    (puthash "\\:crescent_moon:" "🌙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crescent_moon:" "🌙" nil :system t)
    (puthash "\\:new_moon_with_face:" "🌚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:new_moon_with_face:" "🌚" nil :system t)
    (puthash "\\:first_quarter_moon_with_face:" "🌛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:first_quarter_moon_with_face:" "🌛" nil :system t)
    (puthash "\\:last_quarter_moon_with_face:" "🌜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:last_quarter_moon_with_face:" "🌜" nil :system t)
    (puthash "\\:full_moon_with_face:" "🌝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:full_moon_with_face:" "🌝" nil :system t)
    (puthash "\\:sun_with_face:" "🌞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sun_with_face:" "🌞" nil :system t)
    (puthash "\\:star2:" "🌟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:star2:" "🌟" nil :system t)
    (puthash "\\:stars:" "🌠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stars:" "🌠" nil :system t)
    (puthash "\\:hotdog:" "🌭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hotdog:" "🌭" nil :system t)
    (puthash "\\:taco:" "🌮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:taco:" "🌮" nil :system t)
    (puthash "\\:burrito:" "🌯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:burrito:" "🌯" nil :system t)
    (puthash "\\:chestnut:" "🌰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chestnut:" "🌰" nil :system t)
    (puthash "\\:seedling:" "🌱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:seedling:" "🌱" nil :system t)
    (puthash "\\:evergreen_tree:" "🌲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:evergreen_tree:" "🌲" nil :system t)
    (puthash "\\:deciduous_tree:" "🌳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:deciduous_tree:" "🌳" nil :system t)
    (puthash "\\:palm_tree:" "🌴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:palm_tree:" "🌴" nil :system t)
    (puthash "\\:cactus:" "🌵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cactus:" "🌵" nil :system t)
    (puthash "\\:tulip:" "🌷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tulip:" "🌷" nil :system t)
    (puthash "\\:cherry_blossom:" "🌸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cherry_blossom:" "🌸" nil :system t)
    (puthash "\\:rose:" "🌹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rose:" "🌹" nil :system t)
    (puthash "\\:hibiscus:" "🌺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hibiscus:" "🌺" nil :system t)
    (puthash "\\:sunflower:" "🌻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sunflower:" "🌻" nil :system t)
    (puthash "\\:blossom:" "🌼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blossom:" "🌼" nil :system t)
    (puthash "\\:corn:" "🌽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:corn:" "🌽" nil :system t)
    (puthash "\\:ear_of_rice:" "🌾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ear_of_rice:" "🌾" nil :system t)
    (puthash "\\:herb:" "🌿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:herb:" "🌿" nil :system t)
    (puthash "\\:four_leaf_clover:" "🍀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:four_leaf_clover:" "🍀" nil :system t)
    (puthash "\\:maple_leaf:" "🍁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:maple_leaf:" "🍁" nil :system t)
    (puthash "\\:fallen_leaf:" "🍂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fallen_leaf:" "🍂" nil :system t)
    (puthash "\\:leaves:" "🍃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:leaves:" "🍃" nil :system t)
    (puthash "\\:mushroom:" "🍄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mushroom:" "🍄" nil :system t)
    (puthash "\\:tomato:" "🍅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tomato:" "🍅" nil :system t)
    (puthash "\\:eggplant:" "🍆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:eggplant:" "🍆" nil :system t)
    (puthash "\\:grapes:" "🍇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:grapes:" "🍇" nil :system t)
    (puthash "\\:melon:" "🍈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:melon:" "🍈" nil :system t)
    (puthash "\\:watermelon:" "🍉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:watermelon:" "🍉" nil :system t)
    (puthash "\\:tangerine:" "🍊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tangerine:" "🍊" nil :system t)
    (puthash "\\:lemon:" "🍋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lemon:" "🍋" nil :system t)
    (puthash "\\:banana:" "🍌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:banana:" "🍌" nil :system t)
    (puthash "\\:pineapple:" "🍍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pineapple:" "🍍" nil :system t)
    (puthash "\\:apple:" "🍎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:apple:" "🍎" nil :system t)
    (puthash "\\:green_apple:" "🍏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:green_apple:" "🍏" nil :system t)
    (puthash "\\:pear:" "🍐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pear:" "🍐" nil :system t)
    (puthash "\\:peach:" "🍑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:peach:" "🍑" nil :system t)
    (puthash "\\:cherries:" "🍒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cherries:" "🍒" nil :system t)
    (puthash "\\:strawberry:" "🍓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:strawberry:" "🍓" nil :system t)
    (puthash "\\:hamburger:" "🍔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hamburger:" "🍔" nil :system t)
    (puthash "\\:pizza:" "🍕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pizza:" "🍕" nil :system t)
    (puthash "\\:meat_on_bone:" "🍖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:meat_on_bone:" "🍖" nil :system t)
    (puthash "\\:poultry_leg:" "🍗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:poultry_leg:" "🍗" nil :system t)
    (puthash "\\:rice_cracker:" "🍘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rice_cracker:" "🍘" nil :system t)
    (puthash "\\:rice_ball:" "🍙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rice_ball:" "🍙" nil :system t)
    (puthash "\\:rice:" "🍚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rice:" "🍚" nil :system t)
    (puthash "\\:curry:" "🍛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:curry:" "🍛" nil :system t)
    (puthash "\\:ramen:" "🍜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ramen:" "🍜" nil :system t)
    (puthash "\\:spaghetti:" "🍝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:spaghetti:" "🍝" nil :system t)
    (puthash "\\:bread:" "🍞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bread:" "🍞" nil :system t)
    (puthash "\\:fries:" "🍟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fries:" "🍟" nil :system t)
    (puthash "\\:sweet_potato:" "🍠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sweet_potato:" "🍠" nil :system t)
    (puthash "\\:dango:" "🍡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dango:" "🍡" nil :system t)
    (puthash "\\:oden:" "🍢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:oden:" "🍢" nil :system t)
    (puthash "\\:sushi:" "🍣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sushi:" "🍣" nil :system t)
    (puthash "\\:fried_shrimp:" "🍤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fried_shrimp:" "🍤" nil :system t)
    (puthash "\\:fish_cake:" "🍥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fish_cake:" "🍥" nil :system t)
    (puthash "\\:icecream:" "🍦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:icecream:" "🍦" nil :system t)
    (puthash "\\:shaved_ice:" "🍧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shaved_ice:" "🍧" nil :system t)
    (puthash "\\:ice_cream:" "🍨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ice_cream:" "🍨" nil :system t)
    (puthash "\\:doughnut:" "🍩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:doughnut:" "🍩" nil :system t)
    (puthash "\\:cookie:" "🍪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cookie:" "🍪" nil :system t)
    (puthash "\\:chocolate_bar:" "🍫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chocolate_bar:" "🍫" nil :system t)
    (puthash "\\:candy:" "🍬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:candy:" "🍬" nil :system t)
    (puthash "\\:lollipop:" "🍭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lollipop:" "🍭" nil :system t)
    (puthash "\\:custard:" "🍮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:custard:" "🍮" nil :system t)
    (puthash "\\:honey_pot:" "🍯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:honey_pot:" "🍯" nil :system t)
    (puthash "\\:cake:" "🍰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cake:" "🍰" nil :system t)
    (puthash "\\:bento:" "🍱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bento:" "🍱" nil :system t)
    (puthash "\\:stew:" "🍲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stew:" "🍲" nil :system t)
    (puthash "\\:fried_egg:" "🍳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fried_egg:" "🍳" nil :system t)
    (puthash "\\:fork_and_knife:" "🍴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fork_and_knife:" "🍴" nil :system t)
    (puthash "\\:tea:" "🍵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tea:" "🍵" nil :system t)
    (puthash "\\:sake:" "🍶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sake:" "🍶" nil :system t)
    (puthash "\\:wine_glass:" "🍷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wine_glass:" "🍷" nil :system t)
    (puthash "\\:cocktail:" "🍸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cocktail:" "🍸" nil :system t)
    (puthash "\\:tropical_drink:" "🍹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tropical_drink:" "🍹" nil :system t)
    (puthash "\\:beer:" "🍺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:beer:" "🍺" nil :system t)
    (puthash "\\:beers:" "🍻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:beers:" "🍻" nil :system t)
    (puthash "\\:baby_bottle:" "🍼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baby_bottle:" "🍼" nil :system t)
    (puthash "\\:champagne:" "🍾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:champagne:" "🍾" nil :system t)
    (puthash "\\:popcorn:" "🍿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:popcorn:" "🍿" nil :system t)
    (puthash "\\:ribbon:" "🎀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ribbon:" "🎀" nil :system t)
    (puthash "\\:gift:" "🎁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gift:" "🎁" nil :system t)
    (puthash "\\:birthday:" "🎂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:birthday:" "🎂" nil :system t)
    (puthash "\\:jack_o_lantern:" "🎃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:jack_o_lantern:" "🎃" nil :system t)
    (puthash "\\:christmas_tree:" "🎄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:christmas_tree:" "🎄" nil :system t)
    (puthash "\\:santa:" "🎅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:santa:" "🎅" nil :system t)
    (puthash "\\:fireworks:" "🎆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fireworks:" "🎆" nil :system t)
    (puthash "\\:sparkler:" "🎇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sparkler:" "🎇" nil :system t)
    (puthash "\\:balloon:" "🎈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:balloon:" "🎈" nil :system t)
    (puthash "\\:tada:" "🎉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tada:" "🎉" nil :system t)
    (puthash "\\:confetti_ball:" "🎊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:confetti_ball:" "🎊" nil :system t)
    (puthash "\\:tanabata_tree:" "🎋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tanabata_tree:" "🎋" nil :system t)
    (puthash "\\:crossed_flags:" "🎌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crossed_flags:" "🎌" nil :system t)
    (puthash "\\:bamboo:" "🎍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bamboo:" "🎍" nil :system t)
    (puthash "\\:dolls:" "🎎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dolls:" "🎎" nil :system t)
    (puthash "\\:flags:" "🎏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flags:" "🎏" nil :system t)
    (puthash "\\:wind_chime:" "🎐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wind_chime:" "🎐" nil :system t)
    (puthash "\\:rice_scene:" "🎑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rice_scene:" "🎑" nil :system t)
    (puthash "\\:school_satchel:" "🎒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:school_satchel:" "🎒" nil :system t)
    (puthash "\\:mortar_board:" "🎓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mortar_board:" "🎓" nil :system t)
    (puthash "\\:carousel_horse:" "🎠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:carousel_horse:" "🎠" nil :system t)
    (puthash "\\:ferris_wheel:" "🎡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ferris_wheel:" "🎡" nil :system t)
    (puthash "\\:roller_coaster:" "🎢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:roller_coaster:" "🎢" nil :system t)
    (puthash "\\:fishing_pole_and_fish:" "🎣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fishing_pole_and_fish:" "🎣" nil :system t)
    (puthash "\\:microphone:" "🎤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:microphone:" "🎤" nil :system t)
    (puthash "\\:movie_camera:" "🎥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:movie_camera:" "🎥" nil :system t)
    (puthash "\\:cinema:" "🎦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cinema:" "🎦" nil :system t)
    (puthash "\\:headphones:" "🎧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:headphones:" "🎧" nil :system t)
    (puthash "\\:art:" "🎨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:art:" "🎨" nil :system t)
    (puthash "\\:tophat:" "🎩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tophat:" "🎩" nil :system t)
    (puthash "\\:circus_tent:" "🎪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:circus_tent:" "🎪" nil :system t)
    (puthash "\\:ticket:" "🎫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ticket:" "🎫" nil :system t)
    (puthash "\\:clapper:" "🎬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clapper:" "🎬" nil :system t)
    (puthash "\\:performing_arts:" "🎭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:performing_arts:" "🎭" nil :system t)
    (puthash "\\:video_game:" "🎮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:video_game:" "🎮" nil :system t)
    (puthash "\\:dart:" "🎯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dart:" "🎯" nil :system t)
    (puthash "\\:slot_machine:" "🎰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:slot_machine:" "🎰" nil :system t)
    (puthash "\\:8ball:" "🎱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:8ball:" "🎱" nil :system t)
    (puthash "\\:game_die:" "🎲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:game_die:" "🎲" nil :system t)
    (puthash "\\:bowling:" "🎳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bowling:" "🎳" nil :system t)
    (puthash "\\:flower_playing_cards:" "🎴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flower_playing_cards:" "🎴" nil :system t)
    (puthash "\\:musical_note:" "🎵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:musical_note:" "🎵" nil :system t)
    (puthash "\\:notes:" "🎶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:notes:" "🎶" nil :system t)
    (puthash "\\:saxophone:" "🎷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:saxophone:" "🎷" nil :system t)
    (puthash "\\:guitar:" "🎸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:guitar:" "🎸" nil :system t)
    (puthash "\\:musical_keyboard:" "🎹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:musical_keyboard:" "🎹" nil :system t)
    (puthash "\\:trumpet:" "🎺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:trumpet:" "🎺" nil :system t)
    (puthash "\\:violin:" "🎻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:violin:" "🎻" nil :system t)
    (puthash "\\:musical_score:" "🎼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:musical_score:" "🎼" nil :system t)
    (puthash "\\:running_shirt_with_sash:" "🎽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:running_shirt_with_sash:" "🎽" nil :system t)
    (puthash "\\:tennis:" "🎾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tennis:" "🎾" nil :system t)
    (puthash "\\:ski:" "🎿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ski:" "🎿" nil :system t)
    (puthash "\\:basketball:" "🏀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:basketball:" "🏀" nil :system t)
    (puthash "\\:checkered_flag:" "🏁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:checkered_flag:" "🏁" nil :system t)
    (puthash "\\:snowboarder:" "🏂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:snowboarder:" "🏂" nil :system t)
    (puthash "\\:runner:" "🏃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:runner:" "🏃" nil :system t)
    (puthash "\\:surfer:" "🏄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:surfer:" "🏄" nil :system t)
    (puthash "\\:sports_medal:" "🏅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sports_medal:" "🏅" nil :system t)
    (puthash "\\:trophy:" "🏆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:trophy:" "🏆" nil :system t)
    (puthash "\\:horse_racing:" "🏇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:horse_racing:" "🏇" nil :system t)
    (puthash "\\:football:" "🏈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:football:" "🏈" nil :system t)
    (puthash "\\:rugby_football:" "🏉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rugby_football:" "🏉" nil :system t)
    (puthash "\\:swimmer:" "🏊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:swimmer:" "🏊" nil :system t)
    (puthash "\\:cricket_bat_and_ball:" "🏏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cricket_bat_and_ball:" "🏏" nil :system t)
    (puthash "\\:volleyball:" "🏐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:volleyball:" "🏐" nil :system t)
    (puthash "\\:field_hockey_stick_and_ball:" "🏑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:field_hockey_stick_and_ball:" "🏑" nil :system t)
    (puthash "\\:ice_hockey_stick_and_puck:" "🏒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ice_hockey_stick_and_puck:" "🏒" nil :system t)
    (puthash "\\:table_tennis_paddle_and_ball:" "🏓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:table_tennis_paddle_and_ball:" "🏓" nil :system t)
    (puthash "\\:house:" "🏠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:house:" "🏠" nil :system t)
    (puthash "\\:house_with_garden:" "🏡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:house_with_garden:" "🏡" nil :system t)
    (puthash "\\:office:" "🏢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:office:" "🏢" nil :system t)
    (puthash "\\:post_office:" "🏣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:post_office:" "🏣" nil :system t)
    (puthash "\\:european_post_office:" "🏤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:european_post_office:" "🏤" nil :system t)
    (puthash "\\:hospital:" "🏥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hospital:" "🏥" nil :system t)
    (puthash "\\:bank:" "🏦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bank:" "🏦" nil :system t)
    (puthash "\\:atm:" "🏧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:atm:" "🏧" nil :system t)
    (puthash "\\:hotel:" "🏨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hotel:" "🏨" nil :system t)
    (puthash "\\:love_hotel:" "🏩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:love_hotel:" "🏩" nil :system t)
    (puthash "\\:convenience_store:" "🏪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:convenience_store:" "🏪" nil :system t)
    (puthash "\\:school:" "🏫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:school:" "🏫" nil :system t)
    (puthash "\\:department_store:" "🏬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:department_store:" "🏬" nil :system t)
    (puthash "\\:factory:" "🏭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:factory:" "🏭" nil :system t)
    (puthash "\\:izakaya_lantern:" "🏮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:izakaya_lantern:" "🏮" nil :system t)
    (puthash "\\:japanese_castle:" "🏯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:japanese_castle:" "🏯" nil :system t)
    (puthash "\\:european_castle:" "🏰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:european_castle:" "🏰" nil :system t)
    (puthash "\\:waving_black_flag:" "🏴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:waving_black_flag:" "🏴" nil :system t)
    (puthash "\\:badminton_racquet_and_shuttlecock:" "🏸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:badminton_racquet_and_shuttlecock:" "🏸" nil :system t)
    (puthash "\\:bow_and_arrow:" "🏹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bow_and_arrow:" "🏹" nil :system t)
    (puthash "\\:amphora:" "🏺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:amphora:" "🏺" nil :system t)
    (puthash "\\:skin-tone-2:" "🏻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skin-tone-2:" "🏻" nil :system t)
    (puthash "\\:skin-tone-3:" "🏼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skin-tone-3:" "🏼" nil :system t)
    (puthash "\\:skin-tone-4:" "🏽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skin-tone-4:" "🏽" nil :system t)
    (puthash "\\:skin-tone-5:" "🏾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skin-tone-5:" "🏾" nil :system t)
    (puthash "\\:skin-tone-6:" "🏿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skin-tone-6:" "🏿" nil :system t)
    (puthash "\\:rat:" "🐀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rat:" "🐀" nil :system t)
    (puthash "\\:mouse2:" "🐁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mouse2:" "🐁" nil :system t)
    (puthash "\\:ox:" "🐂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ox:" "🐂" nil :system t)
    (puthash "\\:water_buffalo:" "🐃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:water_buffalo:" "🐃" nil :system t)
    (puthash "\\:cow2:" "🐄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cow2:" "🐄" nil :system t)
    (puthash "\\:tiger2:" "🐅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tiger2:" "🐅" nil :system t)
    (puthash "\\:leopard:" "🐆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:leopard:" "🐆" nil :system t)
    (puthash "\\:rabbit2:" "🐇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rabbit2:" "🐇" nil :system t)
    (puthash "\\:cat2:" "🐈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cat2:" "🐈" nil :system t)
    (puthash "\\:dragon:" "🐉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dragon:" "🐉" nil :system t)
    (puthash "\\:crocodile:" "🐊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crocodile:" "🐊" nil :system t)
    (puthash "\\:whale2:" "🐋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:whale2:" "🐋" nil :system t)
    (puthash "\\:snail:" "🐌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:snail:" "🐌" nil :system t)
    (puthash "\\:snake:" "🐍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:snake:" "🐍" nil :system t)
    (puthash "\\:racehorse:" "🐎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:racehorse:" "🐎" nil :system t)
    (puthash "\\:ram:" "🐏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ram:" "🐏" nil :system t)
    (puthash "\\:goat:" "🐐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:goat:" "🐐" nil :system t)
    (puthash "\\:sheep:" "🐑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sheep:" "🐑" nil :system t)
    (puthash "\\:monkey:" "🐒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:monkey:" "🐒" nil :system t)
    (puthash "\\:rooster:" "🐓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rooster:" "🐓" nil :system t)
    (puthash "\\:chicken:" "🐔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chicken:" "🐔" nil :system t)
    (puthash "\\:dog2:" "🐕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dog2:" "🐕" nil :system t)
    (puthash "\\:pig2:" "🐖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pig2:" "🐖" nil :system t)
    (puthash "\\:boar:" "🐗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boar:" "🐗" nil :system t)
    (puthash "\\:elephant:" "🐘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:elephant:" "🐘" nil :system t)
    (puthash "\\:octopus:" "🐙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:octopus:" "🐙" nil :system t)
    (puthash "\\:shell:" "🐚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shell:" "🐚" nil :system t)
    (puthash "\\:bug:" "🐛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bug:" "🐛" nil :system t)
    (puthash "\\:ant:" "🐜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ant:" "🐜" nil :system t)
    (puthash "\\:bee:" "🐝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bee:" "🐝" nil :system t)
    (puthash "\\:ladybug:" "🐞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ladybug:" "🐞" nil :system t)
    (puthash "\\:fish:" "🐟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fish:" "🐟" nil :system t)
    (puthash "\\:tropical_fish:" "🐠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tropical_fish:" "🐠" nil :system t)
    (puthash "\\:blowfish:" "🐡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blowfish:" "🐡" nil :system t)
    (puthash "\\:turtle:" "🐢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:turtle:" "🐢" nil :system t)
    (puthash "\\:hatching_chick:" "🐣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hatching_chick:" "🐣" nil :system t)
    (puthash "\\:baby_chick:" "🐤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baby_chick:" "🐤" nil :system t)
    (puthash "\\:hatched_chick:" "🐥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hatched_chick:" "🐥" nil :system t)
    (puthash "\\:bird:" "🐦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bird:" "🐦" nil :system t)
    (puthash "\\:penguin:" "🐧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:penguin:" "🐧" nil :system t)
    (puthash "\\:koala:" "🐨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:koala:" "🐨" nil :system t)
    (puthash "\\:poodle:" "🐩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:poodle:" "🐩" nil :system t)
    (puthash "\\:dromedary_camel:" "🐪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dromedary_camel:" "🐪" nil :system t)
    (puthash "\\:camel:" "🐫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:camel:" "🐫" nil :system t)
    (puthash "\\:dolphin:" "🐬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dolphin:" "🐬" nil :system t)
    (puthash "\\:mouse:" "🐭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mouse:" "🐭" nil :system t)
    (puthash "\\:cow:" "🐮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cow:" "🐮" nil :system t)
    (puthash "\\:tiger:" "🐯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tiger:" "🐯" nil :system t)
    (puthash "\\:rabbit:" "🐰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rabbit:" "🐰" nil :system t)
    (puthash "\\:cat:" "🐱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cat:" "🐱" nil :system t)
    (puthash "\\:dragon_face:" "🐲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dragon_face:" "🐲" nil :system t)
    (puthash "\\:whale:" "🐳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:whale:" "🐳" nil :system t)
    (puthash "\\:horse:" "🐴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:horse:" "🐴" nil :system t)
    (puthash "\\:monkey_face:" "🐵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:monkey_face:" "🐵" nil :system t)
    (puthash "\\:dog:" "🐶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dog:" "🐶" nil :system t)
    (puthash "\\:pig:" "🐷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pig:" "🐷" nil :system t)
    (puthash "\\:frog:" "🐸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:frog:" "🐸" nil :system t)
    (puthash "\\:hamster:" "🐹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hamster:" "🐹" nil :system t)
    (puthash "\\:wolf:" "🐺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wolf:" "🐺" nil :system t)
    (puthash "\\:bear:" "🐻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bear:" "🐻" nil :system t)
    (puthash "\\:panda_face:" "🐼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:panda_face:" "🐼" nil :system t)
    (puthash "\\:pig_nose:" "🐽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pig_nose:" "🐽" nil :system t)
    (puthash "\\:feet:" "🐾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:feet:" "🐾" nil :system t)
    (puthash "\\:eyes:" "👀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:eyes:" "👀" nil :system t)
    (puthash "\\:ear:" "👂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ear:" "👂" nil :system t)
    (puthash "\\:nose:" "👃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nose:" "👃" nil :system t)
    (puthash "\\:lips:" "👄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lips:" "👄" nil :system t)
    (puthash "\\:tongue:" "👅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tongue:" "👅" nil :system t)
    (puthash "\\:point_up_2:" "👆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:point_up_2:" "👆" nil :system t)
    (puthash "\\:point_down:" "👇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:point_down:" "👇" nil :system t)
    (puthash "\\:point_left:" "👈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:point_left:" "👈" nil :system t)
    (puthash "\\:point_right:" "👉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:point_right:" "👉" nil :system t)
    (puthash "\\:facepunch:" "👊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:facepunch:" "👊" nil :system t)
    (puthash "\\:wave:" "👋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wave:" "👋" nil :system t)
    (puthash "\\:ok_hand:" "👌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ok_hand:" "👌" nil :system t)
    (puthash "\\:+1:" "👍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:+1:" "👍" nil :system t)
    (puthash "\\:-1:" "👎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:-1:" "👎" nil :system t)
    (puthash "\\:clap:" "👏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clap:" "👏" nil :system t)
    (puthash "\\:open_hands:" "👐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:open_hands:" "👐" nil :system t)
    (puthash "\\:crown:" "👑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crown:" "👑" nil :system t)
    (puthash "\\:womans_hat:" "👒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:womans_hat:" "👒" nil :system t)
    (puthash "\\:eyeglasses:" "👓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:eyeglasses:" "👓" nil :system t)
    (puthash "\\:necktie:" "👔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:necktie:" "👔" nil :system t)
    (puthash "\\:shirt:" "👕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shirt:" "👕" nil :system t)
    (puthash "\\:jeans:" "👖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:jeans:" "👖" nil :system t)
    (puthash "\\:dress:" "👗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dress:" "👗" nil :system t)
    (puthash "\\:kimono:" "👘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kimono:" "👘" nil :system t)
    (puthash "\\:bikini:" "👙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bikini:" "👙" nil :system t)
    (puthash "\\:womans_clothes:" "👚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:womans_clothes:" "👚" nil :system t)
    (puthash "\\:purse:" "👛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:purse:" "👛" nil :system t)
    (puthash "\\:handbag:" "👜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:handbag:" "👜" nil :system t)
    (puthash "\\:pouch:" "👝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pouch:" "👝" nil :system t)
    (puthash "\\:mans_shoe:" "👞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mans_shoe:" "👞" nil :system t)
    (puthash "\\:athletic_shoe:" "👟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:athletic_shoe:" "👟" nil :system t)
    (puthash "\\:high_heel:" "👠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:high_heel:" "👠" nil :system t)
    (puthash "\\:sandal:" "👡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sandal:" "👡" nil :system t)
    (puthash "\\:boot:" "👢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boot:" "👢" nil :system t)
    (puthash "\\:footprints:" "👣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:footprints:" "👣" nil :system t)
    (puthash "\\:bust_in_silhouette:" "👤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bust_in_silhouette:" "👤" nil :system t)
    (puthash "\\:busts_in_silhouette:" "👥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:busts_in_silhouette:" "👥" nil :system t)
    (puthash "\\:boy:" "👦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boy:" "👦" nil :system t)
    (puthash "\\:girl:" "👧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:girl:" "👧" nil :system t)
    (puthash "\\:man:" "👨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:man:" "👨" nil :system t)
    (puthash "\\:woman:" "👩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:woman:" "👩" nil :system t)
    (puthash "\\:family:" "👪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:family:" "👪" nil :system t)
    (puthash "\\:couple:" "👫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:couple:" "👫" nil :system t)
    (puthash "\\:man_and_woman_holding_hands:" "👫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:man_and_woman_holding_hands:" "👫" nil :system t)
    (puthash "\\:two_men_holding_hands:" "👬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:two_men_holding_hands:" "👬" nil :system t)
    (puthash "\\:two_women_holding_hands:" "👭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:two_women_holding_hands:" "👭" nil :system t)
    (puthash "\\:cop:" "👮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cop:" "👮" nil :system t)
    (puthash "\\:dancers:" "👯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dancers:" "👯" nil :system t)
    (puthash "\\:bride_with_veil:" "👰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bride_with_veil:" "👰" nil :system t)
    (puthash "\\:person_with_blond_hair:" "👱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_with_blond_hair:" "👱" nil :system t)
    (puthash "\\:man_with_gua_pi_mao:" "👲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:man_with_gua_pi_mao:" "👲" nil :system t)
    (puthash "\\:man_with_turban:" "👳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:man_with_turban:" "👳" nil :system t)
    (puthash "\\:older_man:" "👴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:older_man:" "👴" nil :system t)
    (puthash "\\:older_woman:" "👵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:older_woman:" "👵" nil :system t)
    (puthash "\\:baby:" "👶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baby:" "👶" nil :system t)
    (puthash "\\:construction_worker:" "👷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:construction_worker:" "👷" nil :system t)
    (puthash "\\:princess:" "👸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:princess:" "👸" nil :system t)
    (puthash "\\:japanese_ogre:" "👹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:japanese_ogre:" "👹" nil :system t)
    (puthash "\\:japanese_goblin:" "👺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:japanese_goblin:" "👺" nil :system t)
    (puthash "\\:ghost:" "👻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ghost:" "👻" nil :system t)
    (puthash "\\:angel:" "👼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:angel:" "👼" nil :system t)
    (puthash "\\:alien:" "👽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:alien:" "👽" nil :system t)
    (puthash "\\:space_invader:" "👾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:space_invader:" "👾" nil :system t)
    (puthash "\\:imp:" "👿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:imp:" "👿" nil :system t)
    (puthash "\\:skull:" "💀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skull:" "💀" nil :system t)
    (puthash "\\:information_desk_person:" "💁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:information_desk_person:" "💁" nil :system t)
    (puthash "\\:guardsman:" "💂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:guardsman:" "💂" nil :system t)
    (puthash "\\:dancer:" "💃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dancer:" "💃" nil :system t)
    (puthash "\\:lipstick:" "💄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lipstick:" "💄" nil :system t)
    (puthash "\\:nail_care:" "💅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nail_care:" "💅" nil :system t)
    (puthash "\\:massage:" "💆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:massage:" "💆" nil :system t)
    (puthash "\\:haircut:" "💇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:haircut:" "💇" nil :system t)
    (puthash "\\:barber:" "💈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:barber:" "💈" nil :system t)
    (puthash "\\:syringe:" "💉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:syringe:" "💉" nil :system t)
    (puthash "\\:pill:" "💊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pill:" "💊" nil :system t)
    (puthash "\\:kiss:" "💋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kiss:" "💋" nil :system t)
    (puthash "\\:love_letter:" "💌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:love_letter:" "💌" nil :system t)
    (puthash "\\:ring:" "💍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ring:" "💍" nil :system t)
    (puthash "\\:gem:" "💎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gem:" "💎" nil :system t)
    (puthash "\\:couplekiss:" "💏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:couplekiss:" "💏" nil :system t)
    (puthash "\\:bouquet:" "💐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bouquet:" "💐" nil :system t)
    (puthash "\\:couple_with_heart:" "💑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:couple_with_heart:" "💑" nil :system t)
    (puthash "\\:wedding:" "💒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wedding:" "💒" nil :system t)
    (puthash "\\:heartbeat:" "💓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heartbeat:" "💓" nil :system t)
    (puthash "\\:broken_heart:" "💔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:broken_heart:" "💔" nil :system t)
    (puthash "\\:two_hearts:" "💕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:two_hearts:" "💕" nil :system t)
    (puthash "\\:sparkling_heart:" "💖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sparkling_heart:" "💖" nil :system t)
    (puthash "\\:heartpulse:" "💗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heartpulse:" "💗" nil :system t)
    (puthash "\\:cupid:" "💘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cupid:" "💘" nil :system t)
    (puthash "\\:blue_heart:" "💙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blue_heart:" "💙" nil :system t)
    (puthash "\\:green_heart:" "💚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:green_heart:" "💚" nil :system t)
    (puthash "\\:yellow_heart:" "💛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:yellow_heart:" "💛" nil :system t)
    (puthash "\\:purple_heart:" "💜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:purple_heart:" "💜" nil :system t)
    (puthash "\\:gift_heart:" "💝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gift_heart:" "💝" nil :system t)
    (puthash "\\:revolving_hearts:" "💞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:revolving_hearts:" "💞" nil :system t)
    (puthash "\\:heart_decoration:" "💟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heart_decoration:" "💟" nil :system t)
    (puthash "\\:diamond_shape_with_a_dot_inside:" "💠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:diamond_shape_with_a_dot_inside:" "💠" nil :system t)
    (puthash "\\:bulb:" "💡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bulb:" "💡" nil :system t)
    (puthash "\\:anger:" "💢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:anger:" "💢" nil :system t)
    (puthash "\\:bomb:" "💣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bomb:" "💣" nil :system t)
    (puthash "\\:zzz:" "💤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:zzz:" "💤" nil :system t)
    (puthash "\\:boom:" "💥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boom:" "💥" nil :system t)
    (puthash "\\:sweat_drops:" "💦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sweat_drops:" "💦" nil :system t)
    (puthash "\\:droplet:" "💧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:droplet:" "💧" nil :system t)
    (puthash "\\:dash:" "💨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dash:" "💨" nil :system t)
    (puthash "\\:hankey:" "💩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hankey:" "💩" nil :system t)
    (puthash "\\:muscle:" "💪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:muscle:" "💪" nil :system t)
    (puthash "\\:dizzy:" "💫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dizzy:" "💫" nil :system t)
    (puthash "\\:speech_balloon:" "💬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:speech_balloon:" "💬" nil :system t)
    (puthash "\\:thought_balloon:" "💭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:thought_balloon:" "💭" nil :system t)
    (puthash "\\:white_flower:" "💮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_flower:" "💮" nil :system t)
    (puthash "\\:100:" "💯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:100:" "💯" nil :system t)
    (puthash "\\:moneybag:" "💰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:moneybag:" "💰" nil :system t)
    (puthash "\\:currency_exchange:" "💱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:currency_exchange:" "💱" nil :system t)
    (puthash "\\:heavy_dollar_sign:" "💲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heavy_dollar_sign:" "💲" nil :system t)
    (puthash "\\:credit_card:" "💳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:credit_card:" "💳" nil :system t)
    (puthash "\\:yen:" "💴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:yen:" "💴" nil :system t)
    (puthash "\\:dollar:" "💵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dollar:" "💵" nil :system t)
    (puthash "\\:euro:" "💶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:euro:" "💶" nil :system t)
    (puthash "\\:pound:" "💷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pound:" "💷" nil :system t)
    (puthash "\\:money_with_wings:" "💸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:money_with_wings:" "💸" nil :system t)
    (puthash "\\:chart:" "💹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chart:" "💹" nil :system t)
    (puthash "\\:seat:" "💺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:seat:" "💺" nil :system t)
    (puthash "\\:computer:" "💻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:computer:" "💻" nil :system t)
    (puthash "\\:briefcase:" "💼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:briefcase:" "💼" nil :system t)
    (puthash "\\:minidisc:" "💽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:minidisc:" "💽" nil :system t)
    (puthash "\\:floppy_disk:" "💾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:floppy_disk:" "💾" nil :system t)
    (puthash "\\:cd:" "💿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cd:" "💿" nil :system t)
    (puthash "\\:dvd:" "📀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dvd:" "📀" nil :system t)
    (puthash "\\:file_folder:" "📁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:file_folder:" "📁" nil :system t)
    (puthash "\\:open_file_folder:" "📂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:open_file_folder:" "📂" nil :system t)
    (puthash "\\:page_with_curl:" "📃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:page_with_curl:" "📃" nil :system t)
    (puthash "\\:page_facing_up:" "📄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:page_facing_up:" "📄" nil :system t)
    (puthash "\\:date:" "📅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:date:" "📅" nil :system t)
    (puthash "\\:calendar:" "📆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:calendar:" "📆" nil :system t)
    (puthash "\\:card_index:" "📇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:card_index:" "📇" nil :system t)
    (puthash "\\:chart_with_upwards_trend:" "📈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chart_with_upwards_trend:" "📈" nil :system t)
    (puthash "\\:chart_with_downwards_trend:" "📉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chart_with_downwards_trend:" "📉" nil :system t)
    (puthash "\\:bar_chart:" "📊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bar_chart:" "📊" nil :system t)
    (puthash "\\:clipboard:" "📋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clipboard:" "📋" nil :system t)
    (puthash "\\:pushpin:" "📌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pushpin:" "📌" nil :system t)
    (puthash "\\:round_pushpin:" "📍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:round_pushpin:" "📍" nil :system t)
    (puthash "\\:paperclip:" "📎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:paperclip:" "📎" nil :system t)
    (puthash "\\:straight_ruler:" "📏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:straight_ruler:" "📏" nil :system t)
    (puthash "\\:triangular_ruler:" "📐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:triangular_ruler:" "📐" nil :system t)
    (puthash "\\:bookmark_tabs:" "📑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bookmark_tabs:" "📑" nil :system t)
    (puthash "\\:ledger:" "📒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ledger:" "📒" nil :system t)
    (puthash "\\:notebook:" "📓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:notebook:" "📓" nil :system t)
    (puthash "\\:notebook_with_decorative_cover:" "📔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:notebook_with_decorative_cover:" "📔" nil :system t)
    (puthash "\\:closed_book:" "📕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:closed_book:" "📕" nil :system t)
    (puthash "\\:book:" "📖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:book:" "📖" nil :system t)
    (puthash "\\:green_book:" "📗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:green_book:" "📗" nil :system t)
    (puthash "\\:blue_book:" "📘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blue_book:" "📘" nil :system t)
    (puthash "\\:orange_book:" "📙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:orange_book:" "📙" nil :system t)
    (puthash "\\:books:" "📚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:books:" "📚" nil :system t)
    (puthash "\\:name_badge:" "📛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:name_badge:" "📛" nil :system t)
    (puthash "\\:scroll:" "📜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scroll:" "📜" nil :system t)
    (puthash "\\:memo:" "📝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:memo:" "📝" nil :system t)
    (puthash "\\:telephone_receiver:" "📞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:telephone_receiver:" "📞" nil :system t)
    (puthash "\\:pager:" "📟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pager:" "📟" nil :system t)
    (puthash "\\:fax:" "📠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fax:" "📠" nil :system t)
    (puthash "\\:satellite_antenna:" "📡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:satellite_antenna:" "📡" nil :system t)
    (puthash "\\:satellite:" "📡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:satellite:" "📡" nil :system t)
    (puthash "\\:loudspeaker:" "📢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:loudspeaker:" "📢" nil :system t)
    (puthash "\\:mega:" "📣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mega:" "📣" nil :system t)
    (puthash "\\:outbox_tray:" "📤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:outbox_tray:" "📤" nil :system t)
    (puthash "\\:inbox_tray:" "📥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:inbox_tray:" "📥" nil :system t)
    (puthash "\\:package:" "📦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:package:" "📦" nil :system t)
    (puthash "\\:e-mail:" "📧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:e-mail:" "📧" nil :system t)
    (puthash "\\:incoming_envelope:" "📨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:incoming_envelope:" "📨" nil :system t)
    (puthash "\\:envelope_with_arrow:" "📩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:envelope_with_arrow:" "📩" nil :system t)
    (puthash "\\:mailbox_closed:" "📪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mailbox_closed:" "📪" nil :system t)
    (puthash "\\:mailbox:" "📫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mailbox:" "📫" nil :system t)
    (puthash "\\:mailbox_with_mail:" "📬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mailbox_with_mail:" "📬" nil :system t)
    (puthash "\\:mailbox_with_no_mail:" "📭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mailbox_with_no_mail:" "📭" nil :system t)
    (puthash "\\:postbox:" "📮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:postbox:" "📮" nil :system t)
    (puthash "\\:postal_horn:" "📯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:postal_horn:" "📯" nil :system t)
    (puthash "\\:newspaper:" "📰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:newspaper:" "📰" nil :system t)
    (puthash "\\:iphone:" "📱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:iphone:" "📱" nil :system t)
    (puthash "\\:calling:" "📲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:calling:" "📲" nil :system t)
    (puthash "\\:vibration_mode:" "📳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:vibration_mode:" "📳" nil :system t)
    (puthash "\\:mobile_phone_off:" "📴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mobile_phone_off:" "📴" nil :system t)
    (puthash "\\:no_mobile_phones:" "📵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_mobile_phones:" "📵" nil :system t)
    (puthash "\\:signal_strength:" "📶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:signal_strength:" "📶" nil :system t)
    (puthash "\\:camera:" "📷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:camera:" "📷" nil :system t)
    (puthash "\\:camera_with_flash:" "📸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:camera_with_flash:" "📸" nil :system t)
    (puthash "\\:video_camera:" "📹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:video_camera:" "📹" nil :system t)
    (puthash "\\:tv:" "📺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tv:" "📺" nil :system t)
    (puthash "\\:radio:" "📻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:radio:" "📻" nil :system t)
    (puthash "\\:vhs:" "📼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:vhs:" "📼" nil :system t)
    (puthash "\\:prayer_beads:" "📿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:prayer_beads:" "📿" nil :system t)
    (puthash "\\:twisted_rightwards_arrows:" "🔀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:twisted_rightwards_arrows:" "🔀" nil :system t)
    (puthash "\\:repeat:" "🔁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:repeat:" "🔁" nil :system t)
    (puthash "\\:repeat_one:" "🔂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:repeat_one:" "🔂" nil :system t)
    (puthash "\\:arrows_clockwise:" "🔃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrows_clockwise:" "🔃" nil :system t)
    (puthash "\\:arrows_counterclockwise:" "🔄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrows_counterclockwise:" "🔄" nil :system t)
    (puthash "\\:low_brightness:" "🔅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:low_brightness:" "🔅" nil :system t)
    (puthash "\\:high_brightness:" "🔆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:high_brightness:" "🔆" nil :system t)
    (puthash "\\:mute:" "🔇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mute:" "🔇" nil :system t)
    (puthash "\\:speaker:" "🔈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:speaker:" "🔈" nil :system t)
    (puthash "\\:sound:" "🔉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sound:" "🔉" nil :system t)
    (puthash "\\:loud_sound:" "🔊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:loud_sound:" "🔊" nil :system t)
    (puthash "\\:battery:" "🔋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:battery:" "🔋" nil :system t)
    (puthash "\\:electric_plug:" "🔌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:electric_plug:" "🔌" nil :system t)
    (puthash "\\:mag:" "🔍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mag:" "🔍" nil :system t)
    (puthash "\\:mag_right:" "🔎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mag_right:" "🔎" nil :system t)
    (puthash "\\:lock_with_ink_pen:" "🔏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lock_with_ink_pen:" "🔏" nil :system t)
    (puthash "\\:closed_lock_with_key:" "🔐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:closed_lock_with_key:" "🔐" nil :system t)
    (puthash "\\:key:" "🔑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:key:" "🔑" nil :system t)
    (puthash "\\:lock:" "🔒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lock:" "🔒" nil :system t)
    (puthash "\\:unlock:" "🔓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:unlock:" "🔓" nil :system t)
    (puthash "\\:bell:" "🔔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bell:" "🔔" nil :system t)
    (puthash "\\:no_bell:" "🔕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_bell:" "🔕" nil :system t)
    (puthash "\\:bookmark:" "🔖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bookmark:" "🔖" nil :system t)
    (puthash "\\:link:" "🔗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:link:" "🔗" nil :system t)
    (puthash "\\:radio_button:" "🔘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:radio_button:" "🔘" nil :system t)
    (puthash "\\:back:" "🔙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:back:" "🔙" nil :system t)
    (puthash "\\:end:" "🔚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:end:" "🔚" nil :system t)
    (puthash "\\:on:" "🔛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:on:" "🔛" nil :system t)
    (puthash "\\:soon:" "🔜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:soon:" "🔜" nil :system t)
    (puthash "\\:top:" "🔝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:top:" "🔝" nil :system t)
    (puthash "\\:underage:" "🔞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:underage:" "🔞" nil :system t)
    (puthash "\\:keycap_ten:" "🔟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:keycap_ten:" "🔟" nil :system t)
    (puthash "\\:capital_abcd:" "🔠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:capital_abcd:" "🔠" nil :system t)
    (puthash "\\:abcd:" "🔡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:abcd:" "🔡" nil :system t)
    (puthash "\\:1234:" "🔢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:1234:" "🔢" nil :system t)
    (puthash "\\:symbols:" "🔣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:symbols:" "🔣" nil :system t)
    (puthash "\\:abc:" "🔤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:abc:" "🔤" nil :system t)
    (puthash "\\:fire:" "🔥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fire:" "🔥" nil :system t)
    (puthash "\\:flashlight:" "🔦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flashlight:" "🔦" nil :system t)
    (puthash "\\:wrench:" "🔧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wrench:" "🔧" nil :system t)
    (puthash "\\:hammer:" "🔨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hammer:" "🔨" nil :system t)
    (puthash "\\:nut_and_bolt:" "🔩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nut_and_bolt:" "🔩" nil :system t)
    (puthash "\\:hocho:" "🔪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hocho:" "🔪" nil :system t)
    (puthash "\\:gun:" "🔫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gun:" "🔫" nil :system t)
    (puthash "\\:microscope:" "🔬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:microscope:" "🔬" nil :system t)
    (puthash "\\:telescope:" "🔭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:telescope:" "🔭" nil :system t)
    (puthash "\\:crystal_ball:" "🔮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crystal_ball:" "🔮" nil :system t)
    (puthash "\\:six_pointed_star:" "🔯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:six_pointed_star:" "🔯" nil :system t)
    (puthash "\\:beginner:" "🔰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:beginner:" "🔰" nil :system t)
    (puthash "\\:trident:" "🔱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:trident:" "🔱" nil :system t)
    (puthash "\\:black_square_button:" "🔲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_square_button:" "🔲" nil :system t)
    (puthash "\\:white_square_button:" "🔳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_square_button:" "🔳" nil :system t)
    (puthash "\\:red_circle:" "🔴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:red_circle:" "🔴" nil :system t)
    (puthash "\\:large_blue_circle:" "🔵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_blue_circle:" "🔵" nil :system t)
    (puthash "\\:large_orange_diamond:" "🔶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_orange_diamond:" "🔶" nil :system t)
    (puthash "\\:large_blue_diamond:" "🔷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_blue_diamond:" "🔷" nil :system t)
    (puthash "\\:small_orange_diamond:" "🔸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:small_orange_diamond:" "🔸" nil :system t)
    (puthash "\\:small_blue_diamond:" "🔹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:small_blue_diamond:" "🔹" nil :system t)
    (puthash "\\:small_red_triangle:" "🔺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:small_red_triangle:" "🔺" nil :system t)
    (puthash "\\:small_red_triangle_down:" "🔻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:small_red_triangle_down:" "🔻" nil :system t)
    (puthash "\\:arrow_up_small:" "🔼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_up_small:" "🔼" nil :system t)
    (puthash "\\:arrow_down_small:" "🔽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:arrow_down_small:" "🔽" nil :system t)
    (puthash "\\:kaaba:" "🕋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kaaba:" "🕋" nil :system t)
    (puthash "\\:mosque:" "🕌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mosque:" "🕌" nil :system t)
    (puthash "\\:synagogue:" "🕍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:synagogue:" "🕍" nil :system t)
    (puthash "\\:menorah_with_nine_branches:" "🕎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:menorah_with_nine_branches:" "🕎" nil :system t)
    (puthash "\\:clock1:" "🕐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock1:" "🕐" nil :system t)
    (puthash "\\:clock2:" "🕑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock2:" "🕑" nil :system t)
    (puthash "\\:clock3:" "🕒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock3:" "🕒" nil :system t)
    (puthash "\\:clock4:" "🕓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock4:" "🕓" nil :system t)
    (puthash "\\:clock5:" "🕔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock5:" "🕔" nil :system t)
    (puthash "\\:clock6:" "🕕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock6:" "🕕" nil :system t)
    (puthash "\\:clock7:" "🕖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock7:" "🕖" nil :system t)
    (puthash "\\:clock8:" "🕗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock8:" "🕗" nil :system t)
    (puthash "\\:clock9:" "🕘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock9:" "🕘" nil :system t)
    (puthash "\\:clock10:" "🕙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock10:" "🕙" nil :system t)
    (puthash "\\:clock11:" "🕚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock11:" "🕚" nil :system t)
    (puthash "\\:clock12:" "🕛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock12:" "🕛" nil :system t)
    (puthash "\\:clock130:" "🕜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock130:" "🕜" nil :system t)
    (puthash "\\:clock230:" "🕝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock230:" "🕝" nil :system t)
    (puthash "\\:clock330:" "🕞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock330:" "🕞" nil :system t)
    (puthash "\\:clock430:" "🕟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock430:" "🕟" nil :system t)
    (puthash "\\:clock530:" "🕠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock530:" "🕠" nil :system t)
    (puthash "\\:clock630:" "🕡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock630:" "🕡" nil :system t)
    (puthash "\\:clock730:" "🕢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock730:" "🕢" nil :system t)
    (puthash "\\:clock830:" "🕣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock830:" "🕣" nil :system t)
    (puthash "\\:clock930:" "🕤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock930:" "🕤" nil :system t)
    (puthash "\\:clock1030:" "🕥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock1030:" "🕥" nil :system t)
    (puthash "\\:clock1130:" "🕦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock1130:" "🕦" nil :system t)
    (puthash "\\:clock1230:" "🕧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clock1230:" "🕧" nil :system t)
    (puthash "\\:man_dancing:" "🕺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:man_dancing:" "🕺" nil :system t)
    (puthash "\\:middle_finger:" "🖕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:middle_finger:" "🖕" nil :system t)
    (puthash "\\:spock-hand:" "🖖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:spock-hand:" "🖖" nil :system t)
    (puthash "\\:black_heart:" "🖤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:black_heart:" "🖤" nil :system t)
    (puthash "\\:mount_fuji:" "🗻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mount_fuji:" "🗻" nil :system t)
    (puthash "\\:tokyo_tower:" "🗼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tokyo_tower:" "🗼" nil :system t)
    (puthash "\\:statue_of_liberty:" "🗽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:statue_of_liberty:" "🗽" nil :system t)
    (puthash "\\:japan:" "🗾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:japan:" "🗾" nil :system t)
    (puthash "\\:moyai:" "🗿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:moyai:" "🗿" nil :system t)
    (puthash "\\:grinning:" "😀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:grinning:" "😀" nil :system t)
    (puthash "\\:grin:" "😁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:grin:" "😁" nil :system t)
    (puthash "\\:joy:" "😂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:joy:" "😂" nil :system t)
    (puthash "\\:smiley:" "😃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smiley:" "😃" nil :system t)
    (puthash "\\:smile:" "😄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smile:" "😄" nil :system t)
    (puthash "\\:sweat_smile:" "😅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sweat_smile:" "😅" nil :system t)
    (puthash "\\:laughing:" "😆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:laughing:" "😆" nil :system t)
    (puthash "\\:innocent:" "😇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:innocent:" "😇" nil :system t)
    (puthash "\\:smiling_imp:" "😈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smiling_imp:" "😈" nil :system t)
    (puthash "\\:wink:" "😉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wink:" "😉" nil :system t)
    (puthash "\\:blush:" "😊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blush:" "😊" nil :system t)
    (puthash "\\:yum:" "😋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:yum:" "😋" nil :system t)
    (puthash "\\:relieved:" "😌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:relieved:" "😌" nil :system t)
    (puthash "\\:heart_eyes:" "😍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heart_eyes:" "😍" nil :system t)
    (puthash "\\:sunglasses:" "😎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sunglasses:" "😎" nil :system t)
    (puthash "\\:smirk:" "😏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smirk:" "😏" nil :system t)
    (puthash "\\:neutral_face:" "😐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:neutral_face:" "😐" nil :system t)
    (puthash "\\:expressionless:" "😑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:expressionless:" "😑" nil :system t)
    (puthash "\\:unamused:" "😒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:unamused:" "😒" nil :system t)
    (puthash "\\:sweat:" "😓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sweat:" "😓" nil :system t)
    (puthash "\\:pensive:" "😔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pensive:" "😔" nil :system t)
    (puthash "\\:confused:" "😕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:confused:" "😕" nil :system t)
    (puthash "\\:confounded:" "😖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:confounded:" "😖" nil :system t)
    (puthash "\\:kissing:" "😗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kissing:" "😗" nil :system t)
    (puthash "\\:kissing_heart:" "😘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kissing_heart:" "😘" nil :system t)
    (puthash "\\:kissing_smiling_eyes:" "😙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kissing_smiling_eyes:" "😙" nil :system t)
    (puthash "\\:kissing_closed_eyes:" "😚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kissing_closed_eyes:" "😚" nil :system t)
    (puthash "\\:stuck_out_tongue:" "😛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stuck_out_tongue:" "😛" nil :system t)
    (puthash "\\:stuck_out_tongue_winking_eye:" "😜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stuck_out_tongue_winking_eye:" "😜" nil :system t)
    (puthash "\\:stuck_out_tongue_closed_eyes:" "😝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stuck_out_tongue_closed_eyes:" "😝" nil :system t)
    (puthash "\\:disappointed:" "😞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:disappointed:" "😞" nil :system t)
    (puthash "\\:worried:" "😟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:worried:" "😟" nil :system t)
    (puthash "\\:angry:" "😠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:angry:" "😠" nil :system t)
    (puthash "\\:rage:" "😡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rage:" "😡" nil :system t)
    (puthash "\\:cry:" "😢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cry:" "😢" nil :system t)
    (puthash "\\:persevere:" "😣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:persevere:" "😣" nil :system t)
    (puthash "\\:triumph:" "😤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:triumph:" "😤" nil :system t)
    (puthash "\\:disappointed_relieved:" "😥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:disappointed_relieved:" "😥" nil :system t)
    (puthash "\\:frowning:" "😦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:frowning:" "😦" nil :system t)
    (puthash "\\:anguished:" "😧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:anguished:" "😧" nil :system t)
    (puthash "\\:fearful:" "😨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fearful:" "😨" nil :system t)
    (puthash "\\:weary:" "😩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:weary:" "😩" nil :system t)
    (puthash "\\:sleepy:" "😪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sleepy:" "😪" nil :system t)
    (puthash "\\:tired_face:" "😫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tired_face:" "😫" nil :system t)
    (puthash "\\:grimacing:" "😬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:grimacing:" "😬" nil :system t)
    (puthash "\\:sob:" "😭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sob:" "😭" nil :system t)
    (puthash "\\:open_mouth:" "😮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:open_mouth:" "😮" nil :system t)
    (puthash "\\:hushed:" "😯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hushed:" "😯" nil :system t)
    (puthash "\\:cold_sweat:" "😰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cold_sweat:" "😰" nil :system t)
    (puthash "\\:scream:" "😱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scream:" "😱" nil :system t)
    (puthash "\\:astonished:" "😲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:astonished:" "😲" nil :system t)
    (puthash "\\:flushed:" "😳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flushed:" "😳" nil :system t)
    (puthash "\\:sleeping:" "😴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sleeping:" "😴" nil :system t)
    (puthash "\\:dizzy_face:" "😵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dizzy_face:" "😵" nil :system t)
    (puthash "\\:no_mouth:" "😶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_mouth:" "😶" nil :system t)
    (puthash "\\:mask:" "😷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mask:" "😷" nil :system t)
    (puthash "\\:smile_cat:" "😸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smile_cat:" "😸" nil :system t)
    (puthash "\\:joy_cat:" "😹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:joy_cat:" "😹" nil :system t)
    (puthash "\\:smiley_cat:" "😺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smiley_cat:" "😺" nil :system t)
    (puthash "\\:heart_eyes_cat:" "😻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:heart_eyes_cat:" "😻" nil :system t)
    (puthash "\\:smirk_cat:" "😼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smirk_cat:" "😼" nil :system t)
    (puthash "\\:kissing_cat:" "😽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kissing_cat:" "😽" nil :system t)
    (puthash "\\:pouting_cat:" "😾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pouting_cat:" "😾" nil :system t)
    (puthash "\\:crying_cat_face:" "😿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crying_cat_face:" "😿" nil :system t)
    (puthash "\\:scream_cat:" "🙀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scream_cat:" "🙀" nil :system t)
    (puthash "\\:slightly_frowning_face:" "🙁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:slightly_frowning_face:" "🙁" nil :system t)
    (puthash "\\:slightly_smiling_face:" "🙂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:slightly_smiling_face:" "🙂" nil :system t)
    (puthash "\\:upside_down_face:" "🙃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:upside_down_face:" "🙃" nil :system t)
    (puthash "\\:face_with_rolling_eyes:" "🙄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_rolling_eyes:" "🙄" nil :system t)
    (puthash "\\:no_good:" "🙅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_good:" "🙅" nil :system t)
    (puthash "\\:ok_woman:" "🙆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ok_woman:" "🙆" nil :system t)
    (puthash "\\:bow:" "🙇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bow:" "🙇" nil :system t)
    (puthash "\\:see_no_evil:" "🙈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:see_no_evil:" "🙈" nil :system t)
    (puthash "\\:hear_no_evil:" "🙉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hear_no_evil:" "🙉" nil :system t)
    (puthash "\\:speak_no_evil:" "🙊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:speak_no_evil:" "🙊" nil :system t)
    (puthash "\\:raising_hand:" "🙋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:raising_hand:" "🙋" nil :system t)
    (puthash "\\:raised_hands:" "🙌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:raised_hands:" "🙌" nil :system t)
    (puthash "\\:person_frowning:" "🙍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_frowning:" "🙍" nil :system t)
    (puthash "\\:person_with_pouting_face:" "🙎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_with_pouting_face:" "🙎" nil :system t)
    (puthash "\\:pray:" "🙏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pray:" "🙏" nil :system t)
    (puthash "\\:rocket:" "🚀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rocket:" "🚀" nil :system t)
    (puthash "\\:helicopter:" "🚁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:helicopter:" "🚁" nil :system t)
    (puthash "\\:steam_locomotive:" "🚂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:steam_locomotive:" "🚂" nil :system t)
    (puthash "\\:railway_car:" "🚃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:railway_car:" "🚃" nil :system t)
    (puthash "\\:bullettrain_side:" "🚄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bullettrain_side:" "🚄" nil :system t)
    (puthash "\\:bullettrain_front:" "🚅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bullettrain_front:" "🚅" nil :system t)
    (puthash "\\:train2:" "🚆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:train2:" "🚆" nil :system t)
    (puthash "\\:metro:" "🚇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:metro:" "🚇" nil :system t)
    (puthash "\\:light_rail:" "🚈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:light_rail:" "🚈" nil :system t)
    (puthash "\\:station:" "🚉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:station:" "🚉" nil :system t)
    (puthash "\\:tram:" "🚊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tram:" "🚊" nil :system t)
    (puthash "\\:train:" "🚋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:train:" "🚋" nil :system t)
    (puthash "\\:bus:" "🚌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bus:" "🚌" nil :system t)
    (puthash "\\:oncoming_bus:" "🚍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:oncoming_bus:" "🚍" nil :system t)
    (puthash "\\:trolleybus:" "🚎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:trolleybus:" "🚎" nil :system t)
    (puthash "\\:busstop:" "🚏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:busstop:" "🚏" nil :system t)
    (puthash "\\:minibus:" "🚐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:minibus:" "🚐" nil :system t)
    (puthash "\\:ambulance:" "🚑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ambulance:" "🚑" nil :system t)
    (puthash "\\:fire_engine:" "🚒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fire_engine:" "🚒" nil :system t)
    (puthash "\\:police_car:" "🚓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:police_car:" "🚓" nil :system t)
    (puthash "\\:oncoming_police_car:" "🚔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:oncoming_police_car:" "🚔" nil :system t)
    (puthash "\\:taxi:" "🚕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:taxi:" "🚕" nil :system t)
    (puthash "\\:oncoming_taxi:" "🚖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:oncoming_taxi:" "🚖" nil :system t)
    (puthash "\\:car:" "🚗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:car:" "🚗" nil :system t)
    (puthash "\\:oncoming_automobile:" "🚘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:oncoming_automobile:" "🚘" nil :system t)
    (puthash "\\:blue_car:" "🚙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blue_car:" "🚙" nil :system t)
    (puthash "\\:truck:" "🚚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:truck:" "🚚" nil :system t)
    (puthash "\\:articulated_lorry:" "🚛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:articulated_lorry:" "🚛" nil :system t)
    (puthash "\\:tractor:" "🚜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tractor:" "🚜" nil :system t)
    (puthash "\\:monorail:" "🚝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:monorail:" "🚝" nil :system t)
    (puthash "\\:mountain_railway:" "🚞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mountain_railway:" "🚞" nil :system t)
    (puthash "\\:suspension_railway:" "🚟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:suspension_railway:" "🚟" nil :system t)
    (puthash "\\:mountain_cableway:" "🚠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mountain_cableway:" "🚠" nil :system t)
    (puthash "\\:aerial_tramway:" "🚡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:aerial_tramway:" "🚡" nil :system t)
    (puthash "\\:ship:" "🚢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ship:" "🚢" nil :system t)
    (puthash "\\:rowboat:" "🚣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rowboat:" "🚣" nil :system t)
    (puthash "\\:speedboat:" "🚤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:speedboat:" "🚤" nil :system t)
    (puthash "\\:traffic_light:" "🚥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:traffic_light:" "🚥" nil :system t)
    (puthash "\\:vertical_traffic_light:" "🚦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:vertical_traffic_light:" "🚦" nil :system t)
    (puthash "\\:construction:" "🚧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:construction:" "🚧" nil :system t)
    (puthash "\\:rotating_light:" "🚨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rotating_light:" "🚨" nil :system t)
    (puthash "\\:triangular_flag_on_post:" "🚩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:triangular_flag_on_post:" "🚩" nil :system t)
    (puthash "\\:door:" "🚪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:door:" "🚪" nil :system t)
    (puthash "\\:no_entry_sign:" "🚫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_entry_sign:" "🚫" nil :system t)
    (puthash "\\:smoking:" "🚬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smoking:" "🚬" nil :system t)
    (puthash "\\:no_smoking:" "🚭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_smoking:" "🚭" nil :system t)
    (puthash "\\:put_litter_in_its_place:" "🚮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:put_litter_in_its_place:" "🚮" nil :system t)
    (puthash "\\:do_not_litter:" "🚯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:do_not_litter:" "🚯" nil :system t)
    (puthash "\\:potable_water:" "🚰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:potable_water:" "🚰" nil :system t)
    (puthash "\\:non-potable_water:" "🚱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:non-potable_water:" "🚱" nil :system t)
    (puthash "\\:bike:" "🚲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bike:" "🚲" nil :system t)
    (puthash "\\:no_bicycles:" "🚳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_bicycles:" "🚳" nil :system t)
    (puthash "\\:bicyclist:" "🚴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bicyclist:" "🚴" nil :system t)
    (puthash "\\:mountain_bicyclist:" "🚵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mountain_bicyclist:" "🚵" nil :system t)
    (puthash "\\:walking:" "🚶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:walking:" "🚶" nil :system t)
    (puthash "\\:no_pedestrians:" "🚷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:no_pedestrians:" "🚷" nil :system t)
    (puthash "\\:children_crossing:" "🚸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:children_crossing:" "🚸" nil :system t)
    (puthash "\\:mens:" "🚹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mens:" "🚹" nil :system t)
    (puthash "\\:womens:" "🚺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:womens:" "🚺" nil :system t)
    (puthash "\\:restroom:" "🚻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:restroom:" "🚻" nil :system t)
    (puthash "\\:baby_symbol:" "🚼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baby_symbol:" "🚼" nil :system t)
    (puthash "\\:toilet:" "🚽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:toilet:" "🚽" nil :system t)
    (puthash "\\:wc:" "🚾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wc:" "🚾" nil :system t)
    (puthash "\\:shower:" "🚿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shower:" "🚿" nil :system t)
    (puthash "\\:bath:" "🛀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bath:" "🛀" nil :system t)
    (puthash "\\:bathtub:" "🛁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bathtub:" "🛁" nil :system t)
    (puthash "\\:passport_control:" "🛂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:passport_control:" "🛂" nil :system t)
    (puthash "\\:customs:" "🛃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:customs:" "🛃" nil :system t)
    (puthash "\\:baggage_claim:" "🛄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baggage_claim:" "🛄" nil :system t)
    (puthash "\\:left_luggage:" "🛅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:left_luggage:" "🛅" nil :system t)
    (puthash "\\:sleeping_accommodation:" "🛌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sleeping_accommodation:" "🛌" nil :system t)
    (puthash "\\:place_of_worship:" "🛐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:place_of_worship:" "🛐" nil :system t)
    (puthash "\\:octagonal_sign:" "🛑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:octagonal_sign:" "🛑" nil :system t)
    (puthash "\\:shopping_trolley:" "🛒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shopping_trolley:" "🛒" nil :system t)
    (puthash "\\:hindu_temple:" "🛕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hindu_temple:" "🛕" nil :system t)
    (puthash "\\:hut:" "🛖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hut:" "🛖" nil :system t)
    (puthash "\\:elevator:" "🛗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:elevator:" "🛗" nil :system t)
    (puthash "\\:airplane_departure:" "🛫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:airplane_departure:" "🛫" nil :system t)
    (puthash "\\:airplane_arriving:" "🛬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:airplane_arriving:" "🛬" nil :system t)
    (puthash "\\:scooter:" "🛴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scooter:" "🛴" nil :system t)
    (puthash "\\:motor_scooter:" "🛵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:motor_scooter:" "🛵" nil :system t)
    (puthash "\\:canoe:" "🛶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:canoe:" "🛶" nil :system t)
    (puthash "\\:sled:" "🛷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sled:" "🛷" nil :system t)
    (puthash "\\:flying_saucer:" "🛸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flying_saucer:" "🛸" nil :system t)
    (puthash "\\:skateboard:" "🛹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skateboard:" "🛹" nil :system t)
    (puthash "\\:auto_rickshaw:" "🛺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:auto_rickshaw:" "🛺" nil :system t)
    (puthash "\\:pickup_truck:" "🛻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pickup_truck:" "🛻" nil :system t)
    (puthash "\\:roller_skate:" "🛼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:roller_skate:" "🛼" nil :system t)
    (puthash "\\:large_orange_circle:" "🟠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_orange_circle:" "🟠" nil :system t)
    (puthash "\\:large_yellow_circle:" "🟡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_yellow_circle:" "🟡" nil :system t)
    (puthash "\\:large_green_circle:" "🟢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_green_circle:" "🟢" nil :system t)
    (puthash "\\:large_purple_circle:" "🟣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_purple_circle:" "🟣" nil :system t)
    (puthash "\\:large_brown_circle:" "🟤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_brown_circle:" "🟤" nil :system t)
    (puthash "\\:large_red_square:" "🟥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_red_square:" "🟥" nil :system t)
    (puthash "\\:large_blue_square:" "🟦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_blue_square:" "🟦" nil :system t)
    (puthash "\\:large_orange_square:" "🟧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_orange_square:" "🟧" nil :system t)
    (puthash "\\:large_yellow_square:" "🟨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_yellow_square:" "🟨" nil :system t)
    (puthash "\\:large_green_square:" "🟩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_green_square:" "🟩" nil :system t)
    (puthash "\\:large_purple_square:" "🟪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_purple_square:" "🟪" nil :system t)
    (puthash "\\:large_brown_square:" "🟫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:large_brown_square:" "🟫" nil :system t)
    (puthash "\\:pinched_fingers:" "🤌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pinched_fingers:" "🤌" nil :system t)
    (puthash "\\:white_heart:" "🤍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:white_heart:" "🤍" nil :system t)
    (puthash "\\:brown_heart:" "🤎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:brown_heart:" "🤎" nil :system t)
    (puthash "\\:pinching_hand:" "🤏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pinching_hand:" "🤏" nil :system t)
    (puthash "\\:zipper_mouth_face:" "🤐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:zipper_mouth_face:" "🤐" nil :system t)
    (puthash "\\:money_mouth_face:" "🤑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:money_mouth_face:" "🤑" nil :system t)
    (puthash "\\:face_with_thermometer:" "🤒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_thermometer:" "🤒" nil :system t)
    (puthash "\\:nerd_face:" "🤓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nerd_face:" "🤓" nil :system t)
    (puthash "\\:thinking_face:" "🤔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:thinking_face:" "🤔" nil :system t)
    (puthash "\\:face_with_head_bandage:" "🤕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_head_bandage:" "🤕" nil :system t)
    (puthash "\\:robot_face:" "🤖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:robot_face:" "🤖" nil :system t)
    (puthash "\\:hugging_face:" "🤗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hugging_face:" "🤗" nil :system t)
    (puthash "\\:the_horns:" "🤘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:the_horns:" "🤘" nil :system t)
    (puthash "\\:call_me_hand:" "🤙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:call_me_hand:" "🤙" nil :system t)
    (puthash "\\:raised_back_of_hand:" "🤚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:raised_back_of_hand:" "🤚" nil :system t)
    (puthash "\\:left-facing_fist:" "🤛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:left-facing_fist:" "🤛" nil :system t)
    (puthash "\\:right-facing_fist:" "🤜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:right-facing_fist:" "🤜" nil :system t)
    (puthash "\\:handshake:" "🤝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:handshake:" "🤝" nil :system t)
    (puthash "\\:crossed_fingers:" "🤞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crossed_fingers:" "🤞" nil :system t)
    (puthash "\\:i_love_you_hand_sign:" "🤟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:i_love_you_hand_sign:" "🤟" nil :system t)
    (puthash "\\:face_with_cowboy_hat:" "🤠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_cowboy_hat:" "🤠" nil :system t)
    (puthash "\\:clown_face:" "🤡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clown_face:" "🤡" nil :system t)
    (puthash "\\:nauseated_face:" "🤢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nauseated_face:" "🤢" nil :system t)
    (puthash "\\:rolling_on_the_floor_laughing:" "🤣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rolling_on_the_floor_laughing:" "🤣" nil :system t)
    (puthash "\\:drooling_face:" "🤤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:drooling_face:" "🤤" nil :system t)
    (puthash "\\:lying_face:" "🤥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lying_face:" "🤥" nil :system t)
    (puthash "\\:face_palm:" "🤦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_palm:" "🤦" nil :system t)
    (puthash "\\:sneezing_face:" "🤧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sneezing_face:" "🤧" nil :system t)
    (puthash "\\:face_with_raised_eyebrow:" "🤨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_raised_eyebrow:" "🤨" nil :system t)
    (puthash "\\:star-struck:" "🤩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:star-struck:" "🤩" nil :system t)
    (puthash "\\:zany_face:" "🤪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:zany_face:" "🤪" nil :system t)
    (puthash "\\:shushing_face:" "🤫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shushing_face:" "🤫" nil :system t)
    (puthash "\\:face_with_symbols_on_mouth:" "🤬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_symbols_on_mouth:" "🤬" nil :system t)
    (puthash "\\:face_with_hand_over_mouth:" "🤭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_hand_over_mouth:" "🤭" nil :system t)
    (puthash "\\:face_vomiting:" "🤮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_vomiting:" "🤮" nil :system t)
    (puthash "\\:exploding_head:" "🤯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:exploding_head:" "🤯" nil :system t)
    (puthash "\\:pregnant_woman:" "🤰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pregnant_woman:" "🤰" nil :system t)
    (puthash "\\:breast-feeding:" "🤱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:breast-feeding:" "🤱" nil :system t)
    (puthash "\\:palms_up_together:" "🤲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:palms_up_together:" "🤲" nil :system t)
    (puthash "\\:selfie:" "🤳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:selfie:" "🤳" nil :system t)
    (puthash "\\:prince:" "🤴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:prince:" "🤴" nil :system t)
    (puthash "\\:person_in_tuxedo:" "🤵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_in_tuxedo:" "🤵" nil :system t)
    (puthash "\\:mrs_claus:" "🤶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mrs_claus:" "🤶" nil :system t)
    (puthash "\\:shrug:" "🤷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shrug:" "🤷" nil :system t)
    (puthash "\\:person_doing_cartwheel:" "🤸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_doing_cartwheel:" "🤸" nil :system t)
    (puthash "\\:juggling:" "🤹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:juggling:" "🤹" nil :system t)
    (puthash "\\:fencer:" "🤺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fencer:" "🤺" nil :system t)
    (puthash "\\:wrestlers:" "🤼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wrestlers:" "🤼" nil :system t)
    (puthash "\\:water_polo:" "🤽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:water_polo:" "🤽" nil :system t)
    (puthash "\\:handball:" "🤾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:handball:" "🤾" nil :system t)
    (puthash "\\:diving_mask:" "🤿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:diving_mask:" "🤿" nil :system t)
    (puthash "\\:wilted_flower:" "🥀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wilted_flower:" "🥀" nil :system t)
    (puthash "\\:drum_with_drumsticks:" "🥁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:drum_with_drumsticks:" "🥁" nil :system t)
    (puthash "\\:clinking_glasses:" "🥂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:clinking_glasses:" "🥂" nil :system t)
    (puthash "\\:tumbler_glass:" "🥃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tumbler_glass:" "🥃" nil :system t)
    (puthash "\\:spoon:" "🥄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:spoon:" "🥄" nil :system t)
    (puthash "\\:goal_net:" "🥅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:goal_net:" "🥅" nil :system t)
    (puthash "\\:first_place_medal:" "🥇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:first_place_medal:" "🥇" nil :system t)
    (puthash "\\:second_place_medal:" "🥈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:second_place_medal:" "🥈" nil :system t)
    (puthash "\\:third_place_medal:" "🥉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:third_place_medal:" "🥉" nil :system t)
    (puthash "\\:boxing_glove:" "🥊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boxing_glove:" "🥊" nil :system t)
    (puthash "\\:martial_arts_uniform:" "🥋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:martial_arts_uniform:" "🥋" nil :system t)
    (puthash "\\:curling_stone:" "🥌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:curling_stone:" "🥌" nil :system t)
    (puthash "\\:lacrosse:" "🥍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lacrosse:" "🥍" nil :system t)
    (puthash "\\:softball:" "🥎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:softball:" "🥎" nil :system t)
    (puthash "\\:flying_disc:" "🥏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flying_disc:" "🥏" nil :system t)
    (puthash "\\:croissant:" "🥐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:croissant:" "🥐" nil :system t)
    (puthash "\\:avocado:" "🥑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:avocado:" "🥑" nil :system t)
    (puthash "\\:cucumber:" "🥒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cucumber:" "🥒" nil :system t)
    (puthash "\\:bacon:" "🥓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bacon:" "🥓" nil :system t)
    (puthash "\\:potato:" "🥔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:potato:" "🥔" nil :system t)
    (puthash "\\:carrot:" "🥕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:carrot:" "🥕" nil :system t)
    (puthash "\\:baguette_bread:" "🥖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:baguette_bread:" "🥖" nil :system t)
    (puthash "\\:green_salad:" "🥗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:green_salad:" "🥗" nil :system t)
    (puthash "\\:shallow_pan_of_food:" "🥘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shallow_pan_of_food:" "🥘" nil :system t)
    (puthash "\\:stuffed_flatbread:" "🥙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stuffed_flatbread:" "🥙" nil :system t)
    (puthash "\\:egg:" "🥚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:egg:" "🥚" nil :system t)
    (puthash "\\:glass_of_milk:" "🥛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:glass_of_milk:" "🥛" nil :system t)
    (puthash "\\:peanuts:" "🥜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:peanuts:" "🥜" nil :system t)
    (puthash "\\:kiwifruit:" "🥝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kiwifruit:" "🥝" nil :system t)
    (puthash "\\:pancakes:" "🥞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pancakes:" "🥞" nil :system t)
    (puthash "\\:dumpling:" "🥟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dumpling:" "🥟" nil :system t)
    (puthash "\\:fortune_cookie:" "🥠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fortune_cookie:" "🥠" nil :system t)
    (puthash "\\:takeout_box:" "🥡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:takeout_box:" "🥡" nil :system t)
    (puthash "\\:chopsticks:" "🥢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chopsticks:" "🥢" nil :system t)
    (puthash "\\:bowl_with_spoon:" "🥣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bowl_with_spoon:" "🥣" nil :system t)
    (puthash "\\:cup_with_straw:" "🥤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cup_with_straw:" "🥤" nil :system t)
    (puthash "\\:coconut:" "🥥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:coconut:" "🥥" nil :system t)
    (puthash "\\:broccoli:" "🥦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:broccoli:" "🥦" nil :system t)
    (puthash "\\:pie:" "🥧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pie:" "🥧" nil :system t)
    (puthash "\\:pretzel:" "🥨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pretzel:" "🥨" nil :system t)
    (puthash "\\:cut_of_meat:" "🥩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cut_of_meat:" "🥩" nil :system t)
    (puthash "\\:sandwich:" "🥪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sandwich:" "🥪" nil :system t)
    (puthash "\\:canned_food:" "🥫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:canned_food:" "🥫" nil :system t)
    (puthash "\\:leafy_green:" "🥬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:leafy_green:" "🥬" nil :system t)
    (puthash "\\:mango:" "🥭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mango:" "🥭" nil :system t)
    (puthash "\\:moon_cake:" "🥮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:moon_cake:" "🥮" nil :system t)
    (puthash "\\:bagel:" "🥯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bagel:" "🥯" nil :system t)
    (puthash "\\:smiling_face_with_3_hearts:" "🥰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smiling_face_with_3_hearts:" "🥰" nil :system t)
    (puthash "\\:yawning_face:" "🥱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:yawning_face:" "🥱" nil :system t)
    (puthash "\\:smiling_face_with_tear:" "🥲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:smiling_face_with_tear:" "🥲" nil :system t)
    (puthash "\\:partying_face:" "🥳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:partying_face:" "🥳" nil :system t)
    (puthash "\\:woozy_face:" "🥴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:woozy_face:" "🥴" nil :system t)
    (puthash "\\:hot_face:" "🥵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hot_face:" "🥵" nil :system t)
    (puthash "\\:cold_face:" "🥶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cold_face:" "🥶" nil :system t)
    (puthash "\\:ninja:" "🥷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ninja:" "🥷" nil :system t)
    (puthash "\\:disguised_face:" "🥸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:disguised_face:" "🥸" nil :system t)
    (puthash "\\:pleading_face:" "🥺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pleading_face:" "🥺" nil :system t)
    (puthash "\\:sari:" "🥻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sari:" "🥻" nil :system t)
    (puthash "\\:lab_coat:" "🥼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lab_coat:" "🥼" nil :system t)
    (puthash "\\:goggles:" "🥽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:goggles:" "🥽" nil :system t)
    (puthash "\\:hiking_boot:" "🥾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hiking_boot:" "🥾" nil :system t)
    (puthash "\\:womans_flat_shoe:" "🥿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:womans_flat_shoe:" "🥿" nil :system t)
    (puthash "\\:crab:" "🦀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:crab:" "🦀" nil :system t)
    (puthash "\\:lion_face:" "🦁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lion_face:" "🦁" nil :system t)
    (puthash "\\:scorpion:" "🦂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scorpion:" "🦂" nil :system t)
    (puthash "\\:turkey:" "🦃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:turkey:" "🦃" nil :system t)
    (puthash "\\:unicorn_face:" "🦄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:unicorn_face:" "🦄" nil :system t)
    (puthash "\\:eagle:" "🦅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:eagle:" "🦅" nil :system t)
    (puthash "\\:duck:" "🦆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:duck:" "🦆" nil :system t)
    (puthash "\\:bat:" "🦇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bat:" "🦇" nil :system t)
    (puthash "\\:shark:" "🦈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shark:" "🦈" nil :system t)
    (puthash "\\:owl:" "🦉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:owl:" "🦉" nil :system t)
    (puthash "\\:fox_face:" "🦊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fox_face:" "🦊" nil :system t)
    (puthash "\\:butterfly:" "🦋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:butterfly:" "🦋" nil :system t)
    (puthash "\\:deer:" "🦌" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:deer:" "🦌" nil :system t)
    (puthash "\\:gorilla:" "🦍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gorilla:" "🦍" nil :system t)
    (puthash "\\:lizard:" "🦎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lizard:" "🦎" nil :system t)
    (puthash "\\:rhinoceros:" "🦏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rhinoceros:" "🦏" nil :system t)
    (puthash "\\:shrimp:" "🦐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shrimp:" "🦐" nil :system t)
    (puthash "\\:squid:" "🦑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:squid:" "🦑" nil :system t)
    (puthash "\\:giraffe_face:" "🦒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:giraffe_face:" "🦒" nil :system t)
    (puthash "\\:zebra_face:" "🦓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:zebra_face:" "🦓" nil :system t)
    (puthash "\\:hedgehog:" "🦔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hedgehog:" "🦔" nil :system t)
    (puthash "\\:sauropod:" "🦕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sauropod:" "🦕" nil :system t)
    (puthash "\\:t-rex:" "🦖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:t-rex:" "🦖" nil :system t)
    (puthash "\\:cricket:" "🦗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cricket:" "🦗" nil :system t)
    (puthash "\\:kangaroo:" "🦘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kangaroo:" "🦘" nil :system t)
    (puthash "\\:llama:" "🦙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:llama:" "🦙" nil :system t)
    (puthash "\\:peacock:" "🦚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:peacock:" "🦚" nil :system t)
    (puthash "\\:hippopotamus:" "🦛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hippopotamus:" "🦛" nil :system t)
    (puthash "\\:parrot:" "🦜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:parrot:" "🦜" nil :system t)
    (puthash "\\:raccoon:" "🦝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:raccoon:" "🦝" nil :system t)
    (puthash "\\:lobster:" "🦞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lobster:" "🦞" nil :system t)
    (puthash "\\:mosquito:" "🦟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mosquito:" "🦟" nil :system t)
    (puthash "\\:microbe:" "🦠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:microbe:" "🦠" nil :system t)
    (puthash "\\:badger:" "🦡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:badger:" "🦡" nil :system t)
    (puthash "\\:swan:" "🦢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:swan:" "🦢" nil :system t)
    (puthash "\\:mammoth:" "🦣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mammoth:" "🦣" nil :system t)
    (puthash "\\:dodo:" "🦤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dodo:" "🦤" nil :system t)
    (puthash "\\:sloth:" "🦥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sloth:" "🦥" nil :system t)
    (puthash "\\:otter:" "🦦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:otter:" "🦦" nil :system t)
    (puthash "\\:orangutan:" "🦧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:orangutan:" "🦧" nil :system t)
    (puthash "\\:skunk:" "🦨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:skunk:" "🦨" nil :system t)
    (puthash "\\:flamingo:" "🦩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flamingo:" "🦩" nil :system t)
    (puthash "\\:oyster:" "🦪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:oyster:" "🦪" nil :system t)
    (puthash "\\:beaver:" "🦫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:beaver:" "🦫" nil :system t)
    (puthash "\\:bison:" "🦬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bison:" "🦬" nil :system t)
    (puthash "\\:seal:" "🦭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:seal:" "🦭" nil :system t)
    (puthash "\\:guide_dog:" "🦮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:guide_dog:" "🦮" nil :system t)
    (puthash "\\:probing_cane:" "🦯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:probing_cane:" "🦯" nil :system t)
    (puthash "\\:bone:" "🦴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bone:" "🦴" nil :system t)
    (puthash "\\:leg:" "🦵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:leg:" "🦵" nil :system t)
    (puthash "\\:foot:" "🦶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:foot:" "🦶" nil :system t)
    (puthash "\\:tooth:" "🦷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tooth:" "🦷" nil :system t)
    (puthash "\\:superhero:" "🦸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:superhero:" "🦸" nil :system t)
    (puthash "\\:supervillain:" "🦹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:supervillain:" "🦹" nil :system t)
    (puthash "\\:safety_vest:" "🦺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:safety_vest:" "🦺" nil :system t)
    (puthash "\\:ear_with_hearing_aid:" "🦻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ear_with_hearing_aid:" "🦻" nil :system t)
    (puthash "\\:motorized_wheelchair:" "🦼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:motorized_wheelchair:" "🦼" nil :system t)
    (puthash "\\:manual_wheelchair:" "🦽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:manual_wheelchair:" "🦽" nil :system t)
    (puthash "\\:mechanical_arm:" "🦾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mechanical_arm:" "🦾" nil :system t)
    (puthash "\\:mechanical_leg:" "🦿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mechanical_leg:" "🦿" nil :system t)
    (puthash "\\:cheese_wedge:" "🧀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cheese_wedge:" "🧀" nil :system t)
    (puthash "\\:cupcake:" "🧁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cupcake:" "🧁" nil :system t)
    (puthash "\\:salt:" "🧂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:salt:" "🧂" nil :system t)
    (puthash "\\:beverage_box:" "🧃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:beverage_box:" "🧃" nil :system t)
    (puthash "\\:garlic:" "🧄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:garlic:" "🧄" nil :system t)
    (puthash "\\:onion:" "🧅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:onion:" "🧅" nil :system t)
    (puthash "\\:falafel:" "🧆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:falafel:" "🧆" nil :system t)
    (puthash "\\:waffle:" "🧇" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:waffle:" "🧇" nil :system t)
    (puthash "\\:butter:" "🧈" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:butter:" "🧈" nil :system t)
    (puthash "\\:mate_drink:" "🧉" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mate_drink:" "🧉" nil :system t)
    (puthash "\\:ice_cube:" "🧊" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ice_cube:" "🧊" nil :system t)
    (puthash "\\:bubble_tea:" "🧋" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bubble_tea:" "🧋" nil :system t)
    (puthash "\\:standing_person:" "🧍" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:standing_person:" "🧍" nil :system t)
    (puthash "\\:kneeling_person:" "🧎" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kneeling_person:" "🧎" nil :system t)
    (puthash "\\:deaf_person:" "🧏" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:deaf_person:" "🧏" nil :system t)
    (puthash "\\:face_with_monocle:" "🧐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:face_with_monocle:" "🧐" nil :system t)
    (puthash "\\:adult:" "🧑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:adult:" "🧑" nil :system t)
    (puthash "\\:child:" "🧒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:child:" "🧒" nil :system t)
    (puthash "\\:older_adult:" "🧓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:older_adult:" "🧓" nil :system t)
    (puthash "\\:bearded_person:" "🧔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bearded_person:" "🧔" nil :system t)
    (puthash "\\:person_with_headscarf:" "🧕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_with_headscarf:" "🧕" nil :system t)
    (puthash "\\:person_in_steamy_room:" "🧖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_in_steamy_room:" "🧖" nil :system t)
    (puthash "\\:person_climbing:" "🧗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_climbing:" "🧗" nil :system t)
    (puthash "\\:person_in_lotus_position:" "🧘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:person_in_lotus_position:" "🧘" nil :system t)
    (puthash "\\:mage:" "🧙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mage:" "🧙" nil :system t)
    (puthash "\\:fairy:" "🧚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fairy:" "🧚" nil :system t)
    (puthash "\\:vampire:" "🧛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:vampire:" "🧛" nil :system t)
    (puthash "\\:merperson:" "🧜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:merperson:" "🧜" nil :system t)
    (puthash "\\:elf:" "🧝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:elf:" "🧝" nil :system t)
    (puthash "\\:genie:" "🧞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:genie:" "🧞" nil :system t)
    (puthash "\\:zombie:" "🧟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:zombie:" "🧟" nil :system t)
    (puthash "\\:brain:" "🧠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:brain:" "🧠" nil :system t)
    (puthash "\\:orange_heart:" "🧡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:orange_heart:" "🧡" nil :system t)
    (puthash "\\:billed_cap:" "🧢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:billed_cap:" "🧢" nil :system t)
    (puthash "\\:scarf:" "🧣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:scarf:" "🧣" nil :system t)
    (puthash "\\:gloves:" "🧤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:gloves:" "🧤" nil :system t)
    (puthash "\\:coat:" "🧥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:coat:" "🧥" nil :system t)
    (puthash "\\:socks:" "🧦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:socks:" "🧦" nil :system t)
    (puthash "\\:red_envelope:" "🧧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:red_envelope:" "🧧" nil :system t)
    (puthash "\\:firecracker:" "🧨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:firecracker:" "🧨" nil :system t)
    (puthash "\\:jigsaw:" "🧩" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:jigsaw:" "🧩" nil :system t)
    (puthash "\\:test_tube:" "🧪" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:test_tube:" "🧪" nil :system t)
    (puthash "\\:petri_dish:" "🧫" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:petri_dish:" "🧫" nil :system t)
    (puthash "\\:dna:" "🧬" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:dna:" "🧬" nil :system t)
    (puthash "\\:compass:" "🧭" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:compass:" "🧭" nil :system t)
    (puthash "\\:abacus:" "🧮" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:abacus:" "🧮" nil :system t)
    (puthash "\\:fire_extinguisher:" "🧯" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fire_extinguisher:" "🧯" nil :system t)
    (puthash "\\:toolbox:" "🧰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:toolbox:" "🧰" nil :system t)
    (puthash "\\:bricks:" "🧱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bricks:" "🧱" nil :system t)
    (puthash "\\:magnet:" "🧲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:magnet:" "🧲" nil :system t)
    (puthash "\\:luggage:" "🧳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:luggage:" "🧳" nil :system t)
    (puthash "\\:lotion_bottle:" "🧴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lotion_bottle:" "🧴" nil :system t)
    (puthash "\\:thread:" "🧵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:thread:" "🧵" nil :system t)
    (puthash "\\:yarn:" "🧶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:yarn:" "🧶" nil :system t)
    (puthash "\\:safety_pin:" "🧷" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:safety_pin:" "🧷" nil :system t)
    (puthash "\\:teddy_bear:" "🧸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:teddy_bear:" "🧸" nil :system t)
    (puthash "\\:broom:" "🧹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:broom:" "🧹" nil :system t)
    (puthash "\\:basket:" "🧺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:basket:" "🧺" nil :system t)
    (puthash "\\:roll_of_paper:" "🧻" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:roll_of_paper:" "🧻" nil :system t)
    (puthash "\\:soap:" "🧼" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:soap:" "🧼" nil :system t)
    (puthash "\\:sponge:" "🧽" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sponge:" "🧽" nil :system t)
    (puthash "\\:receipt:" "🧾" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:receipt:" "🧾" nil :system t)
    (puthash "\\:nazar_amulet:" "🧿" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nazar_amulet:" "🧿" nil :system t)
    (puthash "\\:ballet_shoes:" "🩰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ballet_shoes:" "🩰" nil :system t)
    (puthash "\\:one-piece_swimsuit:" "🩱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:one-piece_swimsuit:" "🩱" nil :system t)
    (puthash "\\:briefs:" "🩲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:briefs:" "🩲" nil :system t)
    (puthash "\\:shorts:" "🩳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:shorts:" "🩳" nil :system t)
    (puthash "\\:thong_sandal:" "🩴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:thong_sandal:" "🩴" nil :system t)
    (puthash "\\:drop_of_blood:" "🩸" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:drop_of_blood:" "🩸" nil :system t)
    (puthash "\\:adhesive_bandage:" "🩹" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:adhesive_bandage:" "🩹" nil :system t)
    (puthash "\\:stethoscope:" "🩺" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:stethoscope:" "🩺" nil :system t)
    (puthash "\\:yo-yo:" "🪀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:yo-yo:" "🪀" nil :system t)
    (puthash "\\:kite:" "🪁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:kite:" "🪁" nil :system t)
    (puthash "\\:parachute:" "🪂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:parachute:" "🪂" nil :system t)
    (puthash "\\:boomerang:" "🪃" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:boomerang:" "🪃" nil :system t)
    (puthash "\\:magic_wand:" "🪄" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:magic_wand:" "🪄" nil :system t)
    (puthash "\\:pinata:" "🪅" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:pinata:" "🪅" nil :system t)
    (puthash "\\:nesting_dolls:" "🪆" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:nesting_dolls:" "🪆" nil :system t)
    (puthash "\\:ringed_planet:" "🪐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ringed_planet:" "🪐" nil :system t)
    (puthash "\\:chair:" "🪑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:chair:" "🪑" nil :system t)
    (puthash "\\:razor:" "🪒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:razor:" "🪒" nil :system t)
    (puthash "\\:axe:" "🪓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:axe:" "🪓" nil :system t)
    (puthash "\\:diya_lamp:" "🪔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:diya_lamp:" "🪔" nil :system t)
    (puthash "\\:banjo:" "🪕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:banjo:" "🪕" nil :system t)
    (puthash "\\:military_helmet:" "🪖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:military_helmet:" "🪖" nil :system t)
    (puthash "\\:accordion:" "🪗" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:accordion:" "🪗" nil :system t)
    (puthash "\\:long_drum:" "🪘" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:long_drum:" "🪘" nil :system t)
    (puthash "\\:coin:" "🪙" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:coin:" "🪙" nil :system t)
    (puthash "\\:carpentry_saw:" "🪚" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:carpentry_saw:" "🪚" nil :system t)
    (puthash "\\:screwdriver:" "🪛" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:screwdriver:" "🪛" nil :system t)
    (puthash "\\:ladder:" "🪜" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:ladder:" "🪜" nil :system t)
    (puthash "\\:hook:" "🪝" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:hook:" "🪝" nil :system t)
    (puthash "\\:mirror:" "🪞" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mirror:" "🪞" nil :system t)
    (puthash "\\:window:" "🪟" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:window:" "🪟" nil :system t)
    (puthash "\\:plunger:" "🪠" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:plunger:" "🪠" nil :system t)
    (puthash "\\:sewing_needle:" "🪡" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:sewing_needle:" "🪡" nil :system t)
    (puthash "\\:knot:" "🪢" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:knot:" "🪢" nil :system t)
    (puthash "\\:bucket:" "🪣" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bucket:" "🪣" nil :system t)
    (puthash "\\:mouse_trap:" "🪤" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:mouse_trap:" "🪤" nil :system t)
    (puthash "\\:toothbrush:" "🪥" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:toothbrush:" "🪥" nil :system t)
    (puthash "\\:headstone:" "🪦" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:headstone:" "🪦" nil :system t)
    (puthash "\\:placard:" "🪧" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:placard:" "🪧" nil :system t)
    (puthash "\\:rock:" "🪨" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:rock:" "🪨" nil :system t)
    (puthash "\\:fly:" "🪰" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fly:" "🪰" nil :system t)
    (puthash "\\:worm:" "🪱" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:worm:" "🪱" nil :system t)
    (puthash "\\:beetle:" "🪲" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:beetle:" "🪲" nil :system t)
    (puthash "\\:cockroach:" "🪳" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:cockroach:" "🪳" nil :system t)
    (puthash "\\:potted_plant:" "🪴" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:potted_plant:" "🪴" nil :system t)
    (puthash "\\:wood:" "🪵" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:wood:" "🪵" nil :system t)
    (puthash "\\:feather:" "🪶" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:feather:" "🪶" nil :system t)
    (puthash "\\:anatomical_heart:" "🫀" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:anatomical_heart:" "🫀" nil :system t)
    (puthash "\\:lungs:" "🫁" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:lungs:" "🫁" nil :system t)
    (puthash "\\:people_hugging:" "🫂" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:people_hugging:" "🫂" nil :system t)
    (puthash "\\:blueberries:" "🫐" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:blueberries:" "🫐" nil :system t)
    (puthash "\\:bell_pepper:" "🫑" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:bell_pepper:" "🫑" nil :system t)
    (puthash "\\:olive:" "🫒" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:olive:" "🫒" nil :system t)
    (puthash "\\:flatbread:" "🫓" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:flatbread:" "🫓" nil :system t)
    (puthash "\\:tamale:" "🫔" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:tamale:" "🫔" nil :system t)
    (puthash "\\:fondue:" "🫕" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:fondue:" "🫕" nil :system t)
    (puthash "\\:teapot:" "🫖" table)
    (define-abbrev julia-latexsub-abbrev-table "\\:teapot:" "🫖" nil :system t)
    table))

(provide 'julia-mode-latexsubs)
